/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, Clipboard routines */
#include "teraterm.h"
#include "tttypes.h"
#include "ttwinman.h"
#include "buffer.h"
#include "ttcommon.h"

HANDLE CBMemHandle = 0;
PCHAR CBMemPtr = NULL;
static BOOL CBAddCR = FALSE;
static BYTE CBByte;
BOOL CBRetrySend;
BOOL CBRetryEcho;
BOOL CBSendCR;
BOOL CBDDE;

void CBCopy(HWND HWin)
{
  LONG MemSize, TmpPtr;
  int i, j, k, IStart, IEnd;

  if (! Select) return;
  if (TalkStatus==IdTalkCB) return;

  MemSize = (SelectEnd.y-SelectStart.y)*(NumOfColumns+2)+
    SelectEnd.x-SelectStart.x+1;
  CBMemHandle = GlobalAlloc(GMEM_MOVEABLE, MemSize);
  if (CBMemHandle == 0) MessageBeep(0);
  else {
    CBMemPtr = GlobalLock(CBMemHandle);
    if (CBMemPtr == NULL)
    {
      GlobalFree(CBMemHandle);
      MessageBeep(0);
    }
    else {
      TmpPtr = GetLinePtr(SelectStart.y);
      k = 0;
      for (j = SelectStart.y ; j<=SelectEnd.y ; j++)
      {
        IStart = 0;
        IEnd = NumOfColumns-1;
        if (j==SelectStart.y) IStart = SelectStart.x;
        if (j==SelectEnd.y) IEnd = SelectEnd.x-1;

        if ((IStart>0) && (IStart < NumOfColumns-1) &&
            ((AttrBuff[TmpPtr+IStart-1] & AttrKanji) != 0)) /* DBCS first byte? */
          IStart++;

        while ((IEnd>=0) && (CodeBuff[TmpPtr+IEnd]==0x20))
          IEnd--;

        if ((IEnd>=0) && (IEnd < NumOfColumns-1) &&
            ((AttrBuff[TmpPtr+IEnd] & AttrKanji) != 0)) /* DBCS first byte? */
          IEnd++;

        for (i = IStart ; i <= IEnd ; i++)
        {
          CBMemPtr[k] = CodeBuff[TmpPtr+i];
          k++;
        }
        if (j < SelectEnd.y)
        {
          CBMemPtr[k] = 0x0d;
          k++;
          CBMemPtr[k] = 0x0a;
          k++;
        }

        TmpPtr = NextLinePtr(TmpPtr);
      }
      CBMemPtr[k] = 0;

      GlobalUnlock(CBMemHandle);

      if (OpenClipboard(HWin))
      {
        EmptyClipboard();
        SetClipboardData(CF_TEXT, CBMemHandle);
        CloseClipboard();
      }

    }
  }
}

void CBStartPaste(HWND HWin, BOOL AddCR)
{
  UINT Cf;

  if (! cv.Ready) return;
  if (TalkStatus!=IdTalkKeyb) return;

  CBAddCR = AddCR;
  if (IsClipboardFormatAvailable(CF_TEXT))
    Cf = CF_TEXT;
  else if (IsClipboardFormatAvailable(CF_OEMTEXT))
    Cf = CF_OEMTEXT;
  else return;

  if (OpenClipboard(HWin))
  {
    CBMemHandle = GetClipboardData(Cf);
    if (CBMemHandle != 0)
    {
      CBMemPtr = GlobalLock(CBMemHandle);
      if (CBMemPtr != NULL) TalkStatus = IdTalkCB;
    }
  }
  CBRetrySend = FALSE;
  CBRetryEcho = FALSE;
  CBSendCR = FALSE;
  CBDDE = FALSE;
  if (TalkStatus != IdTalkCB) CBEndPaste();
}

void CBSend()
{
  int c;

  if (CBMemPtr == NULL) return;

  if (CBRetrySend)
  {
    CBRetryEcho = (ts.LocalEcho>0);
    c = CommTextOut(&cv,(PCHAR)&CBByte,1);
    CBRetrySend = (c==0);
    if (CBRetrySend) return;
  }

  if (CBRetryEcho)
  {
    c = CommTextEcho(&cv,(PCHAR)&CBByte,1);
    CBRetryEcho = (c==0);
    if (CBRetryEcho) return;
  }

  do {
    if (CBSendCR && (CBMemPtr[0]==0x0a))
      CBMemPtr = &CBMemPtr[1];

    if (CBMemPtr[0]!=0)
    {
      CBByte = CBMemPtr[0];
      CBMemPtr = &CBMemPtr[1];
      if (CBByte==0xFF) /* FF from TTMACRO */
      {
        CBByte = CBMemPtr[0];
        CBMemPtr = &CBMemPtr[1];
        if ((CBByte & 1) == 0)
          CBByte = 0;
        else
          CBByte = 0xFF;
      }
    }
    else if (CBAddCR)
    {
      CBAddCR = FALSE;
      CBByte = 0x0d;
    }
    else {
      CBEndPaste();
      return;
    }

    c = CommTextOut(&cv,(PCHAR)&CBByte,1);
    CBSendCR = (CBByte==0x0D);
    CBRetrySend = (c==0);
    if (CBRetrySend) return;
    if (ts.LocalEcho>0)
    {
      c = CommTextEcho(&cv,(PCHAR)&CBByte,1);
      CBRetryEcho = (c==0);
      if (CBRetryEcho) return;
    }
  }
  while (c>0);
}

void CBEndPaste()
{
  TalkStatus = IdTalkKeyb;
  if (CBMemHandle!=0) GlobalUnlock(CBMemHandle);

  if (CBDDE)
  {
    if (CBMemHandle!=0) GlobalFree(CBMemHandle);
  }
  else
    CloseClipboard();

  CBDDE = FALSE;
  CBMemHandle = 0;
  CBMemPtr = NULL;
  CBAddCR = FALSE;
}

