/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, file transfer routines */
#include "teraterm.h"
#include "tttypes.h"
#include "ttftypes.h"
#include "ttwinman.h"
#include "commlib.h"
#include "ttcommon.h"
#include "ttdde.h"

#include "filesys.h"

PFileVar LogVar = NULL;
PFileVar SendVar = NULL;
PFileVar FileVar = NULL;
static BYTE LogLast = 0;
BOOL FileLog = FALSE;
BOOL BinLog = FALSE;
BOOL DDELog = FALSE;
BOOL FileRetrySend, FileRetryEcho, FileCRSend;
BYTE FileByte;

static BOOL FSend = FALSE;
BOOL ProtoFlag = FALSE;

static HMODULE HTTFILE = NULL;
static int TTFILECount = 0;

PGetSetupFname GetSetupFname;
PGetTransFname GetTransFname;
PGetMultiFname GetMultiFname;
PGetGetFname GetGetFname;
PSetFileVar SetFileVar;
PGetXFname GetXFname;
PProtoInit ProtoInit;
PProtoParse ProtoParse;
PProtoTimeOutProc ProtoTimeOutProc;
PProtoCancel ProtoCancel;
PVTPrint VTPrint;

#define Id_FileTransDlg 2100
#define Id_TransFname 2101
#define Id_TransBytes 2102
#define Id_TransPauseStart 2103
#define Id_TransHelp 2199

#define IdTTIcon 100

#define IdGetSetupFname  1
#define IdGetTransFname  2
#define IdGetMultiFname  3
#define IdGetGetFname    4
#define IdSetFileVar     5
#define IdGetXFname      6

#define IdProtoInit      7
#define IdProtoParse	 8
#define IdProtoTimeOutProc 9
#define IdProtoCancel    10

#define IdVTPrint        20

BOOL LoadTTFILE()
{
  BOOL Err;

  if (HTTFILE != NULL)
  {
    TTFILECount++;
    return TRUE;
  }
  else
    TTFILECount = 0;

  HTTFILE = LoadLibrary("TTPFILE.DLL");
  if (HTTFILE == NULL) return FALSE;

  Err = FALSE;
  GetSetupFname = (PGetSetupFname)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdGetSetupFname));
  if (GetSetupFname==NULL) Err = TRUE;

  GetTransFname = (PGetTransFname)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdGetTransFname));
  if (GetTransFname==NULL) Err = TRUE;

  GetMultiFname = (PGetMultiFname)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdGetMultiFname));
  if (GetMultiFname==NULL) Err = TRUE;

  GetGetFname = (PGetGetFname)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdGetGetFname));
  if (GetGetFname==NULL) Err = TRUE;

  SetFileVar = (PSetFileVar)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdSetFileVar));
  if (SetFileVar==NULL) Err = TRUE;

  GetXFname = (PGetXFname)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdGetXFname));
  if (GetXFname==NULL) Err = TRUE;

  ProtoInit = (PProtoInit)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdProtoInit));
  if (ProtoInit==NULL) Err = TRUE;

  ProtoParse = (PProtoParse)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdProtoParse));
  if (ProtoParse==NULL) Err = TRUE;

  ProtoTimeOutProc = (PProtoTimeOutProc)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdProtoTimeOutProc));
  if (ProtoTimeOutProc==NULL) Err = TRUE;

  ProtoCancel = (PProtoCancel)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdProtoCancel));
  if (ProtoCancel==NULL) Err = TRUE;

  VTPrint = (PVTPrint)GetProcAddress(HTTFILE,
    MAKEINTRESOURCE(IdVTPrint));
  if (VTPrint==NULL) Err = TRUE;

  if (Err)
  {
    FreeLibrary(HTTFILE);
    HTTFILE = NULL;
    return FALSE; 
 }
  else {
    TTFILECount = 1;
    return TRUE;
  }
}

BOOL FreeTTFILE()
{
  if (TTFILECount==0) return FALSE;
  TTFILECount--;
  if ((TTFILECount==0) && (HTTFILE!=NULL))
  {
    FreeLibrary(HTTFILE);
    HTTFILE = NULL;
  }
  return TRUE;
}

/* text file transfer dialog box */
class CFileTransDlg : public CDialog
{
private:
  BOOL LogFlag;
  PFileVar fv;

public:
  BOOL Pause;
  BOOL Create(PFileVar pfv);

// Dialog Data
  //{{AFX_DATA(CFileTransDlg)
    enum { IDD = Id_FileTransDlg };
	// NOTE: the ClassWizard will add data members here
  //}}AFX_DATA

  void ChangeButton();
  void RefreshNum();

  virtual BOOL DestroyWindow();
protected:
  virtual BOOL OnCommand(WPARAM wParam, LPARAM lParam);
  virtual BOOL OnInitDialog();

  //{{AFX_MSG( CFileTransDlg )
  //}}AFX_MSG

  DECLARE_MESSAGE_MAP()
};
typedef CFileTransDlg *PFileTransDlg;

BEGIN_MESSAGE_MAP( CFileTransDlg, CDialog )
  //{{AFX_MSG_MAP( CFileTransDlg )
  //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/* File transfer protocol dialog box */
class CProtoDlg : public CDialog
{
private:
  int Proto;
  BOOL LogFlag;
  PFileVar fv;
  PCHAR pv;

public:
  BOOL Create
    (PFileVar pfv, int ProtoId, int Mode, WORD Opt1, WORD Opt2);
// Dialog Data
  //{{AFX_DATA(CProtoDlg)
    enum { IDD = Id_ProtoDialog };
	// NOTE: the ClassWizard will add data members here
  //}}AFX_DATA

  int Parse();
  void TimeOutProc();

protected:
  virtual BOOL OnCommand(WPARAM wParam, LPARAM lParam);
  virtual BOOL OnInitDialog();
  virtual void OnCancel();

  //{{AFX_MSG( CProtoDlg )
    afx_msg void OnDestroy();
  //}}AFX_MSG

  DECLARE_MESSAGE_MAP()
};
typedef CProtoDlg *PProtoDlg;

BEGIN_MESSAGE_MAP( CProtoDlg, CDialog )
  //{{AFX_MSG_MAP( CProtoDlg )
  ON_WM_DESTROY()
  //}}AFX_MSG_MAP
END_MESSAGE_MAP()

PFileTransDlg FLogDlg = NULL;
PFileTransDlg SendDlg = NULL;
PProtoDlg ProtoDlg = NULL;

BOOL CFileTransDlg::Create(PFileVar pfv)
{
  BOOL Ok;

  fv = pfv;
  LogFlag = (fv->OpId==OpLog);
  Pause = FALSE;
  Ok = CDialog::Create(CFileTransDlg::IDD, NULL);
  SetClassLong(GetSafeHwnd(),GCL_HICON,
    (LONG)LoadIcon(AfxGetInstanceHandle(),MAKEINTRESOURCE(IdTTIcon)));
  return Ok;
}

void CFileTransDlg::ChangeButton()
{
  if (Pause)
    SetDlgItemText(Id_TransPauseStart, "&Start");
  else
    SetDlgItemText(Id_TransPauseStart, "Pau&se");

  cv.FilePause =
    (FileVar!=NULL) ||
    LogFlag && Pause;
}

void CFileTransDlg::RefreshNum()
{
  char NumStr[13];

  sprintf(NumStr,"%u",fv->ByteCount);
  SetDlgItemText(Id_TransBytes, NumStr);
}

BOOL CFileTransDlg::DestroyWindow()
{
  if (LogFlag)
  {
    FLogDlg = NULL;
    LogEnd();
  }
  else {
    SendDlg = NULL;
    FileSendEnd();
  }

  return (CDialog::DestroyWindow());
}

BOOL CFileTransDlg::OnCommand(WPARAM wParam, LPARAM lParam)
{
  switch (wParam) {
    case IDCANCEL:
      DestroyWindow();
      return TRUE;
    case Id_TransPauseStart:
      Pause = ! Pause;
      ChangeButton();
      return TRUE;
    case Id_TransHelp:
      ::PostMessage(HVTWin,WM_USER+MsgDlgHelp2,0,0);
      return TRUE;
    default:
      return (CDialog::OnCommand(wParam,lParam));
  }
}

BOOL CFileTransDlg::OnInitDialog()
{
  SetWindowText(fv->DlgCaption);
  SetDlgItemText(Id_TransFname, &(fv->FullName[fv->DirLen]));
  return TRUE;
}

BOOL CProtoDlg::Create(PFileVar pfv, int ProtoId, int Mode, WORD Opt1, WORD Opt2)
{
  int vsize;

  fv = pfv;
  Proto = ProtoId;
  switch (Proto) {
    case PROTO_KMT:
      vsize = sizeof(TKmtVar);
      break;
    case PROTO_XM:
      vsize = sizeof(TXVar);
      break;
    case PROTO_ZM:
      vsize = sizeof(TZVar);
      break;
    case PROTO_BP:
      vsize = sizeof(TBPVar);
      break;
    case PROTO_QV:
      vsize = sizeof(TQVVar);
      break;
  }
  pv = (PCHAR)malloc(vsize);
  if (pv!=NULL)
    switch (Proto) {
      case PROTO_KMT:
        ((PKmtVar)pv)->KmtMode = Mode;
        break;
      case PROTO_XM:
	((PXVar)pv)->XMode = Mode;
	((PXVar)pv)->XOpt = Opt1;
	((PXVar)pv)->TextFlag = 1 - (Opt2 & 1);
	break;
      case PROTO_ZM:
	((PZVar)pv)->BinFlag = (Opt1 & 1) != 0;
	((PZVar)pv)->ZMode = Mode;
	break;
      case PROTO_BP:
	((PBPVar)pv)->BPMode = Mode;
	break;
      case PROTO_QV:
	((PQVVar)pv)->QVMode = Mode;
	break;
    }

  return (CDialog::Create(CProtoDlg::IDD, NULL));
}

void CProtoDlg::OnDestroy()
{
  ::KillTimer(fv->HMainWin,IdProtoTimer);
  if ((Proto==PROTO_QV) &&
      (((PQVVar)pv)->QVMode==IdQVSend))
    CommTextOut(&cv,"\015",1);
  if (fv->LogFlag)
    _lclose(fv->LogFile); 
  fv->LogFile = NULL;
  if (pv!=NULL) free(pv);
  ProtoDlg = NULL;
  ProtoEnd();
  CDialog::OnDestroy();
}

BOOL CProtoDlg::OnCommand(WPARAM wParam, LPARAM lParam)
{
  switch (wParam) {
    case Id_TransHelp:
      ::PostMessage(HVTWin,WM_USER+MsgDlgHelp2,0,0);
      return TRUE;
    default:
      return (CDialog::OnCommand(wParam,lParam));
  }
}

void CProtoDlg::OnCancel()
{
  if ((*ProtoCancel)(Proto,fv,pv,&cv))
    DestroyWindow();
}

BOOL CProtoDlg::OnInitDialog()
{
  if (pv==NULL)
  {
    CDialog::DestroyWindow();
    return TRUE;
  }
  fv->HWin = GetSafeHwnd();

  (*ProtoInit)(Proto,fv,pv,&cv,&ts);
  return TRUE;
}

int CProtoDlg::Parse()
{
  int P;

  P = ActiveWin;
  if ((*ProtoParse)(Proto,fv,pv,&cv))
    P = 0; /* continue */
  else {
    CommSend(&cv);
    DestroyWindow();
  }
  return P;
}

void CProtoDlg::TimeOutProc()
{
  (*ProtoTimeOutProc)(Proto,fv,pv,&cv);
}

BOOL OpenFTDlg(PFileVar fv)
{
  PFileTransDlg FTDlg;

  FTDlg = new CFileTransDlg();
  if (fv->OpId==OpLog)
    FLogDlg = FTDlg; /* Log */
  else
    SendDlg = FTDlg; /* File send */
  
  if (FTDlg!=NULL)
  {
    FTDlg->Create(fv);
    FTDlg->RefreshNum();
    if (fv->OpId == OpLog)
      FTDlg->ShowWindow(SW_MINIMIZE);
  }
  return (FTDlg!=NULL);
}

BOOL NewFileVar(PFileVar *FV)
{
  if ((*FV)==NULL)
  {
    *FV = (PFileVar)malloc(sizeof(TFileVar));
    if ((*FV)!=NULL)
    {
      memset(*FV, 0, sizeof(TFileVar));        
      strcpy((*FV)->FullName,ts.FileDir);
      (*FV)->DirLen = strlen((*FV)->FullName);
      if ((*FV)->FullName[(*FV)->DirLen-1] != '\\')
      {
        strcat((*FV)->FullName,"\\");
        (*FV)->DirLen++;
      }
      (*FV)->FileOpen = FALSE;
      (*FV)->HMainWin = HVTWin;
    }
  }

  return ((*FV)!=NULL);
}

void FreeFileVar(PFileVar *FV)
{
  if ((*FV)!=NULL)
  {
    if ((*FV)->FileOpen) _lclose((*FV)->FileHandle);
    if ((*FV)->FnStrMemHandle>0)
    {
      GlobalUnlock((*FV)->FnStrMemHandle);
      GlobalFree((*FV)->FnStrMemHandle);
    }
    free(*FV);
    *FV = NULL;
  }
}

extern "C" {
void LogStart()
{
  LONG Option;

  if ((FileLog) || (BinLog)) return;

  if (! LoadTTFILE()) return;
  if (! NewFileVar(&LogVar))
  {
    FreeTTFILE();
    return;
  }
  LogVar->OpId = OpLog;

  if (strlen(&(LogVar->FullName[LogVar->DirLen]))==0)
  {
    Option = MAKELONG(ts.TransBin,ts.Append);
    if (! (*GetTransFname)(HVTWin, GTF_LOG, &Option, LogVar, ts.FileDir))
    {
      LogEnd();
      return;
    }
    ts.TransBin = LOWORD(Option);
    ts.Append = HIWORD(Option);
  }
  else
    (*SetFileVar)(LogVar);

  if (ts.TransBin > 0)
  {
    if (! CreateBinBuf())
    {
      LogEnd();
      return;
    }
    BinLog = TRUE;
    FileLog = FALSE;
  }
  else {
    if (! CreateLogBuf())
    {
      LogEnd();
      return;
    }
    BinLog = FALSE;
    FileLog = TRUE;
  }
  cv.LStart = cv.LogPtr;
  cv.LCount = 0;

  HelpId = HlpFileLog;
  if (ts.Append>0)
  {
    LogVar->FileHandle = _lopen(LogVar->FullName,OF_WRITE);
    if (LogVar->FileHandle>0)
      _llseek(LogVar->FileHandle,0,2);
    else
      LogVar->FileHandle = _lcreat(LogVar->FullName,0);
  }
  else
    LogVar->FileHandle = _lcreat(LogVar->FullName,0);
  LogVar->FileOpen = (LogVar->FileHandle>0);
  if (! LogVar->FileOpen)
  {
    LogEnd();
    return;
  }
  LogVar->ByteCount = 0;

  if (! OpenFTDlg(LogVar))
    LogEnd();
}
}

void LogEnd()
{
  if (FLogDlg!=NULL)
  {
    FLogDlg->DestroyWindow();
    FLogDlg = NULL;
  }

  FreeTTFILE();
  FreeFileVar(&LogVar);

  FileLog = FALSE;
  FreeLogBuf();
  BinLog = FALSE;
  FreeBinBuf();
}

void LogPut1(BYTE b)
{
  LogLast = b;
  cv.LogBuf[cv.LogPtr] = b;
  cv.LogPtr++;
  if (cv.LogPtr>=InBuffSize)
    cv.LogPtr = cv.LogPtr-InBuffSize;
  if (cv.LCount>=InBuffSize)
  {
    cv.LCount = InBuffSize;
    cv.LStart = cv.LogPtr;
  }
  else cv.LCount++;
  if (cv.DCount>=InBuffSize)
  {
    cv.DCount = InBuffSize;
    cv.DStart = cv.LogPtr;
  }
  else cv.DCount++;
}

void Log1Byte(BYTE b)
{
  if (b==0x0d)
  {
    LogLast = b;
    return;
  }
  if ((b==0x0a) && (LogLast==0x0d))
    LogPut1(0x0d);
  LogPut1(b);
}

  static BOOL Get1(PCHAR Buf, int *Start, int *Count, PBYTE b)
  {
    if (*Count<=0) return FALSE;
    *b = Buf[*Start];
    (*Start)++;
    if (*Start>=InBuffSize)
      *Start = *Start-InBuffSize;
    (*Count)--;
    return TRUE;
  }

void LogToFile()
{
  PCHAR Buf;
  int Start, Count;
  BYTE b;

  if (! LogVar->FileOpen) return;
  if (FileLog)
  {
    Buf = cv.LogBuf;
    Start = cv.LStart;
    Count = cv.LCount;
  }
  else if (BinLog)
  {
    Buf = cv.BinBuf;
    Start = cv.BStart;
    Count = cv.BCount;
  }
  else
    return;

  if (Buf==NULL) return;
  if (Count==0) return;

  while (Get1(Buf,&Start,&Count,&b))
  {
    if (! cv.FilePause)
    {
      _lwrite(LogVar->FileHandle,(PCHAR)&b,1);
      (LogVar->ByteCount)++;
    }
  }

  if (FileLog)
  {
    cv.LStart = Start;
    cv.LCount = Count;
  }
  else {
    cv.BStart = Start;
    cv.BCount = Count;
  }
  if (cv.FilePause) return;
  if (FLogDlg!=NULL)
    FLogDlg->RefreshNum();
}

BOOL CreateLogBuf()
{
  if (cv.Ready && (cv.HLogBuf==0))
  {
    cv.HLogBuf = GlobalAlloc(GMEM_MOVEABLE,InBuffSize);
    cv.LogBuf = NULL;
    cv.LogPtr = 0;
    cv.LStart = 0;
    cv.LCount = 0;
    cv.DStart = 0;
    cv.DCount = 0;
  }
  return (cv.Ready && (cv.HLogBuf!=0));
}

void FreeLogBuf()
{
  if ((cv.HLogBuf==0) || FileLog || DDELog) return;
  if (cv.LogBuf!=NULL) GlobalUnlock(cv.HLogBuf);
  GlobalFree(cv.HLogBuf);
  cv.HLogBuf = 0;
  cv.LogBuf = NULL;
  cv.LogPtr = 0;
  cv.LStart = 0;
  cv.LCount = 0;
  cv.DStart = 0;
  cv.DCount = 0;
}

BOOL CreateBinBuf()
{
  if (cv.Ready && (cv.HBinBuf==0))
  {
    cv.HBinBuf = GlobalAlloc(GMEM_MOVEABLE,InBuffSize);
    cv.BinBuf = NULL;
    cv.BinPtr = 0;
    cv.BStart = 0;
    cv.BCount = 0;
  }
  return (cv.Ready && (cv.HBinBuf!=0));
}

void FreeBinBuf()
{
  if ((cv.HBinBuf==0) || BinLog) return;
  if (cv.BinBuf!=NULL) GlobalUnlock(cv.HBinBuf);
  GlobalFree(cv.HBinBuf);
  cv.HBinBuf = 0;
  cv.BinBuf = NULL;
  cv.BinPtr = 0;
  cv.BStart = 0;
  cv.BCount = 0;
}

extern "C" {
void FileSendStart()
{
  LONG Option;

  if (FSend) return;
  if (ProtoFlag)
  {
    FreeFileVar(&SendVar);
    return;
  }

  if (! LoadTTFILE()) return;
  if (! NewFileVar(&SendVar))
  {
    FreeTTFILE();
    return;
  }
  SendVar->OpId = OpSendFile;

  FSend = TRUE;

  if (strlen(&(SendVar->FullName[SendVar->DirLen]))==0)
  {
    Option = MAKELONG(ts.TransBin,0);
    if (! (*GetTransFname)(HVTWin, GTF_SEND, &Option, SendVar, ts.FileDir))
    {
      FileSendEnd();
      return;
    }
    ts.TransBin = LOWORD(Option);
  }
  else
    (*SetFileVar)(SendVar);

  SendVar->FileHandle = _lopen(SendVar->FullName,OF_READ);
  SendVar->FileOpen = (SendVar->FileHandle>0);
  if (! SendVar->FileOpen)
  {
    FileSendEnd();
    return;
  }
  SendVar->ByteCount = 0;

  TalkStatus = IdTalkFile;
  FileRetrySend = FALSE;
  FileRetryEcho = FALSE;
  FileCRSend = FALSE;

  if (! OpenFTDlg(SendVar))
    FileSendEnd();
}
}

void FileSendEnd()
{
  if (! FSend) return;
  TalkStatus = IdTalkKeyb;

  if (SendDlg!=NULL)
  {
    SendDlg->DestroyWindow();
    SendDlg = NULL;
  }

  FreeTTFILE();
  FreeFileVar(&SendVar);
  FSend = FALSE;
  EndDdeCmnd();
}

  int FSOut1(BYTE b)
  {
    if (ts.TransBin > 0)
      return CommBinaryOut(&cv,(PCHAR)&b,1);
    else if ((b>=0x20) || (b==0x09) || (b==0x0A) || (b==0x0D))
      return CommTextOut(&cv,(PCHAR)&b,1);
    else
      return 1;
  }

  int FSEcho1(BYTE b)
  {
    if (ts.TransBin > 0)
      return CommBinaryEcho(&cv,(PCHAR)&b,1);
    else
      return CommTextEcho(&cv,(PCHAR)&b,1);
  }

extern "C" {
void FileSend()
{
  WORD c, fc;

  if ((SendDlg==NULL) || (SendDlg->Pause)) return;

  if (FileRetrySend)
  {
    FileRetryEcho = (ts.LocalEcho>0);
    c = FSOut1(FileByte);
    FileRetrySend = (c==0);
    if (FileRetrySend) return;
  }

  if (FileRetryEcho)
  {
    c = FSEcho1(FileByte);
    FileRetryEcho = (c==0);
    if (FileRetryEcho) return;
  }

  SendDlg->RefreshNum();  

  do {
    fc = _lread(SendVar->FileHandle,&FileByte,1);
    SendVar->ByteCount = SendVar->ByteCount + fc;

    if (FileCRSend && (fc==1) && (FileByte==0x0A))
    {
      fc = _lread(SendVar->FileHandle,&FileByte,1);
      SendVar->ByteCount = SendVar->ByteCount + fc;
    }

    if (fc!=0)
    {
      c = FSOut1(FileByte);
      FileCRSend = (ts.TransBin==0) && (FileByte==0x0D);
      FileRetrySend = (c==0);
      if (FileRetrySend) return;
      if (ts.LocalEcho>0)
      {
        c = FSEcho1(FileByte);
        FileRetryEcho = (c==0);
        if (FileRetryEcho) return;
      }
    }
  } while (fc!=0);

  if (fc==0) FileSendEnd();
}
}

extern "C" {
void FLogChangeButton(BOOL Pause)
{
  if (FLogDlg!=NULL)
  {
    FLogDlg->Pause = Pause;
    FLogDlg->ChangeButton();
  }  
}
}

extern "C" {
void FLogRefreshNum()
{
  if (FLogDlg!=NULL)
    FLogDlg->RefreshNum();
}
}

BOOL OpenProtoDlg(PFileVar fv, int IdProto, int Mode, WORD Opt1, WORD Opt2)
{
  ProtoDlg = new CProtoDlg();
  if (ProtoDlg!=NULL)
  {
    ProtoDlg->Create(fv,IdProto,Mode,Opt1,Opt2);
  }
  return (ProtoDlg!=NULL);
}

extern "C" {
void CloseProtoDlg()
{
  if (ProtoDlg!=NULL)
    ProtoDlg->DestroyWindow();
  ProtoDlg = NULL;
  cv.FilePause =
    (FileVar!=NULL) ||
    (FLogDlg!=NULL) && FLogDlg->Pause;
}
}

BOOL ProtoStart()
{
  if (ProtoFlag) return FALSE;
  if (FSend)
  {
    FreeFileVar(&FileVar);
    return FALSE;
  }

  if (! LoadTTFILE()) return FALSE;
  NewFileVar(&FileVar);
  cv.FilePause =
    (FileVar!=NULL) ||
    (FLogDlg!=NULL) && FLogDlg->Pause;

  if (FileVar==NULL)
  {
    FreeTTFILE();
    return FALSE;
  }
  ProtoFlag = TRUE;
  return TRUE;
}

void ProtoEnd()
{
  if (! ProtoFlag) return;

  /* Enable transmit delay (serial port) */
  cv.DelayFlag = TRUE;
  TalkStatus = IdTalkKeyb;

  CloseProtoDlg();

  FreeTTFILE();
  FreeFileVar(&FileVar);
  cv.FilePause =
    (FileVar!=NULL) ||
    (FLogDlg!=NULL) && FLogDlg->Pause;

  ProtoFlag = FALSE;
  EndDdeCmnd();
}

extern "C" {
int ProtoDlgParse()
{
  if (ProtoDlg!=NULL)
    return (ProtoDlg->Parse());
  else
    return (ActiveWin);
}
}

extern "C" {
void ProtoDlgTimeOut()
{
  if (ProtoDlg!=NULL)
    ProtoDlg->TimeOutProc();
}
}

extern "C" {
void KermitStart(int mode)
{
  WORD w;

  if (! ProtoStart()) return;

  switch (mode) {
    case IdKmtSend:
      if (strlen(&FileVar->FullName[FileVar->DirLen])==0)
      {
        if (! (*GetMultiFname)(HVTWin,FileVar,ts.FileDir,GMF_KERMIT,&w) ||
            (FileVar->NumFname==0))
        {
          ProtoEnd();
          return;
        }
      }
      else {
        FileVar->OpId = OpKmtSend;
        PSetFileVar(FileVar);
      }
      break;
    case IdKmtReceive:
      FileVar->OpId = OpKmtRcv;
      break;
    case IdKmtGet:
      if (! (*GetGetFname)(FileVar->HMainWin,FileVar) || (strlen(FileVar->FullName)==0))
      {
        ProtoEnd();
        return;
      }
      FileVar->OpId = OpKmtGet;
      break;
    case IdKmtFinish:
      FileVar->OpId = OpKmtFin;
      break;
    default:
      ProtoEnd();
      return;
  }
  TalkStatus = IdTalkQuiet;

  /* disable transmit delay (serial port) */
  cv.DelayFlag = FALSE;

  if (! OpenProtoDlg(FileVar,PROTO_KMT,mode,0,0))
    ProtoEnd();
}
}

extern "C" {
void XMODEMStart(int mode)
{
  LONG Option;

  if (! ProtoStart()) return;

  if (mode==IdXReceive)
    FileVar->OpId = OpXRcv;
  else
    FileVar->OpId = OpXSend;

  if (strlen(&(FileVar->FullName[FileVar->DirLen]))==0)
  {
    Option = MAKELONG(ts.XmodemBin,ts.XmodemOpt);
    if (! (*GetXFname)(FileVar->HMainWin,mode==IdXReceive,&Option,FileVar,ts.FileDir))
    {
      ProtoEnd();
      return;
    }
    ts.XmodemOpt = HIWORD(Option);
    ts.XmodemBin = LOWORD(Option);
  }
  else
    (*SetFileVar)(FileVar);

  if (mode==IdXReceive)
    FileVar->FileHandle = _lcreat(FileVar->FullName,0);
  else
    FileVar->FileHandle = _lopen(FileVar->FullName,OF_READ);

  FileVar->FileOpen = FileVar->FileHandle>0;
  if (! FileVar->FileOpen)
  {
    ProtoEnd();
    return;
  }
  TalkStatus = IdTalkQuiet;

  /* disable transmit delay (serial port) */
  cv.DelayFlag = FALSE;

  if (! OpenProtoDlg(FileVar,PROTO_XM,mode,ts.XmodemOpt,ts.XmodemBin))
    ProtoEnd();
}
}

extern "C" {
void ZMODEMStart(int mode)
{
  WORD Opt;

  if (! ProtoStart()) return;

  if (mode==IdZSend)
  {
    FileVar->OpId = OpZSend;
    if (strlen(&(FileVar->FullName[FileVar->DirLen]))==0)
    {
      Opt = ts.XmodemBin;
      if (! (*GetMultiFname)(FileVar->HMainWin,FileVar,ts.FileDir,GMF_Z,&Opt) ||
          (FileVar->NumFname==0))
      {
        ProtoEnd();
        return;
      }
      ts.XmodemBin = Opt;
    }
    else
      (*SetFileVar)(FileVar);
  }
  else /* IdZReceive or IdZAuto */
    FileVar->OpId = OpZRcv;

  TalkStatus = IdTalkQuiet;

  /* disable transmit delay (serial port) */
  cv.DelayFlag = FALSE;

  if (! OpenProtoDlg(FileVar,PROTO_ZM,mode,Opt,0))
    ProtoEnd();
}
}

extern "C" {
void BPStart(int mode)
{
  LONG Option;

  if (! ProtoStart()) return;
  if (mode==IdBPSend)
  {
    FileVar->OpId = OpBPSend;
    if (strlen(&(FileVar->FullName[FileVar->DirLen]))==0)
    {
      if (! (*GetTransFname)(FileVar->HMainWin, GTF_BP, &Option, FileVar, ts.FileDir))
      {
        ProtoEnd();
        return;
      }
    }
    else
      (*SetFileVar)(FileVar);

  }
  else /* IdBPReceive or IdBPAuto */
    FileVar->OpId = OpBPRcv;

  TalkStatus = IdTalkQuiet;

  /* disable transmit delay (serial port) */
  cv.DelayFlag = FALSE;

  if (! OpenProtoDlg(FileVar,PROTO_BP,mode,0,0))
    ProtoEnd();
}
}

extern "C" {
void QVStart(int mode)
{
  WORD W;

  if (! ProtoStart()) return;

  if (mode==IdQVSend)
  {
    FileVar->OpId = OpQVSend;
    if (strlen(&(FileVar->FullName[FileVar->DirLen]))==0)
    {
      if (! (*GetMultiFname)(FileVar->HMainWin,FileVar,ts.FileDir,GMF_QV, &W) ||
          (FileVar->NumFname==0))
      {
        ProtoEnd();
        return;
      }
    }
    else
      (*SetFileVar)(FileVar);
  }
  else
    FileVar->OpId = OpQVRcv;

  TalkStatus = IdTalkQuiet;

  /* disable transmit delay (serial port) */
  cv.DelayFlag = FALSE;

  if (! OpenProtoDlg(FileVar,PROTO_QV,mode,0,0))
    ProtoEnd();
}
}
