/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TTCMN.DLL, main */
#include "teraterm.h"
#include <direct.h>
#include "tttypes.h"
#include "ttlib.h"
#include "ttsetup.h"
#include "kanji.h"

/* first instance flag */
static BOOL FirstInstance;

/* shared memory */
typedef struct {
  /* Setup information from "teraterm.ini" */
  TTTSet ts;
  /* Key code map from "keyboard.def" */
  TKeyMap km;
  /* COM port use flag - bit0-3 : COM1-4 */
  BYTE ComFlag;
} TMap;  
typedef TMap *PMap;

static PMap  pm;
static HANDLE HMap = NULL;

void PASCAL FAR StartTeraTerm(PTTSet ts)
{
  char Temp[MAXPATHLEN];

  if (FirstInstance)
  {
    FirstInstance = FALSE;
    /* Get home directory */
    GetModuleFileName(NULL,Temp,sizeof(Temp));
    ExtractDirNameJ(Temp, pm->ts.HomeDir);
    chdir(pm->ts.HomeDir);
    strcpy(pm->ts.SetupFName, pm->ts.HomeDir);
    strcat(pm->ts.SetupFName, "\\TERATERM.INI");
    strcpy(Temp, pm->ts.HomeDir);
    strcat(Temp,"\\KEYBOARD.CNF");
    if (LoadTTSET())
    {
      /* read setup info from "teraterm.ini" */
      (*ReadIniFile)(pm->ts.SetupFName, &(pm->ts));
      /* read keycode map from "keyboard.cnf" */
      (*ReadKeyboardCnf)(Temp,&(pm->km));
      FreeTTSET();
    }
  }
  else
  {
    /* only the first instance uses saved position */
    pm->ts.VTPos.x = CW_USEDEFAULT;
    pm->ts.VTPos.y = CW_USEDEFAULT;
    pm->ts.TEKPos.x = CW_USEDEFAULT;
    pm->ts.TEKPos.y = CW_USEDEFAULT;
  }

  memcpy(ts,&(pm->ts),sizeof(TTTSet));
}

void PASCAL FAR ChangeDefaultSet(PTTSet ts)
{
  if (stricmp(ts->SetupFName, pm->ts.SetupFName) != 0) return;

  memcpy(&(pm->ts),ts,sizeof(TTTSet));
}

void PASCAL FAR LoadDefaultSet(PCHAR SetupFName)
{
  if (stricmp(SetupFName,pm->ts.SetupFName) != 0) return;

  if (LoadTTSET())
  {
    /* read setup info from "teraterm.ini" */
    (*ReadIniFile)(pm->ts.SetupFName, &(pm->ts));
    FreeTTSET();
  }
}

/* Key scan code -> Tera Term key code */
WORD PASCAL FAR GetKeyCode(PKeyMap KeyMap, WORD Scan)
{
  WORD Key;

  if (KeyMap==NULL)
    KeyMap = &(pm->km);
  Key = IdKeyMax;
  while ((Key>0) && (KeyMap->Map[Key-1] != Scan))
    Key--;
  return Key;
}

void PASCAL FAR GetKeyStr(HWND HWin, PKeyMap KeyMap, WORD KeyCode,
                        BOOL AppliKeyMode, BOOL AppliCursorMode,
                        PCHAR KeyStr, LPINT Len, LPBOOL Conv)
{
  MSG Msg;

  if (KeyMap==NULL)
    KeyMap = &(pm->km);

  *Conv = TRUE;
  *Len = 0;
  switch (KeyCode) {
    case IdUp:
      *Len = 3;
      if (AppliCursorMode)
        strcpy(KeyStr,"\033OA");
      else
        strcpy(KeyStr,"\033[A");
      break;
    case IdDown:
      *Len = 3;
      if (AppliCursorMode) strcpy(KeyStr,"\033OB");
                      else strcpy(KeyStr,"\033[B");
      break;
    case IdRight:
      *Len = 3;
      if (AppliCursorMode) strcpy(KeyStr,"\033OC");
                      else strcpy(KeyStr,"\033[C");
      break;
    case IdLeft:
      *Len = 3;
      if (AppliCursorMode) strcpy(KeyStr,"\033OD");
                      else strcpy(KeyStr,"\033[D");
      break;
    case Id0:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Op");
      }
      else {
        *Len = 1;
        KeyStr[0] = '0';
      }
      break;
    case Id1:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Oq");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '1';
      }
      break;
    case Id2:       
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Or");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '2';
      }
      break;
    case Id3:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Os");
      }
      else {
        *Len = 1;
        KeyStr[0] = '3';
      }
      break;
    case Id4:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Ot");
      }
      else {
        *Len = 1;
        KeyStr[0] = '4';
      }
      break;
    case Id5:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Ou");
      }
      else {
        *Len = 1;
        KeyStr[0] = '5';
      }
      break;
    case Id6:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Ov");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '6';
      }
      break;
    case Id7:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Ow");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '7';
      }
      break;
    case Id8:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Ox");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '8';
      }
      break;
    case Id9:
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Oy");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '9';
      }
      break;
    case IdMinus: /* numaric pad - key (DEC) */
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Om");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '-';
      }
      break;
    case IdComma: /* numaric pad , key (DEC) */
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033Ol");
      }  
      else {
        *Len = 1;
        KeyStr[0] = ',';
      }
      break;
    case IdPeriod: /* numaric pad . key */
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033On");
      }  
      else {
        *Len = 1;
        KeyStr[0] = '.';
      }
      break;
    case IdEnter: /* numaric pad enter key */
      if (AppliKeyMode)
      {
        *Len = 3;
        strcpy(KeyStr,"\033OM");
      }  
      else {
        *Len = 1;
        KeyStr[0] = 0x0D;
      }
      break;
    case IdPF1: /* DEC Key: PF1 */
        *Len = 3;
        strcpy(KeyStr,"\033OP");
      break;
    case IdPF2: /* DEC Key: PF2 */
        *Len = 3;
        strcpy(KeyStr,"\033OQ");
      break;
    case IdPF3: /* DEC Key: PF3 */
        *Len = 3;
        strcpy(KeyStr,"\033OR");
      break;
    case IdPF4: /* DEC Key: PF4 */
        *Len = 3;
        strcpy(KeyStr,"\033OS");
      break;
    case IdFind: /* DEC Key: Find */
        *Len = 4;
        strcpy(KeyStr,"\033[1~");
      break;
    case IdInsert: /* DEC Key: Insert Here */
        *Len = 4;
        strcpy(KeyStr,"\033[2~");
      break;
    case IdRemove: /* DEC Key: Remove */
        *Len = 4;
        strcpy(KeyStr,"\033[3~");
      break;
    case IdSelect: /* DEC Key: Select */
        *Len = 4;
        strcpy(KeyStr,"\033[4~");
      break;
    case IdPrev: /* DEC Key: Prev */
        *Len = 4;
        strcpy(KeyStr,"\033[5~");
      break;
    case IdNext: /* DEC Key: Next */
        *Len = 4;
        strcpy(KeyStr,"\033[6~");
      break;
    case IdF6: /* DEC Key: F6 */
        *Len = 5;
        strcpy(KeyStr,"\033[17~");
      break;
    case IdF7: /* DEC Key: F7 */
        *Len = 5;
        strcpy(KeyStr,"\033[18~");
      break;
    case IdF8: /* DEC Key: F8 */
        *Len = 5;
        strcpy(KeyStr,"\033[19~");
      break;
    case IdF9: /* DEC Key: F9 */
        *Len = 5;
        strcpy(KeyStr,"\033[20~");
      break;
    case IdF10: /* DEC Key: F10 */
        *Len = 5;
        strcpy(KeyStr,"\033[21~");
      break;
    case IdF11: /* DEC Key: F11 */
        *Len = 5;
        strcpy(KeyStr,"\033[23~");
      break;
    case IdF12: /* DEC Key: F12 */
        *Len = 5;
        strcpy(KeyStr,"\033[24~");
      break;
    case IdF13: /* DEC Key: F13 */
        *Len = 5;
        strcpy(KeyStr,"\033[25~");
      break;
    case IdF14: /* DEC Key: F14 */
        *Len = 5;
        strcpy(KeyStr,"\033[26~");
      break;
    case IdHelp: /* DEC Key: Help */
        *Len = 5;
        strcpy(KeyStr,"\033[28~");
      break;
    case IdDo: /* DEC Key: Do */
        *Len = 5;
        strcpy(KeyStr,"\033[29~");
      break;
    case IdF17: /* DEC Key: F17 */
        *Len = 5;
        strcpy(KeyStr,"\033[31~");
      break;
    case IdF18: /* DEC Key: F18 */
        *Len = 5;
        strcpy(KeyStr,"\033[32~");
      break;
    case IdF19: /* DEC Key: F19 */
        *Len = 5;
        strcpy(KeyStr,"\033[33~");
      break;
    case IdF20: /* DEC Key: F20 */
        *Len = 5;
        strcpy(KeyStr,"\033[34~");
      break;
    case IdAccEditCopy:
    case IdAccEditPaste:
    case IdAccEditPasteCR:
    case IdAccLineUp:
    case IdAccLineDown:
    case IdAccPageUp:
    case IdAccPageDown:
    case IdAccBuffTop:
    case IdAccBuffBottom:
    case IdAccSendBreak:
      PostMessage(HWin,WM_USER+MsgAccelCommand,KeyCode,0);
      break;
  default:
    if ((KeyCode >= IdUser1) && (KeyCode <= IdKeyMax))
    {
      if ((*KeyMap).UserKeyConv[KeyCode-IdUser1]==0)
        *Conv = FALSE;
      else
        *Conv = TRUE;
      *Len = KeyMap->UserKeyLen[KeyCode-IdUser1];
      memcpy(&KeyStr[0],
                     &KeyMap->UserKeyStr[KeyMap->UserKeyPtr[KeyCode-IdUser1]],
                     *Len);
    }
    else
      return;
  }
  /* remove wm_char message for used keycode */
  PeekMessage(&Msg,HWin, WM_CHAR, WM_CHAR,PM_REMOVE);
}

void FAR PASCAL SetCOMFlag(BYTE Com)
{
  pm->ComFlag = Com & 0x0F;
}

BYTE FAR PASCAL GetCOMFlag()
{
  return (pm->ComFlag & 0x0F);
}

int FAR PASCAL CommReadRawByte(PComVar cv, PBYTE b)
{
  if ( ! cv->Ready ) return 0;

  if ( cv->InBuffCount>0 )
  {
    *b = cv->InBuff[cv->InPtr];
    cv->InPtr++;
    cv->InBuffCount--;
    if ( cv->InBuffCount==0 ) cv->InPtr = 0;
    return 1;
  }
  else {
    cv->InPtr = 0;
    return 0;
  }
}

void PASCAL FAR CommInsert1Byte(PComVar cv, BYTE b)
{
  if ( ! cv->Ready ) return;

  if (cv->InPtr == 0)
    memmove(&(cv->InBuff[1]),&(cv->InBuff[0]),cv->InBuffCount);
  else
    cv->InPtr--;
  cv->InBuff[cv->InPtr] = b;
  cv->InBuffCount++;

  if (cv->HBinBuf!=0 ) cv->BinSkip++;
}

void Log1Bin(PComVar cv, BYTE b)
{
  if ( cv->FilePause ) return;
  if ( cv->BinSkip > 0 )
  {
    cv->BinSkip--;
    return;
  }
  cv->BinBuf[cv->BinPtr] = b;
  cv->BinPtr++;
  if ( cv->BinPtr>=InBuffSize )
    cv->BinPtr = cv->BinPtr-InBuffSize;
  if ( cv->BCount>=InBuffSize )
  {
    cv->BCount = InBuffSize;
    cv->BStart = cv->BinPtr;
  }
  else cv->BCount++;
}

int FAR PASCAL CommRead1Byte(PComVar cv, PBYTE b)
{
  int c;

  if ( ! cv->Ready ) return 0;

  if ( cv->TelMode )
    c = 0;
  else
    c = CommReadRawByte(cv,b);

  if ( c==1 )
  {
    if ( cv->IACFlag )
    {
      cv->IACFlag = FALSE;
      if ( *b != 0xFF )
      {
        cv->TelMode = TRUE;
        CommInsert1Byte(cv,*b);
        if ( cv->HBinBuf!=0 ) cv->BinSkip--;
        c = 0;
      }
    }
    else if ( cv->TelCRFlag )
    {
      cv->TelCRFlag = FALSE;
      if ( *b==0 ) c = 0;
    }
    else if ( cv->TelFlag && (*b==0xFF) )
    {
      cv->IACFlag = TRUE;
      c = 0;
    }
    else if ( cv->TelFlag && ! cv->TelBinRecv && (*b==0x0D) )
      cv->TelCRFlag = TRUE;      
  }

  if ( (c==1) && (cv->HBinBuf!=0) )
    Log1Bin(cv, *b);

  return c;
}

int FAR PASCAL CommRawOut(PComVar cv, PCHAR B, int C)
{
  int a;

  if ( ! cv->Ready ) return C;

  if ( C > OutBuffSize - cv->OutBuffCount)
    a = OutBuffSize - cv->OutBuffCount;
  else
    a = C;
  if ( cv->OutPtr > 0 )
  {
    memmove(&(cv->OutBuff[0]),&(cv->OutBuff[cv->OutPtr]),cv->OutBuffCount);
    cv->OutPtr = 0;
  }
  memcpy(&(cv->OutBuff[cv->OutBuffCount]),B,a);
  cv->OutBuffCount = cv->OutBuffCount + a;
  return a;
}

int FAR PASCAL CommBinaryOut(PComVar cv, PCHAR B, int C)
{
  int a, i, Len;
  char d[3];

  if ( ! cv->Ready ) return C;

  i = 0;
  a = 1;
  while ((a>0) && (i<C))
  {
    Len = 0;

    d[Len] = B[i];
    Len++;

    if ( cv->TelFlag && (B[i]==0x0D) &&
       ! cv->TelBinSend )
    {
      d[Len] = 0x00;
      Len++;
    };

    if ( cv->TelFlag && (B[i]==0xff) )
    {
      d[Len] = 0xff;
      Len++;
    }

    if ( OutBuffSize-cv->OutBuffCount-Len >=0 )
    {
      CommRawOut(cv,d,Len);
      a = 1;
    }
    else
      a = 0;

    i = i + a;
  }
  return i;
}

int FAR PASCAL CommTextOut(PComVar cv, PCHAR B, int C)
{
  int i, TempLen;
  WORD K;
  char TempStr[11];
  int SendCodeNew;
  BOOL Full, KanjiFlagNew;

  if (! cv->Ready ) return C;

  Full = FALSE;
  i = 0;
  while (! Full && (i < C))
  {
    TempLen = 0;
    SendCodeNew = cv->SendCode;

    if ( cv->SendKanjiFlag )
    {
      KanjiFlagNew = FALSE;
      SendCodeNew = IdKanji;

      K = (cv->SendKanjiFirst << 8) + (BYTE)B[i];
      if ( cv->KanjiCodeSend == IdEUC ) K = SJIS2EUC(K);
      else if ( cv->KanjiCodeSend != IdSJIS ) K = SJIS2JIS(K);

      TempStr[0] = HIBYTE(K);
      TempStr[1] = LOBYTE(K);
      TempLen = 2;
    }
    else

      if ( (cv->Language==IdJapanese) && IsDBCSLeadByte(B[i])
      ) {
        KanjiFlagNew = TRUE;
        cv->SendKanjiFirst = (BYTE)B[i];
        SendCodeNew = IdKanji;

        if ((cv->SendCode!=IdKanji) &&
            (cv->KanjiCodeSend==IdJIS))
        {
          TempStr[0] = 0x1B;
          TempStr[1] = '$';
          if ( cv->KanjiIn == IdKanjiInB )
            TempStr[2] = 'B';
          else
            TempStr[2] = '@';
          TempLen = 3;
        }
        else TempLen = 0;

      }
      else {
        KanjiFlagNew = FALSE;

        if ((cv->SendCode==IdKanji) &&
            (cv->KanjiCodeSend==IdJIS))
        {
          TempStr[0] = 0x1B;
          TempStr[1] = '(';
          switch (cv->KanjiOut) {
            case IdKanjiOutJ: TempStr[2] = 'J'; break;
            case IdKanjiOutH: TempStr[2] = 'H'; break;
            default:
              TempStr[2] = 'B';
          }
          TempLen = 3;
        }
        else TempLen = 0;

        if ((cv->Language==IdJapanese) &&
            (0xA0<B[i]) && (B[i]<0xE0) )
        {
          SendCodeNew = IdKatakana;
          if ((cv->SendCode!=IdKatakana) &&
              (cv->KanjiCodeSend==IdJIS) &&
              (cv->JIS7KatakanaSend==1) )
          {
            TempStr[TempLen] = SO;
            TempLen++;
          }
        }
        else {
          SendCodeNew = IdASCII;
          if ((cv->SendCode==IdKatakana) &&
              (cv->KanjiCodeSend==IdJIS) &&
              (cv->JIS7KatakanaSend==1) )
          {
            TempStr[TempLen] = SI;
            TempLen++;
          }
        }

        if (B[i] == 0x0d)
	{
          TempStr[TempLen] = 0x0d;
          TempLen++;
          if ( cv->CRSend==IdCRLF )
          {
            TempStr[TempLen] = 0x0a;
            TempLen++;
          }
          else if ( cv->TelFlag && ! cv->TelBinSend )
          {
            TempStr[TempLen] = 0;
            TempLen++;
          }
        }
        else if ((B[i]>=0xa1) &&
		 (B[i]<=0xe0))
	{
	  if (cv->Language==IdJapanese)
          { /* Katakana */
            if ( cv->KanjiCodeSend==IdEUC )
            {
              TempStr[TempLen] = 0x8E;
              TempLen++;
            }
            if ((cv->KanjiCodeSend==IdJIS) &&
                (cv->JIS7KatakanaSend==1) )
              TempStr[TempLen] = B[i] && 0x7f;
            else
              TempStr[TempLen] = B[i];
            TempLen++;
          }
          else {
            TempStr[TempLen] = B[i];
            TempLen++;
          }
	}
        else {
          TempStr[TempLen] = B[i];
          TempLen++;
          if ( cv->TelFlag && (B[i]==0xFF) )
          {
            TempStr[TempLen] = 0xFF;
            TempLen++;
          }
        }

      } /* if ISDBCSLeadByte..then ... else ... end */

    if ( TempLen == 0 )
    {
      i++;
      cv->SendCode = SendCodeNew;
      cv->SendKanjiFlag = KanjiFlagNew;
    }
    else {
      Full = OutBuffSize-cv->OutBuffCount-TempLen < 0;
      if ( ! Full )
      {
        i++;
        cv->SendCode = SendCodeNew;
        cv->SendKanjiFlag = KanjiFlagNew;
        CommRawOut(cv,TempStr,TempLen);
      }
    }

  }

  return i;
}

int FAR PASCAL CommBinaryEcho(PComVar cv, PCHAR B, int C)
{
  int a, i, Len;
  char d[3];

  if ( ! cv->Ready ) return C;

  if ( (cv->InPtr>0) && (cv->InBuffCount>0) )
  {
    memmove(cv->InBuff,&(cv->InBuff[cv->InPtr]),cv->InBuffCount);
    cv->InPtr = 0;
  }

  i = 0;
  a = 1;
  while ((a>0) && (i<C))
  {
    Len = 0;

    d[Len] = B[i];
    Len++;

    if ( cv->TelFlag && (B[i]==0x0D) &&
        ! cv->TelBinSend )
    {
      d[Len] = 0x00;
      Len++;
    }

    if ( cv->TelFlag && (B[i]==0xff) )
    {
      d[Len] = 0xff;
      Len++;
    }

    if ( InBuffSize-cv->InBuffCount-Len >=0 )
    {
      memcpy(&(cv->InBuff[cv->InBuffCount]),d,Len);
      cv->InBuffCount = cv->InBuffCount + Len;
      a = 1;
    }
    else
      a = 0;
    i = i + a;
  }
  return i;
}

int FAR PASCAL CommTextEcho(PComVar cv, PCHAR B, int C)
{
  int i, TempLen;
  WORD K;
  char TempStr[11];
  int EchoCodeNew;
  BOOL Full, KanjiFlagNew;

  if ( ! cv->Ready ) return C;

  if ( (cv->InPtr>0) && (cv->InBuffCount>0) )
  {
    memmove(cv->InBuff,&(cv->InBuff[cv->InPtr]),cv->InBuffCount);
    cv->InPtr = 0;
  }

  Full = FALSE;
  i = 0;
  while (! Full && (i < C))
  {
    TempLen = 0;
    EchoCodeNew = cv->EchoCode;

    if ( cv->EchoKanjiFlag )
    {
      KanjiFlagNew = FALSE;
      EchoCodeNew = IdKanji;

      K = (cv->EchoKanjiFirst << 8) + B[i];
      if ( cv->KanjiCodeEcho == IdEUC ) K = SJIS2EUC(K);
      else if ( cv->KanjiCodeEcho != IdSJIS ) K = SJIS2JIS(K);

      TempStr[0] = HIBYTE(K);
      TempStr[1] = LOBYTE(K);
      TempLen = 2;
    }
    else

      if ((cv->Language==IdJapanese) && IsDBCSLeadByte(B[i]))
      {
        KanjiFlagNew = TRUE;
        cv->EchoKanjiFirst = B[i];
        EchoCodeNew = IdKanji;

        if ((cv->EchoCode!=IdKanji) &&
            (cv->KanjiCodeEcho==IdJIS) )
        {
          TempStr[0] = 0x1B;
          TempStr[1] = '$';
          if ( cv->KanjiIn == IdKanjiInB )
            TempStr[2] = 'B';
          else
            TempStr[2] = '@';
          TempLen = 3;
        }
        else TempLen = 0;

      }
      else {
        KanjiFlagNew = FALSE;

        if ((cv->EchoCode==IdKanji) &&
            (cv->KanjiCodeEcho==IdJIS) )
        {
          TempStr[0] = 0x1B;
          TempStr[1] = '(';
          switch (cv->KanjiOut) {
            case IdKanjiOutJ: TempStr[2] = 'J'; break;
            case IdKanjiOutH: TempStr[2] = 'H'; break;
            default:
              TempStr[2] = 'B';
          }
          TempLen = 3;
        }
        else TempLen = 0;

        if ((cv->Language==IdJapanese) &&
            (0xA0<B[i]) && (B[i]<0xE0))
        {
          EchoCodeNew = IdKatakana;
          if ((cv->EchoCode!=IdKatakana) &&
              (cv->KanjiCodeEcho==IdJIS) &&
              (cv->JIS7KatakanaEcho==1) )
          {
            TempStr[TempLen] = SO;
            TempLen++;
          }
        }
        else {
          EchoCodeNew = IdASCII;
          if ((cv->EchoCode==IdKatakana) &&
              (cv->KanjiCodeEcho==IdJIS) &&
              (cv->JIS7KatakanaEcho==1))
          {
            TempStr[TempLen] = SI;
            TempLen++;
          }
        }

        if (B[i]==0x0d)
        {
          TempStr[TempLen] = 0x0d;
          TempLen++;
          if ( cv->CRSend==IdCRLF )
          {
            TempStr[TempLen] = 0x0a;
            TempLen++;
          }
          else if ( cv->TelFlag && ! cv->TelBinSend )
          {
            TempStr[TempLen] = 0;
            TempLen++;
          }
	}
	else if ((B[i]>=0xa1) && (B[i]<=0xe0))
	{
          if (cv->Language==IdJapanese)
          { /* Katakana */
            if ( cv->KanjiCodeEcho==IdEUC )
            {
              TempStr[TempLen] = 0x8E;
              TempLen++;
            }
            if ((cv->KanjiCodeEcho==IdJIS) &&
                (cv->JIS7KatakanaEcho==1))
              TempStr[TempLen] = (char)(B[i]) && 0x7f;
            else
              TempStr[TempLen] = B[i];
            TempLen++;
          }
          else {
            TempStr[TempLen] = B[i];
            TempLen++;
          }
	}
        else    
        {
          TempStr[TempLen] = B[i];
          TempLen++;
          if ( cv->TelFlag && (B[i]==0xFF) )
          {
            TempStr[TempLen] = 0xFF;
            TempLen++;
          }
        }
      } /* if ISDBCSLeadByte..then ... else ... end */

    if ( TempLen == 0 )
    {
      i++;
      cv->EchoCode = EchoCodeNew;
      cv->EchoKanjiFlag = KanjiFlagNew;
    }
    else {
      Full = InBuffSize-cv->InBuffCount-TempLen < 0;
      if ( ! Full )
      {
        i++;
        cv->EchoCode = EchoCodeNew;
        cv->EchoKanjiFlag = KanjiFlagNew;
        memcpy(&(cv->InBuff[cv->InBuffCount]),TempStr,TempLen);
        cv->InBuffCount = cv->InBuffCount + TempLen;
      }
    }

  }

  return i;
}

#pragma off (unreferenced);
int APIENTRY LibMain(HANDLE hdll,
                                      DWORD  reason,
                                      LPVOID reserved)
#pragma on (unreferenced);
{
  switch( reason ) {
    case DLL_THREAD_ATTACH:
     /* do thread initialization */
      break;
    case DLL_THREAD_DETACH:
      /* do thread cleanup */
      break;
    case DLL_PROCESS_ATTACH:
      /* do process initialization */
      HMap = CreateFileMapping(
        (HANDLE) 0xFFFFFFFF, NULL, PAGE_READWRITE,
        0, sizeof(TMap), "ttset_memfilemap");
      if (HMap == NULL) return FALSE;
      FirstInstance = (GetLastError() != ERROR_ALREADY_EXISTS);

      pm = (PMap)MapViewOfFile(
        HMap,FILE_MAP_WRITE,0,0,0);
      if (pm == NULL) return FALSE;
      break;
    case DLL_PROCESS_DETACH:
      /* do process cleanup */
      UnmapViewOfFile(pm);
      CloseHandle(HMap);
      break;
  }
  return( 1 );
}
