/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, IME interface */
#include "teraterm.h"
#include "tttypes.h"
#include "ttwinman.h"
#include "ttcommon.h"

#include "ttime.h"

static HANDLE HWINNLS = NULL;
static HANDLE HIMElf = NULL;

typedef LRESULT (WINAPI *TSendIMEMessageEx)
  (HWND, LPARAM);

typedef BOOL (WINAPI *TWINNLSEnableIME)
  (HWND HWin, BOOL bEnable);

TSendIMEMessageEx PSendIMEMessageEx;
TWINNLSEnableIME PWINNLSEnableIME;

BOOL LoadWINNLS()
{
  BOOL Err;

  if ( HWINNLS != NULL )
    return TRUE;

  HWINNLS = LoadLibrary("USER32.DLL");

  if ( HWINNLS == NULL )
  {
    MessageBox(0,"Can't use IME",
               "Tera Term: Error",MB_ICONEXCLAMATION);
    WritePrivateProfileString("Tera Term","IME","off",ts.SetupFName);
    LoadDefaultSet(ts.SetupFName);
    return FALSE;
  }

  Err = FALSE;
  /* Win 95J and Win NT 3.5J */
  PSendIMEMessageEx = (TSendIMEMessageEx)GetProcAddress(HWINNLS, "SendIMEMessageExA");
  if ( PSendIMEMessageEx==NULL ) /* Win NT 3.51J */
    PSendIMEMessageEx = (TSendIMEMessageEx)GetProcAddress(HWINNLS, "ImmSendIMEMessageExA");
  if ( PSendIMEMessageEx==NULL )
    Err = TRUE;

  PWINNLSEnableIME = (TWINNLSEnableIME)GetProcAddress(HWINNLS, "WINNLSEnableIME");
  if ( PWINNLSEnableIME==NULL ) Err = TRUE;

  if ( Err )
  {
    FreeLibrary(HWINNLS);
    HWINNLS = NULL;
    return FALSE;
  }
  else
    return TRUE;

}

void FreeWINNLS()
{
  if (HIMElf!=NULL)
    GlobalFree(HIMElf);
  HIMElf = NULL;

  if ( HWINNLS != NULL )
  {
    /* position of conv. window -> default */
    SetConversionWindow(HVTWin,-1,0);

    FreeLibrary(HWINNLS);
    HWINNLS = NULL;
  }
}

BOOL CanUseIME()
{
  return (HWINNLS != NULL);
}

void EnableIME(BOOL On)
{
  PWINNLSEnableIME(0,On);
}

void SetConversionWindow(HWND HWin, int X, int Y)
{
  HANDLE HIME;
  PIMESTRUCT PIME;

  if ( HWINNLS == NULL ) return;

  HIME = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE,sizeof(IMESTRUCT));

  /* Set position of conversion window */
  PIME = (PIMESTRUCT)GlobalLock(HIME);
  (*PIME).fnc = IME_SETCONVERSIONWINDOW;
  if ( X<0 ) (*PIME).wParam = MCW_DEFAULT;
        else (*PIME).wParam = MCW_WINDOW;
  (*PIME).lParam1 = MAKELONG(X,Y);
  GlobalUnlock(HIME);
  PSendIMEMessageEx(HWin, (LPARAM)HIME);

  /* Set font of conversion window */
  PIME = (PIMESTRUCT)GlobalLock(HIME);
  (*PIME).fnc = IME_SETCONVERSIONFONTEX;
  if ( X<0 ) (*PIME).lParam1 = NULL;
        else (*PIME).lParam1 = (LPARAM)HIMElf;
  GlobalUnlock(HIME);
  PSendIMEMessageEx(HWin, (LPARAM)HIME);

  GlobalFree(HIME);
}

void SetConversionLogFont(PLOGFONT lf)
{
  PLOGFONT PIMElf;

  if ( HWINNLS == NULL ) return;

  if (HIMElf==NULL)
    HIMElf = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE,sizeof(LOGFONT));
  if (HIMElf==NULL) return;

  PIMElf = (PLOGFONT)GlobalLock(HIMElf);
  memcpy(PIMElf,lf,sizeof(LOGFONT));
  GlobalUnlock(HIMElf);
}
