/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TTSET.DLL, setup file routines*/
#include "teraterm.h"
#include <stdio.h>
#include <direct.h>
#include "tttypes.h"
#include "ttlib.h"

#define Section "Tera Term"

static PCHAR TermList[] =
  {"VT100","VT100J","VT101","VT102","VT102J","VT220J","VT282","VT320","VT382",NULL};
static PCHAR BaudList[] =
  {"110","300","600","1200","2400","4800","9600",
   "19200","38400","57600","115200",NULL};

WORD str2id(PCHAR * List, PCHAR str, WORD DefId)
{
  WORD i;
  i = 0;
  while ((List[i] != NULL) && (stricmp(List[i],str) != 0))
    i++;
  if (List[i] == NULL)
    i = DefId;
  else
    i++;

  return i;
}

void id2str(PCHAR * List, WORD Id, WORD DefId, PCHAR str)
{
  int i;

  if (Id==0)
    i = DefId - 1;
  else {
    i = 0;
    while ((List[i] != NULL) && (i < Id-1))
      i++;
    if (List[i]==NULL)
      i = DefId - 1;
  } 
  strcpy(str,List[i]);
}

void GetNthString(PCHAR Source, int Nth, int Size, PCHAR Dest)
{
  int i, j, k;
  char c;

  i = 1;
  j = 0;
  k = 0;
  do {
    c = Source[j];
    if ( c==',' ) i++;
    j++;
    if ( (i==Nth) && (c!=',') && (k<Size-1) )
    {
      Dest[k] = c;
      k++;
    }
  }  
  while ((i<=Nth) && (c!=0));
  Dest[k] = 0;
}

void GetNthNum(PCHAR Source, int Nth, int *Num)
{
  char T[6];

  GetNthString(Source,Nth,sizeof(T),&T[0]);
  if (sscanf(T, "%d", Num) != 1)
    Num = 0;
}

WORD GetOnOff(PCHAR Sect, PCHAR Key, PCHAR FName, BOOL Default)
{
  char Temp[4];
  GetPrivateProfileString(Sect,Key,"",
                          Temp,sizeof(Temp),FName);
  if ( Default )
  {
    if ( stricmp(Temp,"off")==0 )
      return 0;
    else
      return 1;
  }
  else {
    if ( stricmp(Temp,"on")==0 )
      return 1;
    else
      return 0;
  }
}

void WriteOnOff(PCHAR Sect, PCHAR Key, PCHAR FName, WORD Flag)
{
  char Temp[4];

  if ( Flag>0 ) strcpy(Temp,"on");
             else strcpy(Temp,"off");
  WritePrivateProfileString(Sect,Key,Temp,FName);
}

void FAR PASCAL ReadIniFile(PCHAR FName, PTTSet ts)
{
  int i;
  HDC TmpDC;
  char Temp[MAXPATHLEN];
  BOOL AskUser;

  AskUser = FALSE;
  /* Version number */
  GetPrivateProfileString(Section,"Version","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"2.0")!=0 )
    AskUser = TRUE;

  /* Language */
  GetPrivateProfileString(Section,"Language","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"Japanese")==0 )
    ts->Language = IdJapanese;
  else if ( stricmp(Temp,"English")==0 )
    ts->Language = IdEnglish;
  else {
    ts->Language = IdEnglish;
    AskUser = TRUE;
  }

  /* Port type */
  GetPrivateProfileString(Section,"Port","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"tcpip")==0 )
    ts->PortType = IdTCPIP;
  else if ( stricmp(Temp,"serial")==0 )
    ts->PortType = IdSerial;
  else {
    ts->PortType = IdSerial;
    AskUser = TRUE;
  }

  /* VT win position */
  GetPrivateProfileString(Section,"VTPos","-32768,-32768",
                          Temp,sizeof(Temp),FName);  /* default: random position */
  GetNthNum(Temp,1,(int *)(&ts->VTPos.x));
  GetNthNum(Temp,2,(int *)(&ts->VTPos.y));

  if ( (ts->VTPos.x < -20) || (ts->VTPos.y < -20) )
  {
    ts->VTPos.x = CW_USEDEFAULT;
    ts->VTPos.y = CW_USEDEFAULT;
  }
  else {
    if ( ts->VTPos.x < 0 ) ts->VTPos.x = 0;
    if ( ts->VTPos.y < 0 ) ts->VTPos.y = 0;
  }

  /* TEK win position */
  GetPrivateProfileString(Section,"TEKPos","-32768,-32768",
                          Temp,sizeof(Temp),FName);  /* default: random position */
  GetNthNum(Temp,1,(int *)&(ts->TEKPos.x));
  GetNthNum(Temp,2,(int *)&(ts->TEKPos.y));

  if ( (ts->TEKPos.x < -20) || (ts->TEKPos.y < -20) )
  {
    ts->TEKPos.x = CW_USEDEFAULT;
    ts->TEKPos.y = CW_USEDEFAULT;
  }
  else {
    if ( ts->TEKPos.x < 0 ) ts->TEKPos.x = 0;
    if ( ts->TEKPos.y < 0 ) ts->TEKPos.y = 0;
  }

  /* VT terminal size  */
  GetPrivateProfileString(Section,"TerminalSize","80,24",
                          Temp,sizeof(Temp),FName);
  GetNthNum(Temp,1,&ts->TerminalWidth);
  GetNthNum(Temp,2,&ts->TerminalHeight);
  if ( ts->TerminalWidth < 0 ) ts->TerminalWidth = 1;
  if ( ts->TerminalHeight < 0 ) ts->TerminalHeight = 1;

  /* Terminal size = Window size */
  ts->TermIsWin = GetOnOff(Section,"TermIsWin",FName,FALSE);

  /* Auto window resize flag */
  ts->AutoWinResize = GetOnOff(Section,"AutoWinResize",FName,FALSE);

  /* CR Receive */
  GetPrivateProfileString(Section,"CRReceive","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"CRLF")==0 ) ts->CRReceive = IdCRLF;
                             else ts->CRReceive = IdCR;
  /* CR Send */
  GetPrivateProfileString(Section,"CRSend","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"CRLF")==0 ) ts->CRSend = IdCRLF;
                             else ts->CRSend = IdCR;

  /* Local echo */
  ts->LocalEcho = GetOnOff(Section,"LocalEcho",FName,FALSE);

  /* Auto B-Plus activation -- read-only option */
  ts->BPAuto = GetOnOff(Section,"BPAuto",FName,FALSE)>0;

  /* Answerback */
  if ( ! ts->BPAuto )
  {
    GetPrivateProfileString(Section,"Answerback","",Temp,
                            sizeof(Temp),FName);
    ts->AnswerbackLen =
      Hex2Str(Temp,ts->Answerback,sizeof(ts->Answerback));
  }
  else {
    strcpy(ts->Answerback,"\020++\0200");
    ts->AnswerbackLen = 5;
  }

  /* Kanji Code (receive) */
  GetPrivateProfileString(Section,"KanjiReceive","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"EUC")==0 ) ts->KanjiCode = IdEUC;
  else if ( stricmp(Temp,"JIS")==0 ) ts->KanjiCode = IdJIS;
  else ts->KanjiCode = IdSJIS;

  /* Katakana (receive) */
  GetPrivateProfileString(Section,"KatakanaReceive","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"7")==0 ) ts->JIS7Katakana = 1;
  else ts->JIS7Katakana = 0;


  /* Kanji Code (transmit) */
  GetPrivateProfileString(Section,"KanjiSend","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"EUC")==0 ) ts->KanjiCodeSend = IdEUC;
  else if ( stricmp(Temp,"JIS")==0 ) ts->KanjiCodeSend = IdJIS;
  else ts->KanjiCodeSend = IdSJIS;

  /* Katakana (receive) */
  GetPrivateProfileString(Section,"KatakanaSend","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"7")==0 ) ts->JIS7KatakanaSend = 1;
  else ts->JIS7KatakanaSend = 0;

  /* KanjiIn */
  GetPrivateProfileString(Section,"KanjiIn","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"@")==0 ) ts->KanjiIn = IdKanjiInA;
  else ts->KanjiIn = IdKanjiInB;

  /* KanjiOut */
  GetPrivateProfileString(Section,"KanjiOut","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"B")==0 ) ts->KanjiOut = IdKanjiOutB;
  else if ( ts->AllowWrongSequence && (stricmp(Temp,"H")==0) ) ts->KanjiOut = IdKanjiOutH;
  else ts->KanjiOut = IdKanjiOutJ;

  /* Auto Win Switch VT<->TEK */
  ts->AutoWinSwitch = GetOnOff(Section,"AutoWinSwitch",FName,FALSE);

  /* Terminal ID */
  GetPrivateProfileString(Section,"TerminalID","",
                          Temp,sizeof(Temp),FName);
  ts->TerminalID = str2id(TermList,Temp,IdVT100);

  /* Title String */
  GetPrivateProfileString(Section,"Title","Tera Term",
                          ts->Title,sizeof(ts->Title),FName);

  /* Cursor shape */
  GetPrivateProfileString(Section,"CursorShape","",
                          Temp,sizeof(Temp),FName);
       if ( stricmp(Temp,"vertical"  )==0 ) ts->CursorShape = IdVCur;
  else if ( stricmp(Temp,"horizontal")==0 ) ts->CursorShape = IdHCur;
  else ts->CursorShape = IdBlkCur;

  /* Popup menu */
  ts->PopupMenu = GetOnOff(Section,"PopupMenu",FName,FALSE);

  /* Enable scroll buffer */
  ts->EnableScrollBuff =
    GetOnOff(Section,"EnableScrollBuff",FName,TRUE);

  /* Scroll buffer size */
  ts->ScrollBuffMax =
    GetPrivateProfileInt(Section,"ScrollBuffSize",100,FName);

  /* VT Color */
  GetPrivateProfileString(Section,"VTColor","0,0,0,255,255,255",
                          Temp,sizeof(Temp),FName);
  for (i = 0; i<=5; i++)
    GetNthNum(Temp,i+1,(int *)&(ts->TmpColor[0][i]));
  for (i = 0; i<=1; i++)
    ts->VTColor[i] = RGB( (BYTE)ts->TmpColor[0][i*3],
                              (BYTE)ts->TmpColor[0][i*3+1],
                              (BYTE)ts->TmpColor[0][i*3+2]);

  /* VT Bold Color */
  GetPrivateProfileString(Section,"VTBoldColor","0,0,255,255,255,255",
                          Temp,sizeof(Temp),FName);
  for (i = 0; i<=5; i++)
    GetNthNum(Temp,i+1,(int *)&(ts->TmpColor[0][i]));
  for (i = 0; i<=1; i++)
    ts->VTBoldColor[i] = RGB((BYTE)ts->TmpColor[0][i*3],
                              (BYTE)ts->TmpColor[0][i*3+1],
                              (BYTE)ts->TmpColor[0][i*3+2]);

  /* VT Blink Color */
  GetPrivateProfileString(Section,"VTBlinkColor","255,0,0,255,255,255",
                          Temp,sizeof(Temp),FName);
  for (i = 0; i<=5; i++)
    GetNthNum(Temp,i+1,(int *)&(ts->TmpColor[0][i]));
  for (i = 0; i<=1; i++)
    ts->VTBlinkColor[i] = RGB((BYTE)ts->TmpColor[0][i*3],
                              (BYTE)ts->TmpColor[0][i*3+1],
                              (BYTE)ts->TmpColor[0][i*3+2]);

  /* TEK Color */
  GetPrivateProfileString(Section,"TEKColor","0,0,0,255,255,255",
                          Temp,sizeof(Temp),FName);
  for (i = 0; i<=5; i++)
    GetNthNum(Temp,i+1,(int *)&(ts->TmpColor[0][i]));
  for (i = 0; i<=1; i++)
    ts->TEKColor[i] = RGB((BYTE)ts->TmpColor[0][i*3],
                              (BYTE)ts->TmpColor[0][i*3+1],
                              (BYTE)ts->TmpColor[0][i*3+2]);

  TmpDC = GetDC(0); /* Get screen device context */
  for (i = 0; i<=1; i++)
    ts->VTColor[i] = GetNearestColor(TmpDC, ts->VTColor[i]);
  for (i = 0; i<=1; i++)
    ts->VTBoldColor[i] = GetNearestColor(TmpDC, ts->VTBoldColor[i]);
  for (i = 0; i<=1; i++)
    ts->VTBlinkColor[i] = GetNearestColor(TmpDC, ts->VTBlinkColor[i]);
  for (i = 0; i<=1; i++)
    ts->TEKColor[i] = GetNearestColor(TmpDC, ts->TEKColor[i]);
  ReleaseDC(0, TmpDC);

  /* TEK color emulation */
  ts->TEKColorEmu =
    GetOnOff(Section,"TEKColorEmulation",FName,FALSE);

  /* VT Font */
  GetPrivateProfileString(Section,"VTFont","Terminal,0,-13,1",
                          Temp,sizeof(Temp),FName);
  GetNthString(Temp,1,sizeof(ts->VTFont),ts->VTFont);
  GetNthNum(Temp,2,(int *)&(ts->VTFontSize.x));
  GetNthNum(Temp,3,(int *)&(ts->VTFontSize.y));
  GetNthNum(Temp,4,&(ts->VTFontCharSet));
  /* for Windows NT J */
  if ((ts->Language==IdJapanese) &&
      (stricmp(ts->VTFont,"Terminal")==0))
    ts->VTFontCharSet = 128;

  /* Bold font flag */
  ts->EnableBold = GetOnOff(Section,"EnableBold",FName,FALSE);

  /* TEK Font */
  GetPrivateProfileString(Section,"TEKFont","Terminal,0,-8,1",
                          Temp,sizeof(Temp),FName);
  GetNthString(Temp,1,sizeof(ts->TEKFont),ts->TEKFont);
  GetNthNum(Temp,2,(int *)&(ts->TEKFontSize.x));
  GetNthNum(Temp,3,(int *)&(ts->TEKFontSize.y));
  GetNthNum(Temp,4,&(ts->TEKFontCharSet));

  /* BS Key */
  GetPrivateProfileString(Section,"BSKey","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"DEL")==0 ) ts->BSKey = IdDEL;
                            else ts->BSKey = IdBS;

  /* IME Flag */
  ts->UseIME = GetOnOff(Section,"IME",FName,TRUE);

  /* IME Flag */
  ts->IMEInline = GetOnOff(Section,"IMEInline",FName,TRUE);

  /* Meta Key */
  ts->MetaKey = GetOnOff(Section,"MetaKey",FName,FALSE);

  /* Serial port ID */
  ts->ComPort =
    GetPrivateProfileInt(Section,"ComPort",1,FName);
  if ( (ts->ComPort<1) || (ts->ComPort>4) )
    ts->ComPort = 1;

  /* Baud rate */
  GetPrivateProfileString(Section,"BaudRate","9600",
                          Temp,sizeof(Temp),FName);
  ts->Baud = str2id(BaudList,Temp,IdBaud9600);

  /* Parity */
  GetPrivateProfileString(Section,"Parity","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"even")==0 ) ts->Parity = IdParityEven;
  else if ( stricmp(Temp,"odd" )==0 ) ts->Parity = IdParityOdd;
  else ts->Parity = IdParityNone;

  /* Data bit */
  GetPrivateProfileString(Section,"DataBit","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"7")==0 ) ts->DataBit = IdDataBit7;
  else ts->DataBit = IdDataBit8;

  /* Stop bit */
  GetPrivateProfileString(Section,"StopBit","",
                          Temp,sizeof(Temp),FName);
  if ( stricmp(Temp,"2")==0 ) ts->StopBit = IdStopBit2;
  else ts->StopBit = IdStopBit1;

  /* Flow control */
  GetPrivateProfileString(Section,"FlowCtrl","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"x"   )==0 ) ts->Flow = IdFlowX;
  else if ( stricmp(Temp,"hard")==0 ) ts->Flow = IdFlowHard;
  else ts->Flow = IdFlowNone;

  /* Delay per character */
  ts->DelayPerChar =
    GetPrivateProfileInt(Section,"DelayPerChar",0,FName);

  /* Delay per line */
  ts->DelayPerLine =
    GetPrivateProfileInt(Section,"DelayPerLine",0,FName);

  /* TCP port num */
  ts->TCPPort =
    GetPrivateProfileInt(Section,"TCPPort",23,FName);

  /* Telnet flag */
  ts->Telnet = GetOnOff(Section,"Telnet",FName,TRUE);

  /* Telnet terminal type */
  GetPrivateProfileString(Section,"TermType","vt100",ts->TermType,
                          sizeof(ts->TermType),FName);

  /* Auto window close flag */
  ts->AutoWinClose = GetOnOff(Section,"AutoWinClose",FName,TRUE);

  /* File transfer binary flag */
  ts->TransBin = GetOnOff(Section,"TransBinary",FName,FALSE);

  /* XMODEM option */
  GetPrivateProfileString(Section,"XmodemOpt","",
                          Temp,sizeof(Temp),FName);
  if (      stricmp(Temp,"crc"   )==0 ) ts->XmodemOpt = XoptCRC;
  else if ( stricmp(Temp,"1k")==0 ) ts->XmodemOpt = Xopt1K;
  else ts->XmodemOpt = XoptCheck;

  /* XMODEM binary file */
  ts->XmodemBin = GetOnOff(Section,"XmodemBinary",FName,TRUE);

  /* Default directory for file transfer */
  GetPrivateProfileString(Section,"FileDir","",
                          ts->FileDir,sizeof(ts->FileDir),FName);
  if ( strlen(ts->FileDir)==0 )
    strcpy(ts->FileDir,ts->HomeDir);
  else {
    getcwd(Temp,sizeof(Temp));
    if  (chdir(ts->FileDir) != 0)
      strcpy(ts->FileDir,ts->HomeDir);
    chdir(Temp);
  }

  /* Debug flag  -- read-only option */
  ts->Debug = GetOnOff(Section,"Debug",FName,FALSE)>0;

  /* Kermit log  -- read-only option */
  ts->KmtLog = GetOnOff(Section,"KmtLog",FName,FALSE)>0;

  /* XMODEM log  -- read-only option */
  ts->XmodemLog = GetOnOff(Section,"XmodemLog",FName,FALSE)>0;

  /* ZMODEM log  -- read-only option */
  ts->ZmodemLog = GetOnOff(Section,"ZmodemLog",FName,FALSE)>0;

  /* ZMODEM ESCCTL flag  -- read-only option */
  ts->ZmodemEscCtl = GetOnOff(Section,"ZmodemEscCtl",FName,FALSE)>0;

  /* ZMODEM data subpacket length for sending -- read-only */
  ts->ZmodemDataLen =
    GetPrivateProfileInt(Section,"ZmodemDataLen",1024,FName);
  /* ZMODEM window size for sending -- read-only */
  ts->ZmodemWinSize =
    GetPrivateProfileInt(Section,"ZmodemWinSize",0,FName);

  /* Auto ZMODEM activation -- read-only option */
  ts->ZmodemAuto = GetOnOff(Section,"ZmodemAuto",FName,FALSE)>0;

  /* B-Plus log  -- read-only option */
  ts->BPLog = GetOnOff(Section,"BPLog",FName,FALSE)>0;

  /* Quick-VAN log  -- read-only option */
  ts->QVLog = GetOnOff(Section,"QVLog",FName,FALSE)>0;

  /* Quick-VAN window size -- read-only */
  ts->QVWinSize =
    GetPrivateProfileInt(Section,"QVWinSize",8,FName);

  /* Telnet log  -- read-only option */
  ts->TelLog = GetOnOff(Section,"TelLog",FName,FALSE)>0;

  /* Ctrl code in Kanji -- read-only option */
  ts->CtrlInKanji = GetOnOff(Section,"CtrlInKanji",FName,TRUE)>0;

  /* Scroll threshold -- read-only option */
  ts->ScrollThreshold =
    GetPrivateProfileInt(Section,"ScrollThreshold",12,FName);

  /* Wrong sequence flag  -- read-only option */
  ts->AllowWrongSequence =
    GetOnOff(Section,"AllowWrongSequence",FName,FALSE)>0;

  /* 8 bit control code flag  -- read-only option */
  ts->Accept8bitCtrl =
    GetOnOff(Section,"Accept8bitCtrl",FName,TRUE)>0;

  /* TEK GIN Mouse keycode -- read-only option */
  ts->GINMouseCode =
    GetPrivateProfileInt(Section,"TEKGINMouseCode",32,FName);

  /* Confirm disconnect -- read-only option */
  ts->ConfirmDisconn =
    GetOnOff(Section,"ConfirmDisconnect",FName,TRUE);

  /* VT Font space --- read-only option */
  GetPrivateProfileString(Section,"VTFontSpace","0,0,0,0",
                          Temp,sizeof(Temp),FName);
  GetNthNum(Temp,1,&ts->FontDX);
  GetNthNum(Temp,2,&ts->FontDW);
  GetNthNum(Temp,3,&ts->FontDY);
  GetNthNum(Temp,4,&ts->FontDH);
  if ( ts->FontDX<0 )
    ts->FontDX = 0;
  if ( ts->FontDW<0 )
    ts->FontDW = 0;
  ts->FontDW = ts->FontDW +
                   ts->FontDX;
  if ( ts->FontDY<0 )
    ts->FontDY = 0;
  if ( ts->FontDH<0 )
    ts->FontDH = 0;
  ts->FontDH = ts->FontDH +
                   ts->FontDY;

  /* Printer Font --- read-only option */
  GetPrivateProfileString(Section,"PrnFont","",
                          Temp,sizeof(Temp),FName);
  if ( strlen(Temp)==0 )
  {
    ts->PrnFont[0] = 0;
    ts->PrnFontSize.x = 0;
    ts->PrnFontSize.y = 0;
    ts->PrnFontCharSet = 0;
  }
  else {
    GetNthString(Temp,1,sizeof(ts->PrnFont),ts->PrnFont);
    GetNthNum(Temp,2,(int *)&(ts->PrnFontSize.x));
    GetNthNum(Temp,3,(int *)&(ts->PrnFontSize.y));
    GetNthNum(Temp,4,&(ts->PrnFontCharSet));
  }

  /* Telnet binary flag -- read-only option */
  ts->TelBin = GetOnOff(Section,"TelBin",FName,FALSE);

  /* Hide title -- read-only option */
  ts->HideTitle = GetOnOff(Section,"HideTitle",FName,FALSE);

  ts->Ok = ! AskUser;

}

void FAR PASCAL WriteIniFile(PCHAR FName, PTTSet ts)
{
  int i;
  char NumStr[11];
  char Temp[81];

  /* version */
  WritePrivateProfileString(Section,"Version","2.0",FName);

  /* Language */
  if ( ts->Language==IdJapanese )
    strcpy(Temp,"Japanese");
  else
    strcpy(Temp,"English");
  WritePrivateProfileString(Section,"Language",Temp,FName);

  /* Port type */
  if (ts->PortType == IdSerial)
   strcpy(Temp,"serial");
  else /* IdFile -> IdTCPIP */
    strcpy(Temp,"tcpip");

  WritePrivateProfileString(Section,"Port",Temp,FName);

  /* VT win position */
  sprintf(Temp,"%d,%d",ts->VTPos.x,ts->VTPos.y);
  WritePrivateProfileString(Section,"VTPos",Temp,FName);

  /* TEK win position */
  sprintf(Temp,"%d,%d",ts->TEKPos.x,ts->TEKPos.y);
  WritePrivateProfileString(Section,"TEKPos",Temp,FName);

  /* VT terminal size  */
  sprintf(Temp,"%d,%d",ts->TerminalWidth,ts->TerminalHeight);
  WritePrivateProfileString(Section,"TerminalSize",Temp,FName);

  /* Terminal size = Window size */
  WriteOnOff(Section,"TermIsWin",FName,ts->TermIsWin);

  /* Auto window resize flag */
  WriteOnOff(Section,"AutoWinResize",FName,ts->AutoWinResize);

  /* CR Receive */
  if ( ts->CRReceive == IdCRLF ) strcpy(Temp,"CRLF");
                                 else strcpy(Temp,"CR");
  WritePrivateProfileString(Section,"CRReceive",Temp,FName);

  /* CR Send */
  if ( ts->CRSend == IdCRLF ) strcpy(Temp,"CRLF");
                              else strcpy(Temp,"CR");
  WritePrivateProfileString(Section,"CRSend",Temp,FName);

  /* Local echo */
  WriteOnOff(Section,"LocalEcho",FName,ts->LocalEcho);

  /* Answerback */
  if ( ! ts->BPAuto )
  {
    Str2Hex(ts->Answerback,Temp,ts->AnswerbackLen,
            sizeof(Temp)-1,TRUE);
    WritePrivateProfileString(Section,"Answerback",Temp,FName);
  }

  /* Kanji Code (receive)  */
  switch (ts->KanjiCode) {
    case IdEUC: strcpy(Temp,"EUC"); break;
    case IdJIS: strcpy(Temp,"JIS"); break;
    default:
      strcpy(Temp,"SJIS");
  }
  WritePrivateProfileString(Section,"KanjiReceive",Temp,FName);

  /* Katakana (receive)  */
  if (ts->JIS7Katakana ==1)
    strcpy(Temp,"7");
  else
    strcpy(Temp,"8");

  WritePrivateProfileString(Section,"KatakanaReceive",Temp,FName);

  /* Kanji Code (transmit)  */
  switch (ts->KanjiCodeSend) {
    case IdEUC: strcpy(Temp,"EUC"); break;
    case IdJIS: strcpy(Temp,"JIS"); break;
    default:
      strcpy(Temp,"SJIS");
  }
  WritePrivateProfileString(Section,"KanjiSend",Temp,FName);

  /* Katakana (transmit)  */
  if (ts->JIS7KatakanaSend==1)
    strcpy(Temp,"7");
  else
    strcpy(Temp,"8");

  WritePrivateProfileString(Section,"KatakanaSend",Temp,FName);

  /* KanjiIn */
  if (ts->KanjiIn == IdKanjiInA)
    strcpy(Temp,"@");
  else
    strcpy(Temp,"B");

  WritePrivateProfileString(Section,"KanjiIn",Temp,FName);

  /* KanjiOut */
  switch (ts->KanjiOut) {
    case IdKanjiOutB: strcpy(Temp,"B"); break;
    case IdKanjiOutH: strcpy(Temp,"H"); break;
    default:
      strcpy(Temp,"J");
  }
  WritePrivateProfileString(Section,"KanjiOut",Temp,FName);

  /* AutoWinChange VT<->TEK */
  WriteOnOff(Section,"AutoWinSwitch",FName,ts->AutoWinSwitch);

  /* Terminal ID */
  id2str(TermList,ts->TerminalID,IdVT100,Temp);
  WritePrivateProfileString(Section,"TerminalID",Temp,FName);

  /* Title text */
  WritePrivateProfileString(Section,"Title",ts->Title,FName);

  /* Cursor shape */
  switch (ts->CursorShape) {
    case IdVCur  : strcpy(Temp,"vertical"); break;
    case IdHCur  : strcpy(Temp,"horizontal"); break;
    default:
      strcpy(Temp,"block");
  }
  WritePrivateProfileString(Section,"CursorShape",Temp,FName);

  /* Popup menu */
  WriteOnOff(Section,"PopupMenu",FName,ts->PopupMenu);

  /* Enable scroll buffer */
  WriteOnOff(Section,"EnableScrollBuff",FName,ts->EnableScrollBuff);

  /* Scroll buffer size */
  sprintf(Temp,"%d",ts->ScrollBuffMax);
  WritePrivateProfileString(Section,"ScrollBuffSize",Temp,FName);

  /* VT Color */
  for (i = 0;i<=1;i++)
  {
    ts->TmpColor[0][i*3]   = GetRValue(ts->VTColor[i]);
    ts->TmpColor[0][i*3+1] = GetGValue(ts->VTColor[i]);
    ts->TmpColor[0][i*3+2] = GetBValue(ts->VTColor[i]);
  }
  Temp[0] = 0;
  for (i = 0;i<=5;i++)
  {
    sprintf(NumStr,"%d",ts->TmpColor[0][i]);
    strcat(Temp,NumStr);
    if ( i!=5 ) strcat(Temp,",");
  }
  WritePrivateProfileString(Section,"VTColor",Temp,FName);

  /* VT bold color */
  for (i = 0;i<=1;i++)
  {
    ts->TmpColor[0][i*3]   = GetRValue(ts->VTBoldColor[i]);
    ts->TmpColor[0][i*3+1] = GetGValue(ts->VTBoldColor[i]);
    ts->TmpColor[0][i*3+2] = GetBValue(ts->VTBoldColor[i]);
  }
  Temp[0] = 0;
  for (i = 0;i<=5;i++)
  {
    sprintf(NumStr,"%d",ts->TmpColor[0][i]);
    strcat(Temp,NumStr);
    if ( i!=5 ) strcat(Temp,",");
  }
  WritePrivateProfileString(Section,"VTBoldColor",Temp,FName);

  /* VT blink color */
  for (i = 0;i<=1;i++)
  {
    ts->TmpColor[0][i*3]   = GetRValue(ts->VTBlinkColor[i]);
    ts->TmpColor[0][i*3+1] = GetGValue(ts->VTBlinkColor[i]);
    ts->TmpColor[0][i*3+2] = GetBValue(ts->VTBlinkColor[i]);
  }
  Temp[0] = 0;
  for (i = 0;i<=5;i++)
  {
    sprintf(NumStr,"%d",ts->TmpColor[0][i]);
    strcat(Temp,NumStr);
    if ( i!=5 ) strcat(Temp,",");
  }
  WritePrivateProfileString(Section,"VTBlinkColor",Temp,FName);

  /* TEK Color */
  for (i = 0;i<=1;i++)
  {
    ts->TmpColor[0][i*3]   = GetRValue(ts->TEKColor[i]);
    ts->TmpColor[0][i*3+1] = GetGValue(ts->TEKColor[i]);
    ts->TmpColor[0][i*3+2] = GetBValue(ts->TEKColor[i]);
  }
  Temp[0] = 0;
  for (i = 0;i<=5;i++)
  {
    sprintf(NumStr,"%d",ts->TmpColor[0][i]);
    strcat(Temp,NumStr);
    if ( i!=5 ) strcat(Temp,",");
  }
  WritePrivateProfileString(Section,"TEKColor",Temp,FName);

  /* TEK color emulation */
  if ( ts->TEKColorEmu == 1 ) strcpy(Temp,"on");
                             else strcpy(Temp,"off");
  WritePrivateProfileString(Section,"TEKColorEmulation",Temp,FName);

  /* VT Font */
  sprintf(Temp,"%s,%d,%d,%d",ts->VTFont,ts->VTFontSize.x,
             ts->VTFontSize.y,ts->VTFontCharSet);
  WritePrivateProfileString(Section,"VTFont",Temp,FName);

  /* Enable bold font flag */
  WriteOnOff(Section,"EnableBold",FName,ts->EnableBold);

  /* TEK Font */
  sprintf(Temp,"%s,%d,%d,%d",ts->TEKFont,ts->TEKFontSize.x,
             ts->TEKFontSize.y,ts->TEKFontCharSet);
  WritePrivateProfileString(Section,"TEKFont",Temp,FName);

  /* BS Key */
  if ( ts->BSKey == IdDEL ) strcpy(Temp,"DEL");
                            else strcpy(Temp,"BS");
  WritePrivateProfileString(Section,"BSKey",Temp,FName);

  /* IME Flag */
  WriteOnOff(Section,"IME",FName,ts->UseIME);

  /* IME Flag */
  WriteOnOff(Section,"IMEInline",FName,ts->IMEInline);

  /* Meta key */
  WriteOnOff(Section,"MetaKey",FName,ts->MetaKey);

  /* Serial port ID */
  switch (ts->ComPort) {
    case IdCOM2: strcpy(Temp,"2"); break;
    case IdCOM3: strcpy(Temp,"3"); break;
    case IdCOM4: strcpy(Temp,"4"); break;
    default:
      strcpy(Temp,"1");
  }
  WritePrivateProfileString(Section,"ComPort",Temp,FName);

  /* Baud rate */
  id2str(BaudList,ts->Baud,IdBaud9600,Temp);
  WritePrivateProfileString(Section,"BaudRate",Temp,FName);

  /* Parity */
  switch (ts->Parity) {
    case IdParityEven: strcpy(Temp,"even"); break;
    case IdParityOdd:  strcpy(Temp,"odd"); break;
    default:
      strcpy(Temp,"none");
  }
  WritePrivateProfileString(Section,"Parity",Temp,FName);

  /* Data bit */
  if (ts->DataBit==IdDataBit7)
    strcpy(Temp,"7");
  else
    strcpy(Temp,"8");

  WritePrivateProfileString(Section,"DataBit",Temp,FName);

  /* Stop bit */
  if (ts->StopBit == IdStopBit2)
    strcpy(Temp,"2");
  else
    strcpy(Temp,"1");

  WritePrivateProfileString(Section,"StopBit",Temp,FName);

  /* Flow control */
  switch (ts->Flow) {
    case IdFlowX:    strcpy(Temp,"x"); break;
    case IdFlowHard: strcpy(Temp,"hard"); break;
    default:
      strcpy(Temp,"none");
  }
  WritePrivateProfileString(Section,"FlowCtrl",Temp,FName);

  /* Delay per character */
  sprintf(Temp,"%d",ts->DelayPerChar);
  WritePrivateProfileString(Section,"DelayPerChar",Temp,FName);

  /* Delay per line */
  sprintf(Temp,"%d",ts->DelayPerLine);
  WritePrivateProfileString(Section,"DelayPerLine",Temp,FName);

  /* TCP port num */
  sprintf(Temp,"%d",ts->TCPPort);
  WritePrivateProfileString(Section,"TCPPort",Temp,FName);

  /* Telnet flag */
  WriteOnOff(Section,"Telnet",FName,ts->Telnet);

  /* Telnet terminal type */
  WritePrivateProfileString(Section,"TermType",ts->TermType,FName);

  /* Auto close flag */
  WriteOnOff(Section,"AutoWinClose",FName,ts->AutoWinClose);

  /* File transfer binary flag */
  WriteOnOff(Section,"TransBin",FName,ts->TransBin);

  /* XMODEM option */
  switch (ts->XmodemOpt) {
    case XoptCRC: strcpy(Temp,"crc"); break;
    case Xopt1K:  strcpy(Temp,"1k"); break;
    default:
    strcpy(Temp,"checksum");
  }
  WritePrivateProfileString(Section,"XmodemOpt",Temp,FName);

  /* XMODEM binary flag */
  WriteOnOff(Section,"XmodemBin",FName,ts->XmodemBin);

  /* Default directory for file transfer */
  WritePrivateProfileString(Section,"FileDir",ts->FileDir,FName);
}

#define VTEditor "VT editor keypad"
#define VTNumeric "VT numeric keypad"
#define VTFunction "VT function keys"
#define ShortCut "Shortcut keys"

WORD GetInt(PCHAR Sect, PCHAR Key, WORD Def, PCHAR FName)
{
  char Temp[11];
  WORD Num;

  GetPrivateProfileString(Sect,Key,"",
                          Temp,sizeof(Temp),FName);
  if ( Temp[0]==0 )
    return Def;
  else if ( stricmp(Temp,"off")==0 )
    return 0xFFFF;
  else {
    if (sscanf(Temp,"%d",&Num) !=1)
      Num = 0xFFFF;
    return Num;
  }
}

void FAR PASCAL ReadKeyboardCnf(PCHAR FName, PKeyMap KeyMap)
{
  int i, j, Ptr;
  char EntName[7];
  char TempStr[101];
  char KStr[101];

  /* Key Codes for VT editor keypad */
  KeyMap->Map[IdUp-1] = GetInt(VTEditor,"Up",328,FName);

  KeyMap->Map[IdDown-1] = GetInt(VTEditor,"Down",336,FName);

  KeyMap->Map[IdRight-1] = GetInt(VTEditor,"Right",333,FName);

  KeyMap->Map[IdLeft-1] = GetInt(VTEditor,"Left",331,FName);

  KeyMap->Map[IdFind-1] = GetInt(VTEditor,"Find",335,FName);

  KeyMap->Map[IdInsert-1] = GetInt(VTEditor,"Insert",338,FName);

  KeyMap->Map[IdRemove-1] = GetInt(VTEditor,"Remove",339,FName);

  KeyMap->Map[IdSelect-1] = GetInt(VTEditor,"Select",327,FName);

  KeyMap->Map[IdPrev-1] = GetInt(VTEditor,"Prev",329,FName);

  KeyMap->Map[IdNext-1] = GetInt(VTEditor,"Next",337,FName);

  /* Key Codes for VT numeric keypad */
  KeyMap->Map[Id0-1] = GetInt(VTNumeric,"Num0",82,FName);

  KeyMap->Map[Id1-1] = GetInt(VTNumeric,"Num1",79,FName);

  KeyMap->Map[Id2-1] = GetInt(VTNumeric,"Num2",80,FName);

  KeyMap->Map[Id3-1] = GetInt(VTNumeric,"Num3",81,FName);

  KeyMap->Map[Id4-1] = GetInt(VTNumeric,"Num4",75,FName);

  KeyMap->Map[Id5-1] = GetInt(VTNumeric,"Num5",76,FName);

  KeyMap->Map[Id6-1] = GetInt(VTNumeric,"Num6",77,FName);

  KeyMap->Map[Id7-1] = GetInt(VTNumeric,"Num7",71,FName);

  KeyMap->Map[Id8-1] = GetInt(VTNumeric,"Num8",72,FName);

  KeyMap->Map[Id9-1] = GetInt(VTNumeric,"Num9",73,FName);

  KeyMap->Map[IdMinus-1] = GetInt(VTNumeric,"NumMinus",1102,FName);

  KeyMap->Map[IdComma-1] = GetInt(VTNumeric,"NumComma",78,FName);

  KeyMap->Map[IdPeriod-1] = GetInt(VTNumeric,"NumPeriod",83,FName);

  KeyMap->Map[IdEnter-1] = GetInt(VTNumeric,"NumEnter",284,FName);

  KeyMap->Map[IdPF1-1] = GetInt(VTNumeric,"PF1",325,FName);

  KeyMap->Map[IdPF2-1] = GetInt(VTNumeric,"PF2",309,FName);

  KeyMap->Map[IdPF3-1] = GetInt(VTNumeric,"PF3",55,FName);

  KeyMap->Map[IdPF4-1] = GetInt(VTNumeric,"PF4",74,FName);

  /* Key Codes for VT function keys */
  KeyMap->Map[IdF6-1] = GetInt(VTFunction,"F6",64,FName);

  KeyMap->Map[IdF7-1] = GetInt(VTFunction,"F7",65,FName);

  KeyMap->Map[IdF8-1] = GetInt(VTFunction,"F8",66,FName);

  KeyMap->Map[IdF9-1] = GetInt(VTFunction,"F9",67,FName);

  KeyMap->Map[IdF10-1] = GetInt(VTFunction,"F10",68,FName);

  KeyMap->Map[IdF11-1] = GetInt(VTFunction,"F11",87,FName);

  KeyMap->Map[IdF12-1] = GetInt(VTFunction,"F12",88,FName);

  KeyMap->Map[IdF13-1] = GetInt(VTFunction,"F13",573,FName);

  KeyMap->Map[IdF14-1] = GetInt(VTFunction,"F14",574,FName);

  KeyMap->Map[IdHelp-1] = GetInt(VTFunction,"Help",575,FName);

  KeyMap->Map[IdDo-1] = GetInt(VTFunction,"Do",576,FName);

  KeyMap->Map[IdF17-1] = GetInt(VTFunction,"F17",577,FName);

  KeyMap->Map[IdF18-1] = GetInt(VTFunction,"F18",578,FName);

  KeyMap->Map[IdF19-1] = GetInt(VTFunction,"F19",579,FName);

  KeyMap->Map[IdF20-1] = GetInt(VTFunction,"F20",580,FName);

  KeyMap->Map[IdAccEditCopy-1] = GetInt(ShortCut,"EditCopy",65535,FName);

  KeyMap->Map[IdAccEditPaste-1] = GetInt(ShortCut,"EditPaste",65535,FName);

  KeyMap->Map[IdAccEditPasteCR-1] = GetInt(ShortCut,"EditPasteCR",65535,FName);

  KeyMap->Map[IdAccLineUp-1] = GetInt(ShortCut,"LineUp",1352,FName);

  KeyMap->Map[IdAccLineDown-1] = GetInt(ShortCut,"LineDown",1360,FName);

  KeyMap->Map[IdAccPageUp-1] = GetInt(ShortCut,"PageUp",1353,FName);

  KeyMap->Map[IdAccPageDown-1] = GetInt(ShortCut,"PageDown",1361,FName);

  KeyMap->Map[IdAccBuffTop-1] = GetInt(ShortCut,"BuffTop",1351,FName);

  KeyMap->Map[IdAccBuffBottom-1] = GetInt(ShortCut,"BuffBottom",1359,FName);

  KeyMap->Map[IdAccSendBreak-1] = GetInt(ShortCut,"SendBreak",65535,FName);

  /* user keys */

  for (i=0 ; i<=NumOfUserKey-1; i++)
  {
    KeyMap->Map[IdUser1+i-1] = 0xFFFF;
    KeyMap->UserKeyPtr[i] = 0;
    KeyMap->UserKeyLen[i] = 0;
  }

  Ptr = 0;

  strcpy(EntName,"User");
  i = IdUser1;
  do {
    if ( i < IdUser1+9 )
    {
      EntName[4] = (char)(i-IdUser1+0x31);
      EntName[5] = 0;
    }
    else {
      EntName[4] = (char)((i-IdUser1+1) / 10 + 0x30);
      EntName[5] = (char)((i-IdUser1+1) % 10 + 0x30);
      EntName[6] = 0;
    }

    GetPrivateProfileString("User keys",EntName,"",
                             TempStr,sizeof(TempStr),FName);
    if (strlen(TempStr)>0)
    {
      /* scan code */
      GetNthNum(TempStr,1,&j);
      KeyMap->Map[i-1] = j;
      /* conversion flag */
      GetNthNum(TempStr,2,&j);
      if ( j==0 ) KeyMap->UserKeyConv[i-IdUser1] = 0;
             else KeyMap->UserKeyConv[i-IdUser1] = 1;
      /* key string */
      GetNthString(TempStr,3,sizeof(KStr),KStr);
      KeyMap->UserKeyPtr[i-IdUser1] = Ptr;
      KeyMap->UserKeyLen[i-IdUser1] =
        Hex2Str(KStr,&(KeyMap->UserKeyStr[Ptr]),KeyStrMax-Ptr+1);
      Ptr = Ptr + KeyMap->UserKeyLen[i-IdUser1];
    }

    i++;
  }
  while ((i <= IdKeyMax) && (strlen(TempStr)>0) && (Ptr<=KeyStrMax));

}

/* copy hostlist from source IniFile to dest IniFile */
void FAR PASCAL CopyHostList(PCHAR IniSrc, PCHAR IniDest)
{
  int i;
  char EntName[7];
  char TempHost[HostNameMaxLength+1];

  if ( stricmp(IniSrc,IniDest)==0 ) return;

  WritePrivateProfileString("Hosts",NULL,NULL,IniDest);
  strcpy(EntName,"Host");

  i = 1;
  do {
    if ( i < 10 )
    {
      EntName[4] = (char)(i+0x30);
      EntName[5] = 0;
    }
    else {
      EntName[4] = (char)(i / 10 + 0x30);
      EntName[5] = (char)(i % 10 + 0x30);
      EntName[6] = 0;
    }

    /* Get one hostname from file IniSrc */
    GetPrivateProfileString("Hosts",EntName,"",
                            TempHost,sizeof(TempHost),IniSrc);
    /* Copy it to the file IniDest */
    if ( strlen(TempHost) > 0 )
      WritePrivateProfileString("Hosts",EntName,TempHost,IniDest);
    i++;
  }
  while ((i <= 99) && (strlen(TempHost)>0));

}
    
  BOOL NextParam(PCHAR Param, int *i, PCHAR Temp, int Size)
  {
    int j;
    char c;
    BOOL Quoted;
    
    if ( *i >= strlen(Param)) return FALSE;
    j = 0;

    while (Param[*i]==' ')
      (*i)++;
 
    Quoted = FALSE;
    c = Param[*i];
    (*i)++;
    while ((c!=0) && (Quoted || (c!=' ')) &&
           (Quoted || (c!=';')) && (j<Size-1))
    {
      if (c=='"')
        Quoted = ! Quoted;
      Temp[j] = c;
      j++;
      c = Param[*i];
      (*i)++;
    }
    if (! Quoted && (c==';'))
      (*i)--;

    Temp[j] = 0;
    return (strlen(Temp)>0);
  }

  void Dequote(PCHAR Source, PCHAR Dest)
  {
    int i, j;
    char q, c;

    Dest[0] = 0;
    if (Source[0]==0) return;
    i = 0;
    /* quoting char */
    q = Source[i];
    /* only " and ' are used as quoting char */
    if ((q!='"') && (q!='\''))
      q = 0;
    else
      i++;

    c = Source[i];
    i++;
    j = 0;
    while ((c!=0) && (c!=q))
    {
      Dest[j] = c;
      j++;
      c = Source[i];
      i++;
    }

    Dest[j] = 0;
  }

  void ConvFName(PCHAR HomeDir, PCHAR Temp, PCHAR DefExt, PCHAR FName)
  {
    int DirLen, FNPos;

    FName[0] = 0;
    if ( ! GetFileNamePosJ(Temp,&DirLen,&FNPos) ) return;
    FitFileName(&Temp[FNPos],DefExt);
    if ( DirLen==0 )
    {
      strcpy(FName,HomeDir);
      strcat(FName,"\\");
    }
    strcat(FName,Temp);
  }


void FAR PASCAL ParseParam(PCHAR Param, PTTSet ts, PCHAR DDETopic)
{
   int i, pos, c;
   BYTE b;
   char Temp[MAXPATHLEN+3];
   char Temp2[MAXPATHLEN];
   char TempDir[MAXPATHLEN];
   WORD ParamPort;
   WORD ParamCom;
   WORD ParamTCP;
   WORD ParamTel;
   WORD ParamBin;

  ts->HostName[0] = 0;
  ts->KeyCnfFN[0] = 0;
  ParamPort = 0;
  ParamTel = 2;
  ParamBin = 2;
  ParamTCP = 65535;
  ParamCom = 0;

  /* Get command line parameters */
  if (DDETopic != NULL)
    DDETopic[0] = 0;
  i = 0;
  /* the first term shuld be executable filename of Tera Term */
  NextParam(Param, &i, Temp, sizeof(Temp));
  while (NextParam(Param, &i, Temp, sizeof(Temp)))
  {
    if ( strnicmp(Temp,"/B",2)==0 ) /* telnet binary */
    {
      ParamPort = IdTCPIP;
      ParamBin = 1;
    }
    else if ( strnicmp(Temp,"/C=",3)==0 ) /* COM port num */
    {
      ParamPort = IdSerial;
      ParamCom = (WORD)(Temp[3]-0x30);
      if ( (ParamCom<1) || (ParamCom>4) )
        ParamCom = 0;
    }
    else if ( strnicmp(Temp,"/D=",3)==0 )
    {
      if (DDETopic != NULL)
        strcpy(DDETopic,&Temp[3]);
    }
    else if ( strnicmp(Temp,"/F=",3)==0 ) /* setup filename */
    {
      Dequote(&Temp[3],Temp2);
      if (strlen(Temp2)>0)
      {
        ConvFName(ts->HomeDir,Temp2,".INI",Temp);
        if (stricmp(ts->SetupFName,Temp)!=0)
	{
          strcpy(ts->SetupFName,Temp);
          ReadIniFile(ts->SetupFName,ts);
	}
      }
    }
    else if ( strnicmp(Temp,"/FD=",4)==0 ) /* file transfer directory */
    {
      Dequote(&Temp[4],Temp2);
      if (strlen(Temp2)>0)
      {
        getcwd(TempDir,sizeof(TempDir));
        if  (chdir(Temp2) == 0)
          strcpy(ts->FileDir,Temp2);
        chdir(TempDir);
      }
    }
    else if ( strnicmp(Temp,"/H",2)==0 ) /* hide title bar */
      ts->HideTitle = 1;
    else if ( strnicmp(Temp,"/K=",3)==0 ) /* Keyboard setup file */
    {  
      Dequote(&Temp[3],Temp2);
      ConvFName(ts->HomeDir,Temp2, ".CNF", ts->KeyCnfFN);
    }
    else if ( (strnicmp(Temp,"/KR=",4)==0)  ||
              (strnicmp(Temp,"/KT=",4)==0) ) /* kanji code */
    {
      if ( strnicmp(&Temp[4],"SJIS",4)==0 )
        c = IdSJIS;
      else if ( strnicmp(&Temp[4],"EUC",3)==0 )
        c = IdEUC;
      else if ( strnicmp(&Temp[4],"JIS",3)==0 )
        c = IdJIS;
      else
        c = -1;
      if ( c!=-1 )
      {
        if ( strnicmp(Temp,"/KR=",4)==0 )
          ts->KanjiCode = c;
        else
          ts->KanjiCodeSend = c;
      }
    }
    else if ( strnicmp(Temp,"/P=",3)==0 ) /* TCP port num */
    {
      ParamPort = IdTCPIP;
      if (sscanf(&Temp[3],"%d",&ParamTCP)!=1)
        ParamTCP = 65535;
    }
    else if ( strnicmp(Temp,"/R=",3)==0 ) /* Replay filename */
    {
      Dequote(&Temp[3],Temp2);
      ConvFName(ts->HomeDir,Temp2, "", ts->HostName);
      if ( strlen(ts->HostName)>0 )
        ParamPort = IdFile;
    }
    else if ( strnicmp(Temp,"/T=0",4)==0 ) /* telnet disable */
    {
      ParamPort = IdTCPIP;
      ParamTel = 0;
    }
    else if ( strnicmp(Temp,"/T=1",4)==0 ) /* telnet enable */
    {
      ParamPort = IdTCPIP;
      ParamTel = 1;
    }
    else if ( strnicmp(Temp,"/W=",3)==0 ) /* Window title */
    {
      Dequote(&Temp[3],ts->Title);
    }
    else if ( strnicmp(Temp,"/X=",3)==0 ) /* Window pos (X) */
    {
      if (sscanf(&Temp[3],"%d",&pos)==1)
      {
        ts->VTPos.x = pos;
	if (ts->VTPos.y==CW_USEDEFAULT)
	  ts->VTPos.y = 0;
      }
    }
    else if ( strnicmp(Temp,"/Y=",3)==0 ) /* Window pos (Y) */
    {
      if (sscanf(&Temp[3],"%d",&pos)==1)
      {
        ts->VTPos.y = pos;
	if (ts->VTPos.x==CW_USEDEFAULT)
	  ts->VTPos.x = 0;
      }
    }
    else if ( (Temp[0]!='/') && (strlen(Temp)>0) )
    {
      ParamPort = IdTCPIP;
      strcpy(ts->HostName,Temp); /* host name */
    }
  }

  if ((ts->HostName[0]!=0) &&
      (ParamPort==IdTCPIP))
  {
    i = 0;
    do {
      b = ts->HostName[i];
      i++;
    } while ((b!=0) && (b!=':'));
    if (b==':')
    {
      ts->HostName[i-1] = 0;
      if (sscanf(&(ts->HostName[i]),"%d",&ParamTCP)!=1)
        ParamTCP = 65535;
    }
  }

  switch (ParamPort) {
    case IdTCPIP:
      ts->PortType = IdTCPIP;
      if ( ParamTel<2 ) ts->Telnet = ParamTel;
      if ( ParamBin<2 ) ts->TelBin = ParamBin;
      if ( ParamTCP<65535 ) ts->TCPPort = ParamTCP;
      break;
    case IdSerial:
      ts->PortType = IdSerial;
      if ( (ParamCom>0) && (ParamCom<5) )
        ts->ComPort = ParamCom;
      break;
    case IdFile:
      ts->PortType = IdFile;
  }
}

#pragma off (unreferenced);
int APIENTRY LibMain(HANDLE hdll,
                     DWORD  reason,
                     LPVOID reserved)
#pragma on (unreferenced);
{
  switch( reason ) { 
  case DLL_THREAD_ATTACH:
     /* do thread initialization */
    break;
  case DLL_THREAD_DETACH:
    /* do thread cleanup */
    break;
  case DLL_PROCESS_ATTACH:
     /* do process initialization */
    break;
  case DLL_PROCESS_DETACH:
    /* do process cleanup */
    break;
  }
   return( 1 );
 }
