/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, VT terminal display routines */
#include "teraterm.h"
#include "tttypes.h"
#include "ttwinman.h"
#include "buffer.h"
#include "scroll.h"
#include "ttime.h"
#include "kanji.h"
#include "ttftypes.h"
#include "filesys.h"
#include "telnet.h"
#include "teraprn.h"

#define CurWidth 2

HFONT VTFont[AttrFontMask];
LOGFONT VTlf;
int WinWidth, WinHeight, WinWidthOld, WinHeightOld;
BOOL Active, CompletelyVisible, Minimized;
int CursorX, CursorY;
int CursorTop, CursorBottom;
    /* mouse status */
BOOL LButton, MButton, RButton, DblClk, TplClk;
int DblClkX, DblClkY;
BYTE CharAttr, CharAttr2;
BOOL Special;
WORD TabStops[256];
int NTabStops;
    /* GL, GR code group */
int Glr[2];
    /* G0, G1, G2, G3 code group */
int Gn[4];
    /* single shift 2/3 flag */
BOOL SSflag;
    /* GL for single shift 2/3 */
int GLtmp;
    /* single shift 2/3 flag */
BOOL SSflag;
    /* JIS -> SJIS conversion flag */
BOOL ConvJIS;
WORD Kanji;
int StrChangeCount;
int FontHeight, FontWidth, ScreenWidth, ScreenHeight;
BOOL InsertMode;
BOOL LFMode;
BOOL AutoWrapMode;
HBRUSH Background;
BOOL AdjustSize;
BOOL DontChangeSize=FALSE;

static COLORREF ANSIColor[8];
int CRTWidth, CRTHeight;
static INT Dx[256];
static int CaretStatus;

static int StrChangeStart;
static BOOL Wrap;

void InitDisp()
{
  HDC TmpDC;
  int i;

  TmpDC = GetDC(NULL);

  ANSIColor[IdBack ]   = RGB(  0,  0,  0);
  ANSIColor[IdRed  ]   = RGB(255,  0,  0);
  ANSIColor[IdGreen]   = RGB(  0,255,  0);
  ANSIColor[IdYellow]  = RGB(255,255,  0);
  ANSIColor[IdBlue]    = RGB(  0,  0,255);
  ANSIColor[IdMagenta] = RGB(255,  0,255);
  ANSIColor[IdCyan]    = RGB(  0,255,255);
  ANSIColor[IdFore ]   = RGB(255,255,255);

  for (i = IdBack ; i <= IdFore ; i++)
    ANSIColor[i] = GetNearestColor(TmpDC, ANSIColor[i]);

  /* background paintbrush */
  Background = CreateSolidBrush(ts.VTColor[1]);
  /* CRT width & height */
  CRTWidth = GetDeviceCaps(TmpDC,HORZRES);
  CRTHeight = GetDeviceCaps(TmpDC,VERTRES);

  ReleaseDC(NULL, TmpDC);

  if ((ts.VTPos.x > CRTWidth) || (ts.VTPos.y > CRTHeight))
  {
    ts.VTPos.x = CW_USEDEFAULT;
    ts.VTPos.y = CW_USEDEFAULT;
  }

  if ((ts.TEKPos.x > CRTWidth) || (ts.TEKPos.y > CRTHeight))
  {
    ts.TEKPos.x = CW_USEDEFAULT;
    ts.TEKPos.y = CW_USEDEFAULT;
  }
}

void EndDisp()
{
  int i, j;

  /* Delete fonts */
  for (i = 0 ; i <= AttrFontMask; i++)
  {
    for (j = i+1 ; j <= AttrFontMask ; j++)
      if (VTFont[j]==VTFont[i])
        VTFont[j] = 0;
    if (VTFont[i]!=0) DeleteObject(VTFont[i]);
  }

  DeleteObject(Background);
}

void ResetDisp()
{
  int i;

  /* Cursor */
  CursorX = 0;
  CursorY = 0;

  /* Top/bottom mergin */
  CursorTop = 0;
  CursorBottom = NumOfLines-1;

  LButton = FALSE;
  MButton = FALSE;
  RButton = FALSE;

  /* Tab stops */
  NTabStops = (NumOfColumns-1) >> 3;
  for (i=1 ; i<=NTabStops ; i++)
    TabStops[i-1] = i*8;

  /* Attribute */
  CharAttr = AttrDefault;
  CharAttr2 = AttrDefault2;
  Special = FALSE;

  /* Various modes */
  InsertMode = FALSE;
  Wrap = FALSE;
  AutoWrapMode = TRUE;
  LFMode = (ts.CRSend == IdCRLF);

  StrChangeCount = 0;
  ScrollCount = 0;
  dScroll = 0;
}

void SetLogFont()
{
  memset(&VTlf, 0, sizeof(LOGFONT));
  VTlf.lfWeight = FW_NORMAL;
  VTlf.lfItalic = 0;
  VTlf.lfUnderline = 0;
  VTlf.lfStrikeOut = 0;
  VTlf.lfWidth = ts.VTFontSize.x;
  VTlf.lfHeight = ts.VTFontSize.y;
  VTlf.lfCharSet = ts.VTFontCharSet;

  VTlf.lfOutPrecision  = OUT_CHARACTER_PRECIS;
  VTlf.lfClipPrecision = CLIP_CHARACTER_PRECIS;
  VTlf.lfQuality       = DEFAULT_QUALITY;
  VTlf.lfPitchAndFamily= FIXED_PITCH | FF_MODERN;
  strcpy(VTlf.lfFaceName,ts.VTFont);
}

void ChangeFont()
{
  int i, j;
  TEXTMETRIC Metrics;
  HDC TmpDC;

  /* Delete Old Fonts */
  for (i = 0 ; i <= AttrFontMask ; i++)
  {
    for (j = i+1 ; j <= AttrFontMask ; j++)
      if (VTFont[j]==VTFont[i])
        VTFont[j] = 0;
    if (VTFont[i]!=0)
      DeleteObject(VTFont[i]);
  }

  /* Normal Font */
  SetLogFont();
  VTFont[0] = CreateFontIndirect(&VTlf);

  /* set IME font */
  SetConversionLogFont(&VTlf);

  TmpDC = GetDC(HVTWin);

  SelectObject(TmpDC, VTFont[0]);
  GetTextMetrics(TmpDC, &Metrics); 
  FontWidth = Metrics.tmAveCharWidth + ts.FontDW;
  FontHeight = Metrics.tmHeight + ts.FontDH;

  ReleaseDC(HVTWin,TmpDC);

  /* Underline */
  VTlf.lfUnderline = 1;
  VTFont[AttrUnder] = CreateFontIndirect(&VTlf);

  if (ts.EnableBold>0)
  {
    /* Bold */
    VTlf.lfUnderline = 0;
    VTlf.lfWeight = FW_BOLD;
    VTFont[AttrBold] = CreateFontIndirect(&VTlf);
    /* Bold + Underline */
    VTlf.lfUnderline = 1;
    VTFont[AttrBold | AttrUnder] = CreateFontIndirect(&VTlf);
  }
  else {
    VTFont[AttrBold] = VTFont[AttrDefault];
    VTFont[AttrBold | AttrUnder] = VTFont[AttrUnder];
  }
  /* Special font */
  VTlf.lfWeight = FW_NORMAL;
  VTlf.lfUnderline = 0;
  VTlf.lfWidth = FontWidth + 1; /* adjust width */
  VTlf.lfHeight = FontHeight;
  VTlf.lfCharSet = 1;
  VTlf.lfPitchAndFamily= DEFAULT_PITCH | FF_DONTCARE;

  strcpy(VTlf.lfFaceName,"Tera Special");
  VTFont[AttrSpecial] = CreateFontIndirect(&VTlf);
  VTFont[AttrSpecial | AttrBold] = VTFont[AttrSpecial];
  VTFont[AttrSpecial | AttrUnder] = VTFont[AttrSpecial];
  VTFont[AttrSpecial | AttrBold | AttrUnder] = VTFont[AttrSpecial];

  SetLogFont();

  for (i = 0 ; i <= 255; i++)
    Dx[i] = FontWidth;
}

void ResetIME()
{
  /* reset language for communication */
  cv.Language = ts.Language;

  /* reset IME */
  if (ts.Language==IdJapanese)
  {
    if (ts.UseIME==0)
      FreeWINNLS();
    else if (! LoadWINNLS())
      ts.UseIME = 0;

    if ((ts.UseIME==1) &&
        (ts.IMEInline > 0))
    {
      SetConversionLogFont(&VTlf);
      CaretOn();
    }
  }
  else
    FreeWINNLS();
}

void ChangeCaret()
{
  DestroyCaret();
  switch (ts.CursorShape) {
    case IdVCur:
      CreateCaret(HVTWin, 0, CurWidth, FontHeight);
      break;
    case IdHCur:
      CreateCaret(HVTWin, 0, FontWidth, CurWidth);
      break;
  }
  CaretStatus = 1;
  CaretOn();
}

void CaretOn()
{
  int CaretX, CaretY, H;

  if (! Active) return;

  CaretX = (CursorX-WinOrgX)*FontWidth;
  CaretY = (CursorY-WinOrgY)*FontHeight;

  if ((ts.Language==IdJapanese) &&
      CanUseIME() && (ts.IMEInline>0))
    /* set IME conversion window pos. & font */
    SetConversionWindow(HVTWin,CaretX,CaretY);

  if (ts.CursorShape!=IdVCur)
  {
    if (ts.CursorShape==IdHCur)
    {
     CaretY = CaretY+FontHeight-CurWidth;
     H = CurWidth;
    }
    else H = FontHeight;

    DestroyCaret();
    LockBuffer();
    NewLine(PageStart+CursorY);
    if (((BYTE)(AttrLine[CursorX]) & AttrKanji) != 0) /* DBCS first? */
      CreateCaret(HVTWin, 0, FontWidth*2, H); /* double width caret */
    else
      CreateCaret(HVTWin, 0, FontWidth, H); /* single width caret */
    UnlockBuffer();
    CaretStatus = 1;
  }

  SetCaretPos(CaretX,CaretY);

  while (CaretStatus > 0)
  {
    ShowCaret(HVTWin);
    CaretStatus--;
  }

}

void CaretOff()
{
  if (! Active) return;
  if (CaretStatus == 0)
  {
    HideCaret(HVTWin);
    CaretStatus++;
  }
}

void ChangeWindowSize(int Nx, int Ny)
{
  LONG W,H,dW,dH;
  RECT R;


  WinWidth = Nx;
  WinHeight = Ny;

  if (ts.TermIsWin>0)
  {
    LockBuffer();
    ChangeTerminalSize(WinWidth,WinHeight);
    UnlockBuffer();
    CaretOn();
    WinWidth = NumOfColumns;
    WinHeight = NumOfLines;
  }

  ScreenWidth = WinWidth*FontWidth;
  ScreenHeight = WinHeight*FontHeight;

  AdjustScrollBar();

  GetWindowRect(HVTWin,&R);
  W = R.right-R.left;
  H = R.bottom-R.top;
  GetClientRect(HVTWin,&R);
  dW = ScreenWidth - R.right + R.left;
  dH = ScreenHeight - R.bottom + R.top;
  
  if ((dW!=0) || (dH!=0))
  {
    AdjustSize = TRUE;
    SetWindowPos(HVTWin,HWND_TOP,0,0,W+dW,H+dH,SWP_NOMOVE);
  }
  else
    InvalidateRect(HVTWin,NULL,FALSE);
}

void ChangeTerminalSize(int Nx, int Ny)
{
  int i, Nb, W, H;

  if (Nx < 1) Nx = 1;
  if (Ny < 1) Ny = 1;
  if ((Nx==NumOfColumns) && (Ny==NumOfLines)) return;
  if (Nx > BuffXMax) Nx = BuffXMax;
  if (Ny > BuffYMax) Ny = BuffYMax;

  if ((ts.ScrollBuffMax < Ny) ||
      (ts.EnableScrollBuff==0))
   Nb = Ny;
  else Nb = ts.ScrollBuffMax;

  if (! ChangeBuffer(Nx,Nb)) return;
  if (ts.EnableScrollBuff>0)
    ts.ScrollBuffMax = NumOfLinesInBuff;
  if (Ny > NumOfLinesInBuff) Ny = NumOfLinesInBuff;

  NumOfColumns = Nx;
  NumOfLines = Ny;
  WinWidthOld = Nx;
  WinHeightOld = Ny;
  ts.TerminalWidth = Nx;
  ts.TerminalHeight = Ny;

  BuffScroll(NumOfLines);
  /* Set Cursor */
  CursorX = 0;
  CursorY = 0;
  CursorTop = 0;
  CursorBottom = NumOfLines-1;

  SelectStart.x = 0;
  SelectStart.y = 0;
  SelectEnd = SelectStart;
  Select = FALSE;

  /* Tab stops */
  NTabStops = (NumOfColumns-1) >> 3;
  for (i = 1 ; i <= NTabStops ; i++)
    TabStops[i-1] = i*8;

  if (ts.TermIsWin>0)
  {
    W = NumOfColumns;
    H = NumOfLines;
  }
  else {
    W = WinWidth;
    H = WinHeight;
    if ((ts.AutoWinResize>0) ||
        (NumOfColumns < W)) W = NumOfColumns;
    if (ts.AutoWinResize>0) H = NumOfLines;
    else if (BuffEnd < H) H = BuffEnd;
  }
                              
  PageStart = BuffEnd-NumOfLines;

  /* Change Window Size */
  ChangeWindowSize(W,H);
  WinOrgY = -NumOfLines;
  ScrollHomePos();

  if (cv.Ready && cv.TelFlag &&
      (tr.MyOpt[NAWS].Status == Yes) &&
      ((NumOfColumns!=tr.WinSize.x) ||
       (NumOfLines!=tr.WinSize.y)))
  {
    tr.WinSize.x = NumOfColumns;
    tr.WinSize.y = NumOfLines;
    TelInformWinSize();
  }

}

void ResizeWindow(int x, int y, int w, int h, int cw, int ch)
{
  int dw,dh, NewX, NewY, Nx, Ny;
  POINT Point;

  if (AdjustSize)
  {
    dw = ScreenWidth - cw;
    dh = ScreenHeight - ch;
    if ((dw!=0) || (dh!=0))
      SetWindowPos(HVTWin,HWND_TOP,x,y,w+dw,h+dh,SWP_NOMOVE);
    else {
      AdjustSize = FALSE;

      NewX = x;
      NewY = y;
      if (x+w > CRTWidth)
      {
        NewX = CRTWidth-w;
        if (NewX < 0) NewX = 0;
      }
      if (y+h > CRTHeight)
      {
        NewY = CRTHeight-h;
        if (NewY < 0) NewY = 0;
      }
      if ((NewX!=x) || (NewY!=y))
        SetWindowPos(HVTWin,HWND_TOP,NewX,NewY,w,h,SWP_NOSIZE);

      Point.x = 0;
      Point.y = ScreenHeight;
      ClientToScreen(HVTWin,&Point);
      CompletelyVisible = (Point.y <= CRTHeight);
      CaretOn();
    }
  }
  else {
    Nx = cw / FontWidth;
    Ny = ch / FontHeight;
    if (ts.TermIsWin>0)
    {
      if (Nx > BuffXMax) Nx = BuffXMax;
      if (Ny > BuffYMax) Ny = BuffYMax;
    }
    else {
      if (Nx>NumOfColumns) Nx = NumOfColumns;
      if (Ny>BuffEnd) Ny = BuffEnd;
    }
    if (Nx==0) Nx = 1;
    if (Ny==0) Ny = 1;
    ChangeWindowSize(Nx,Ny);
  }
}

void PaintWindow(HDC PaintDC, LPPAINTSTRUCT ps)
{
  int XStart, XEnd, YStart, YEnd; 

  if (ps->fErase)
    FillRect(PaintDC, &(ps->rcPaint),Background);

  XStart = ps->rcPaint.left / FontWidth + WinOrgX;
  YStart = ps->rcPaint.top / FontHeight + WinOrgY;
  XEnd = (ps->rcPaint.right-1) / FontWidth + WinOrgX;
  YEnd = (ps->rcPaint.bottom-1) / FontHeight + WinOrgY;

  LockBuffer();
  UpdateRect(PaintDC,XStart,YStart,XEnd,YEnd);
  UnlockBuffer();
}

BOOL CheckSelect(int x, int y)
{
  LONG L, L1, L2;

  L = MAKELONG(x,y);
  L1 = MAKELONG(SelectStart.x,SelectStart.y);
  L2 = MAKELONG(SelectEnd.x,SelectEnd.y);

  return (Select &&
          ((L1<=L) && (L<L2) || (L2<=L) && (L<L1)));
}

void SetupDC()
{
  COLORREF TextColor, BackColor;
  int IFont;

  IFont = CharAttr & AttrFontMask;
  if (Special) IFont = IFont | AttrSpecial;
  SelectObject(VTDC, VTFont[IFont]);

  if ((CharAttr & AttrBlink) != 0)
    TextColor = ts.VTBlinkColor[0];
  else if ((CharAttr & AttrBold) != 0)
    TextColor = ts.VTBoldColor[0];
  else if ((CharAttr2 & Attr2Fore) != 0)
    TextColor = ANSIColor[CharAttr2 & Attr2ForeMask];
  else
    TextColor = ts.VTColor[0];

  if ((CharAttr & AttrBlink) != 0)
    BackColor = ts.VTBlinkColor[1];
  else if ((CharAttr & AttrBold) != 0)
    BackColor = ts.VTBoldColor[1];    
  else if ((CharAttr2 & Attr2Back) != 0)
    BackColor = ANSIColor[(CharAttr2 & Attr2BackMask) >> SftAttrBack];
  else
    BackColor = ts.VTColor[1];

  if ((CharAttr & AttrReverse) != 0)
  {
    SetTextColor(VTDC,BackColor);
    SetBkColor(VTDC,TextColor);
  }
  else {
    SetTextColor(VTDC,TextColor);
    SetBkColor(  VTDC,BackColor);
  }

  SetBkMode(VTDC,OPAQUE);
}

void UpdateStr()
{
  int CY;
  RECT R;

  if (StrChangeCount==0) return;

  CY = CursorY;

  if ((dScroll != 0) &&
      (CY>=SRegionTop) &&
      (CY<=SRegionBottom))
  {
    CY = CY + dScroll;
    if ((CY<SRegionTop) || (CY>SRegionBottom))
    {
      StrChangeCount = 0;
      return;
    }
  }

  if ((CY<WinOrgY) ||
      (CY>=WinOrgY+WinHeight))
  {
    StrChangeCount = 0;
    return;
  }

  R.left = (StrChangeStart-WinOrgX)*FontWidth;
  R.right = R.left + StrChangeCount*FontWidth;
  R.top = (CY-WinOrgY)*FontHeight;
  R.bottom = R.top + FontHeight;

  ExtTextOut(VTDC,R.left+ts.FontDX,R.top+ts.FontDY,
             6,&R,&CodeLine[StrChangeStart],StrChangeCount,&Dx[0]);
  StrChangeCount = 0;
}

void UpdateRect
  (HDC UpdateDC, int XStart, int YStart, int XEnd, int YEnd)
{
  int i, j, count;
  int IStart, IEnd;
  int X, Y;
  LONG TmpPtr;
  BYTE CurAttr, TempAttr;
  BYTE CurAttr2, TempAttr2;
  BOOL CurSel, TempSel;
  RECT RText;
  HFONT PrevFont;
  COLORREF TextColor, BackColor;

  if (XStart >= WinOrgX+WinWidth) return;
  if (YStart >= WinOrgY+WinHeight) return;
  if (XEnd < WinOrgX) return;
  if (YEnd < WinOrgY) return;

  if (XStart < WinOrgX) XStart = WinOrgX;
  if (YStart < WinOrgY) YStart = WinOrgY;
  if (XEnd >= WinOrgX+WinWidth) XEnd = WinOrgX+WinWidth-1;
  if (YEnd >= WinOrgY+WinHeight) YEnd = WinOrgY+WinHeight-1;

  PrevFont = SelectObject(UpdateDC, VTFont[0]);
  SetTextColor(UpdateDC, ts.VTColor[0]);
  SetBkColor(UpdateDC, ts.VTColor[1]);

  TempAttr = AttrDefault;
  TempAttr2 = AttrDefault2;
  TempSel = FALSE;

  Y = (YStart-WinOrgY)*FontHeight;
  TmpPtr = GetLinePtr(PageStart+YStart);
  for (j = YStart+PageStart ; j <= YEnd+PageStart ; j++)
  {
    IStart = XStart;
    IEnd = XEnd;

    if ((IStart>0) &&
        ((AttrBuff[TmpPtr+IStart-1] &
         AttrKanji) != 0)) /* DBCS first byte? */
      IStart--;

    X = (IStart-WinOrgX)*FontWidth;
    RText.top = Y;
    RText.bottom = Y + FontHeight;

    i = IStart;
    do {
      CurAttr = AttrBuff[TmpPtr+i] & ~ AttrKanji;
      CurAttr2 = AttrBuff2[TmpPtr+i];
      CurSel = CheckSelect(i,j);
      count = 1;
      while
        ( (i+count <= IEnd) &&
          (CurAttr==
           (AttrBuff[TmpPtr+i+count] & ~ AttrKanji)) &&
          (CurAttr2==AttrBuff2[TmpPtr+i+count]) &&
          (CurSel==CheckSelect(i+count,j)) ||
          ((AttrBuff[TmpPtr+i+count-1] & AttrKanji) != 0) )
        count++;
        
      if ((CurAttr != TempAttr) ||
          (CurAttr2 != TempAttr2) ||
          (CurSel != TempSel))
      {
        SelectObject(UpdateDC, VTFont[CurAttr & AttrFontMask]);

        if ((CurAttr & AttrBlink) != 0)
          TextColor = ts.VTBlinkColor[0];
        else if ((CurAttr & AttrBold) != 0)
          TextColor = ts.VTBoldColor[0];          
        else if ((CurAttr2 & Attr2Fore) != 0)
          TextColor = ANSIColor[CurAttr2 & Attr2ForeMask];
        else
          TextColor = ts.VTColor[0];

        if ((CurAttr & AttrBlink) != 0)
          BackColor = ts.VTBlinkColor[1];
        else if ((CurAttr & AttrBold) != 0)
          BackColor = ts.VTBoldColor[1];
        else if ((CurAttr2 & Attr2Back) != 0)
          BackColor = ANSIColor[(CurAttr2 & Attr2BackMask) >> SftAttrBack];
        else
          BackColor = ts.VTColor[1];

        if (CurSel != ((CurAttr & AttrReverse) != 0))
        {
          SetTextColor(UpdateDC,BackColor);
          SetBkColor(  UpdateDC,TextColor);
        }
        else {
          SetTextColor(UpdateDC,TextColor);
          SetBkColor(  UpdateDC,BackColor);
        }
        TempAttr = CurAttr;
        TempAttr2 = CurAttr2;
        TempSel = CurSel;
      }
      RText.left = X;
      RText.right = X + count*FontWidth;
      ExtTextOut(UpdateDC,X+ts.FontDX,Y+ts.FontDY,
                 ETO_CLIPPED | ETO_OPAQUE,
                 &RText,&CodeBuff[TmpPtr+i],count,&Dx[0]);

      X = RText.right;
      i = i+count;
    }
    while (i<=IEnd);
    Y = Y + FontHeight;
    TmpPtr = NextLinePtr(TmpPtr);
  }
  SelectObject(UpdateDC, PrevFont);
  if (UpdateDC==VTDC) SetupDC();
}

void MoveRight()
/* move cursor right, but dont update screen.
  this procedure must be called from DispChar&DispKanji only */
{
  CursorX++;

  ScrollToCursor(CursorX, CursorY);
}

void EraseKanji(int LR)
{
  if ((CursorX-LR>=0) &&
      ((AttrLine[CursorX-LR] & AttrKanji) != 0))
  {
    CodeLine[CursorX-LR] = 0x20;
    AttrLine[CursorX-LR] = AttrDefault;
    AttrLine2[CursorX-LR] = AttrDefault2;
    if (CursorX-LR+1 < NumOfColumns)
    {
      CodeLine[CursorX-LR+1] = 0x20;
      AttrLine[CursorX-LR+1] = AttrDefault;
      AttrLine2[CursorX-LR+1] = AttrDefault2;
    }
  }
}

void MoveCursor(int Xnew, int Ynew)
{
  UpdateStr();

  if (CursorY!=Ynew) NewLine(PageStart+Ynew);

  CursorX = Xnew;
  CursorY = Ynew;
  Wrap = FALSE;

  ScrollToCursor(CursorX, CursorY);
}

void CursorUpWithScroll()
{
  if ((0<CursorY) && (CursorY<CursorTop) ||
      (CursorTop<CursorY))
    MoveCursor(CursorX,CursorY-1);
  else if (CursorY==CursorTop)
    ScrollUp1Line();
}

void SetTabStop()
{
  int i,j;

  if (NTabStops<NumOfColumns)
  {
    i = 0;
    while ((TabStops[i]<CursorX) && (i<NTabStops))
      i++;
    
    if ((i<NTabStops) && (TabStops[i]==CursorX)) return;

    for (j=NTabStops ; j>=i+1 ; j--)
      TabStops[j] = TabStops[j-1];
    TabStops[i] = CursorX;
    NTabStops++;
  }
}

void Tab()
{
  int i;

  if (NTabStops>0)
  {
    i = -1;
    do
      i++;
    while ((TabStops[i]<=CursorX) && (i<NTabStops-1));
    if (TabStops[i]>CursorX)
      MoveCursor(TabStops[i],CursorY);
    else
      MoveCursor(NumOfColumns-1,CursorY);
  }
  else
    MoveCursor(NumOfColumns-1,CursorY);
    if (cv.HLogBuf!=0) Log1Byte(HT);
}

void BackSpace()
{
  if (CursorX > 0)
  {
    MoveCursor(CursorX-1,CursorY);
    if (cv.HLogBuf!=0) Log1Byte(BS);
  }
}

void CarriageReturn()
{
 if (cv.HLogBuf!=0) Log1Byte(CR);
 if (CursorX>0)
   MoveCursor(0,CursorY);
}

void LineFeed(BYTE b)
{
 /* for auto print mode */
 if ((AutoPrintMode) &&
     (b>=LF) && (b<=FF))
   PrintCurrentLine(b);

 if (cv.HLogBuf!=0) Log1Byte(LF);

 if (CursorY < CursorBottom)
   MoveCursor(CursorX,CursorY+1);
 else if (CursorY == CursorBottom) Scroll1Line();
 else if (CursorY < NumOfLines-1)
   MoveCursor(CursorX,CursorY+1);

 if (LFMode) CarriageReturn();
}

void DispChar(BYTE b)
{
  BOOL SpecialNew;
  int XStart;
  BYTE CharAttrTmp;

  if (Wrap)
  {
    CarriageReturn();
    LineFeed(LF);
  }
  if (cv.HLogBuf!=0) Log1Byte(b);
  Wrap = FALSE;

  if (ts.Language==IdJapanese)
  {
    EraseKanji(1); /* if cursor is on right half of a kanji, erase the kanji */
    if (! InsertMode) EraseKanji(0); /* if cursor on left half... */
  }

  SpecialNew = FALSE;
  if ((b>0x5F) && (b<0x80))
  {
    if (SSflag)
      SpecialNew = (Gn[GLtmp]==IdSpecial);
    else
      SpecialNew = (Gn[Glr[0]]==IdSpecial);
  }
  else if (b>0xDF)
  {
    if (SSflag)
      SpecialNew = (Gn[GLtmp]==IdSpecial);
    else
      SpecialNew = (Gn[Glr[1]]==IdSpecial);
  }

  if (SpecialNew != Special)
  {
    UpdateStr();
    Special = SpecialNew;
    SetupDC();
  }

  if (Special)
  {
    b = b & 0x7F;
    CharAttrTmp = CharAttr | AttrSpecial;
  }
  else
    CharAttrTmp = CharAttr;

  if (InsertMode)
  {
    memmove(&CodeLine[CursorX+1],&CodeLine[CursorX],NumOfColumns-1-CursorX);
    memmove(&AttrLine[CursorX+1],&AttrLine[CursorX],NumOfColumns-1-CursorX);
    memmove(&AttrLine2[CursorX+1],&AttrLine2[CursorX],NumOfColumns-1-CursorX);
    CodeLine[CursorX] = b;
    AttrLine[CursorX] = CharAttrTmp;
    AttrLine2[CursorX] = CharAttr2;
    /* last char in current line is kanji first? */
    if ((AttrLine[NumOfColumns-1] & AttrKanji) != 0)
    {
      /* then delete it */
      CodeLine[NumOfColumns-1] = 0x20;
      AttrLine[NumOfColumns-1] = AttrDefault;
      AttrLine2[NumOfColumns-1] = AttrDefault2;
    }

    if (StrChangeCount==0) XStart = CursorX;
                      else XStart = StrChangeStart;
    StrChangeCount = 0;
    UpdateRect(VTDC,XStart,CursorY,NumOfColumns-1,CursorY);
  }
  else {
    CodeLine[CursorX] = b;
    AttrLine[CursorX] = CharAttrTmp;
    AttrLine2[CursorX] = CharAttr2;

    if (StrChangeCount==0) StrChangeStart = CursorX;
    StrChangeCount++;
  }

  if (CursorX < NumOfColumns-1)
    MoveRight();
  else {
    UpdateStr();
    Wrap = AutoWrapMode;
  }
}

void DispKanji(BYTE b)
{
  int XStart;

  Kanji = Kanji + b;

  if (ConvJIS)
    Kanji = JIS2SJIS(Kanji & 0x7f7f);

  if (Wrap)
  {
    CarriageReturn();
    LineFeed(LF);
  }
  else if (CursorX > NumOfColumns-2)
    if (AutoWrapMode)
    {
      CarriageReturn();
      LineFeed(LF);
    }
    else return;

  Wrap = FALSE;

  if (cv.HLogBuf!=0)
  {
    Log1Byte(HIBYTE(Kanji));
    Log1Byte(LOBYTE(Kanji));
  }

  EraseKanji(1); /* if cursor is on right half of a kanji, erase the kanji */

  if (Special)
  {
    UpdateStr();
    Special = FALSE;
    SetupDC();
  }

  if (InsertMode)
  {
    memmove(&CodeLine[CursorX+2],&CodeLine[CursorX],NumOfColumns-2-CursorX);
    memmove(&AttrLine[CursorX+2],&AttrLine[CursorX],NumOfColumns-2-CursorX);
    memmove(&AttrLine2[CursorX+2],&AttrLine2[CursorX],NumOfColumns-2-CursorX);

    CodeLine[CursorX] = HIBYTE(Kanji);
    AttrLine[CursorX] = CharAttr | AttrKanji; /* DBCS first byte */
    AttrLine2[CursorX] = CharAttr2;
    if (CursorX < NumOfColumns-1)
    {
      CodeLine[CursorX+1] = LOBYTE(Kanji);
      AttrLine[CursorX+1] = CharAttr;
      AttrLine2[CursorX+1] = CharAttr2;
    }

    /* last char in current line is kanji first? */
    if ((AttrLine[NumOfColumns-1] & AttrKanji) != 0)
    {
      /* then delete it */
      CodeLine[NumOfColumns-1] = 0x20;
      AttrLine[NumOfColumns-1] = AttrDefault;
      AttrLine2[NumOfColumns-1] = AttrDefault2;
    }

    if (StrChangeCount==0) XStart = CursorX;
                      else XStart = StrChangeStart;
    StrChangeCount = 0;
    UpdateRect(VTDC,XStart,CursorY,NumOfColumns-1,CursorY);
  }
  else {
    CodeLine[CursorX] = HIBYTE(Kanji);
    AttrLine[CursorX] = CharAttr | AttrKanji; /* DBCS first byte */
    AttrLine2[CursorX] = CharAttr2;
    if (CursorX < NumOfColumns-1)
    {
      CodeLine[CursorX+1] = LOBYTE(Kanji);
      AttrLine[CursorX+1] = CharAttr;
      AttrLine2[CursorX+1] = CharAttr2;
    }

    if (StrChangeCount==0) StrChangeStart = CursorX;
    StrChangeCount = StrChangeCount + 2;
  }

  if (CursorX < NumOfColumns-2)
  {
    MoveRight();
    MoveRight();
  }
  else {
    UpdateStr();
    Wrap = AutoWrapMode;
  }
}

void DispAllChar(BYTE b)
{
  InsertMode = FALSE;
  AutoWrapMode = TRUE;

  if ((b & 0x80) == 0x80)
  {
    UpdateStr();
    CharAttr = AttrReverse;
    SetupDC();
    b = b & 0x7f;
  }

  if (b<=US)
  {
    DispChar('^');
    DispChar(b+0x40);
  }
  else if (b==DEL)
  {
    DispChar('<');
    DispChar('D');
    DispChar('E');
    DispChar('L');
    DispChar('>');
  }
  else
    DispChar(b);

  if (CharAttr != AttrDefault)
  {
    UpdateStr();
    CharAttr = AttrDefault;
    SetupDC();
  }
}

void ClearBuffer()
{
  /* Reset buffer */
  ResetBuff();

  /* Home position */
  CursorX = 0;
  CursorY = 0;
  WinOrgX = 0;
  WinOrgY = 0;
  NewOrgX = 0;
  NewOrgY = 0;

  /* Top/bottom mergin */
  CursorTop = 0;
  CursorBottom = NumOfLines-1;

  InvalidateRect(HVTWin,NULL,FALSE);

  ScrollCount = 0;
  dScroll = 0;
  StrChangeCount = 0;
  if (WinHeight > NumOfLines)
    ChangeWindowSize(NumOfColumns,NumOfLines);
  else {
    if ((NumOfLines==WinHeight) && (ts.EnableScrollBuff>0))
    {
      SetScrollRange(HVTWin,SB_VERT,0,1,FALSE);
    }
    else 
      SetScrollRange(HVTWin,SB_VERT,0,NumOfLines-WinHeight,FALSE);

    SetScrollPos(HVTWin,SB_HORZ,0,TRUE);
    SetScrollPos(HVTWin,SB_VERT,0,TRUE);
  }

  LockBuffer(); /* for safety */
}

void ChangeSelectRegion()
{
  POINT TempStart, TempEnd;
  int j, IStart, IEnd;
  HDC TempDC;

  if ((SelectEndOld.x==SelectEnd.x) &&
      (SelectEndOld.y==SelectEnd.y)) return;

  if ((SelectEndOld.y < SelectEnd.y) ||
      (SelectEndOld.y==SelectEnd.y) &&
      (SelectEndOld.x<=SelectEnd.x))
  {
    TempStart = SelectEndOld;
    TempEnd.x = SelectEnd.x-1;
    TempEnd.y = SelectEnd.y;
  }
  else {
    TempStart = SelectEnd;
    TempEnd.x = SelectEndOld.x-1;
    TempEnd.y = SelectEndOld.y;
  }
  if (TempEnd.x < 0)
  {
    TempEnd.x = NumOfColumns - 1;
    TempEnd.y--;
  }

  for (j = TempStart.y ; j <= TempEnd.y ; j++)
  {
    IStart = 0;
    IEnd = NumOfColumns-1;
    if (j==TempStart.y) IStart = TempStart.x;
    if (j==TempEnd.y) IEnd = TempEnd.x;

    if ((IEnd>=IStart) && (j >= PageStart+WinOrgY) &&
        (j < PageStart+WinOrgY+WinHeight))
    {
      TempDC = GetDC(HVTWin);
      UpdateRect(TempDC,IStart,j-PageStart,IEnd,j-PageStart);
      ReleaseDC(HVTWin,TempDC);
    }
  }

  SelectEndOld = SelectEnd;

}

void ChangeWin()
{
  int i, j;
  WORD Ny;

  /* Change window caption */
  ChangeTitle();

  /* Change buffer */
  if (ts.EnableScrollBuff>0)
  {
    if (ts.ScrollBuffMax < NumOfLines)
      ts.ScrollBuffMax = NumOfLines;
    Ny = ts.ScrollBuffMax;
  }
  else
    Ny = NumOfLines;

  if (NumOfLinesInBuff!=Ny)
  {
    ChangeBuffer(NumOfColumns,Ny);
    if (ts.EnableScrollBuff>0)
      ts.ScrollBuffMax = NumOfLinesInBuff;

    if (BuffEnd < WinHeight)
      ChangeWindowSize(WinWidth,BuffEnd);
    else if ((BuffEnd==WinHeight) && (ts.EnableScrollBuff>0))
    {
      i = WinWidth;
      j = WinHeight;
      SetScrollRange(HVTWin,SB_VERT,0,1,TRUE);
      ChangeWindowSize(i,j);
    }
    else {
      if (GetScrollPos(HVTWin,SB_VERT) > BuffEnd-WinHeight)
        SetScrollPos(HVTWin,SB_VERT,BuffEnd-WinHeight,FALSE);
      SetScrollRange(HVTWin,SB_VERT,0,BuffEnd-WinHeight,TRUE);
    }
  }

  /* Menu bar / Popup menu */
  if (ts.HideTitle==0)
    SwitchMenu();

  /* 3D dialog box */
/*  if (ts.Dlg3d>0)
    LoadCtl3d();
  else
    FreeCtl3d(); */

  if (Background != NULL) DeleteObject(Background);
  Background = CreateSolidBrush(ts.VTColor[1]);

  InvalidateRect(HVTWin,NULL,TRUE);

  /* Change caret shape */
  ChangeCaret();
}

