/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, VT terminal emulation */
#include "teraterm.h"
#include "tttypes.h"
#include <stdio.h>
#include "buffer.h"
#include "ttwinman.h"
#include "ttcommon.h"
#include "commlib.h"
#include "vtdisp.h"
#include "scroll.h"
#include "keyboard.h"
#include "ttlib.h"
#include "ttftypes.h"
#include "filesys.h"
#include "teraprn.h"

  /* Parsing modes */
#define ModeFirst 0
#define ModeESC   1
#define ModeDCS   2
#define ModeDCUserKey 3
#define ModeSOS   4
#define ModeCSI   5
#define ModeXS    6
#define ModeDLE   7
#define ModeCAN   8

#define NParamMax 16
#define IntCharMax 5


/* prototypes */
void ParseControl(BYTE b);


static POINT SavedCursor;

static BYTE SavedAttr, SavedAttr2;
    /* various modes for VT emulation */
static BOOL RelativeOrgMode;
static BOOL ReverseColor;

static BOOL ESCFlag, JustAfterESC;
static BOOL KanjiIn;
static BOOL EUCkanaIn, EUCsupIn;
static int  EUCcount;

static int Param[NParamMax+1];
static int NParam;
static BOOL FirstPrm;
static BYTE IntChar[IntCharMax+1];
static int ICount;
static BYTE Prv;
static int ParseMode, SavedMode;
static int ChangeEmu;

/* select parser flag ( ^[%!0 ) */
static BOOL SelectParserFlag;

/* user defined keys */
static BOOL WaitKeyId, WaitHi;


void ResetTerminal() /*reset variables but don't update screen */
{
  ResetDisp();

  /* Cursor */
  NewLine(PageStart);
  WinOrgX = 0;
  WinOrgY = 0;
  NewOrgX = 0;
  NewOrgY = 0;
  SavedCursor.x = 0;
  SavedCursor.y = 0;

  /* Initialize text selection region */
  SelectStart.x = 0;
  SelectStart.y = 0;
  SelectEnd = SelectStart;
  SelectEndOld = SelectStart;
  Select = FALSE;

  /* Various modes */
  AppliKeyMode = FALSE;
  AppliCursorMode = FALSE;
  RelativeOrgMode = FALSE;
  ReverseColor = FALSE;
  AutoRepeatMode = TRUE;

  /* Character sets */
  ResetCharSet();

  /* ESC flag for device control sequence */
  ESCFlag = FALSE;
  /* for TEK sequence */
  JustAfterESC = FALSE;

  /* Parse mode */
  ParseMode = ModeFirst;

  /* Clear printer mode */
  PrinterMode = FALSE;

  /* IME */
  OpenConv = FALSE;
}

void ResetCharSet()
{
  if (ts.Language==IdJapanese)
  {
    Gn[0] = IdASCII;
    Gn[1] = IdKatakana;
    Gn[2] = IdKatakana;
    Gn[3] = IdKanji;
    Glr[0] = 0;
    Glr[1] = 3;
  }
  else {
    Gn[0] = IdASCII;
    Gn[1] = IdSpecial;
    Gn[2] = IdASCII;
    Gn[3] = IdSpecial;
    Glr[0] = 0;
    Glr[1] = 0;
    cv.SendCode = IdASCII;
    cv.SendKanjiFlag = FALSE;
    cv.EchoCode = IdASCII;
    cv.EchoKanjiFlag = FALSE;
  }
  /* Kanji flag */
  KanjiIn = FALSE;
  EUCkanaIn = FALSE;
  EUCsupIn = FALSE;
  SSflag = FALSE;

  cv.Language = ts.Language;
  cv.CRSend = ts.CRSend;
  cv.KanjiCodeEcho = ts.KanjiCode;
  cv.JIS7KatakanaEcho = ts.JIS7Katakana;
  cv.KanjiCodeSend = ts.KanjiCodeSend;
  cv.JIS7KatakanaSend = ts.JIS7KatakanaSend;
  cv.KanjiIn = ts.KanjiIn;
  cv.KanjiOut = ts.KanjiOut;
}

/* called by the [Setup] Terminal command */
void SetupTerm()
{
  if (ts.Language==IdJapanese)
    ResetCharSet();

  if ((ts.TerminalWidth!=NumOfColumns) ||
      (ts.TerminalHeight!=NumOfLines))
  {
    LockBuffer();
    ChangeTerminalSize(ts.TerminalWidth,
                       ts.TerminalHeight);
    UnlockBuffer();
    CaretOn();
  }
  else if ((ts.TermIsWin>0) &&
           ((ts.TerminalWidth!=WinWidth) ||
            (ts.TerminalHeight!=WinHeight)))
    ChangeWindowSize(ts.TerminalWidth,ts.TerminalHeight);
}

void XSequence(BYTE b)
{
  if (NParam==1)
  {
    if ((b>=0x30) && (b<=0x39))
    {
      if (Param[1]<1000)
        Param[1] = Param[1]*10 + b - 0x30;
    }
    else
      NParam = 2;
  }
  else {
    if (b<US) {
      if (Param[1]<=2)
      {
        ts.Title[NParam-2] = 0;
        ChangeTitle();
      }
      ParseMode = ModeFirst;
    }
    else {
      if ((Param[1]<=2) &&
	  (NParam-2<sizeof(ts.Title)-1))
      {
        ts.Title[NParam-2] = b;
        NParam++;
      }
    }
  }
}

void DLESeen(BYTE b)
{
  ParseMode = ModeFirst;
  if (ts.BPAuto && (b=='B'))
    BPStart(IdBPAuto); /* Auto B-Plus activation */
  ChangeEmu = -1;
}

void CANSeen(BYTE b)
{
  ParseMode = ModeFirst;
  if (ts.ZmodemAuto && (b=='B'))
    ZMODEMStart(IdZAuto); /* Auto ZMODEM activation */
  ChangeEmu = -1;
}

  void CSInsertCharacter()
  {
    int Count;

    UpdateStr();
    UpdateScroll();
    if (Param[1]<1) Param[1] = 1;
    Count = Param[1];

    if (ts.Language==IdJapanese)
      EraseKanji(1); /* if cursor is on right half of a kanji, erase the kanji */

    if (Count > NumOfColumns - CursorX)
      Count = NumOfColumns - CursorX;
    memmove(&(CodeLine[CursorX+Count]),&(CodeLine[CursorX]),
            NumOfColumns-Count-CursorX);
    memmove(&(AttrLine[CursorX+Count]),&(AttrLine[CursorX]),
            NumOfColumns-Count-CursorX);
    memmove(&(AttrLine2[CursorX+Count]),&(AttrLine2[CursorX]),
            NumOfColumns-Count-CursorX);
    memset(&(CodeLine[CursorX]),0x20,Count);
    memset(&(AttrLine[CursorX]),AttrDefault,Count);
    memset(&(AttrLine2[CursorX]),AttrDefault,Count);
    /* last char in current line is kanji first? */
    if ((AttrLine[NumOfColumns-1] & AttrKanji) != 0)
    {
      /* then delete it */
      CodeLine[NumOfColumns-1] = 0x20;
      AttrLine[NumOfColumns-1] = AttrDefault;
      AttrLine2[NumOfColumns-1] = AttrDefault2;
    }
    UpdateRect(VTDC,CursorX,CursorY,NumOfColumns-1,CursorY);
  }

  void CSCursorUp()
  {
    if (Param[1]<1) Param[1] = 1;

    if (CursorY >= CursorTop)
    {
      if (CursorY-Param[1] > CursorTop)
        MoveCursor(CursorX,CursorY-Param[1]);
      else
        MoveCursor(CursorX,CursorTop);
    }
    else {
      if (CursorY > 0)
        MoveCursor(CursorX,CursorY-Param[1]);
      else
        MoveCursor(CursorX,0);
    }
  }

  void CSCursorUp1()
  {
    MoveCursor(0,CursorY);
    CSCursorUp();
  }

  void CSCursorDown()
  {
    if (Param[1]<1) Param[1] = 1;

    if (CursorY <= CursorBottom)
    {
      if (CursorY+Param[1] < CursorBottom)
        MoveCursor(CursorX,CursorY+Param[1]);
      else
        MoveCursor(CursorX,CursorBottom);
    }
    else {
      if (CursorY < NumOfLines-1)
        MoveCursor(CursorX,CursorY+Param[1]);
      else
        MoveCursor(CursorX,NumOfLines);
    }
  }

  void CSCursorDown1()
  {
    MoveCursor(0,CursorY);
    CSCursorDown();
  }

  void CSScreenErase()
  {
    LONG TmpPtr;
    int offset;
    int i;
    RECT R;

    if (Param[1] == -1) Param[1] = 0;
    UpdateStr();
    UpdateScroll();
    R.left = 0;
    R.right = ScreenWidth;
    switch (Param[1]) {
      case 0:
        if (ts.Language==IdJapanese)
          EraseKanji(1); /* if cursor is on right half of a kanji, erase the kanji */
        offset = CursorX;
        TmpPtr = GetLinePtr(PageStart+CursorY);
        for (i = CursorY ; i <= NumOfLines-1 ; i++)
        {
          memset(&(CodeBuff[TmpPtr+offset]),0x20,NumOfColumns-offset);
          memset(&(AttrBuff[TmpPtr+offset]),AttrDefault,NumOfColumns-offset);
          memset(&(AttrBuff2[TmpPtr+offset]),AttrDefault2,NumOfColumns-offset);
          offset = 0;
          TmpPtr = NextLinePtr(TmpPtr);
        }
        R.top = (CursorY+1-WinOrgY)*FontHeight;
        R.bottom = ScreenHeight;
        FillRect(VTDC,&R,Background);
        R.left = (CursorX-WinOrgX)*FontWidth;
        R.bottom = R.top;
        R.top = R.bottom-FontHeight;
        FillRect(VTDC,&R,Background);
        break;
      case 1:
        if (ts.Language==IdJapanese)
          EraseKanji(0); /* if cursor is on left half of a kanji, erase the kanji */
        offset = NumOfColumns;
        TmpPtr = GetLinePtr(PageStart);
        for (i = 0 ; i <= CursorY ; i++)
        {
	  if (i==CursorY) offset = CursorX+1;
          memset(&(CodeBuff[TmpPtr]),0x20,offset);
          memset(&(AttrBuff[TmpPtr]),AttrDefault,offset);
          memset(&(AttrBuff2[TmpPtr]),AttrDefault2,offset);
          TmpPtr = NextLinePtr(TmpPtr);
        }
        R.top = -WinOrgY*FontHeight;
        R.bottom = (CursorY-WinOrgY)*FontHeight;
        FillRect(VTDC,&R,Background);
        R.top = R.bottom;
        R.bottom = R.top + FontHeight;
        R.right = (CursorX+1-WinOrgX)*FontWidth;
        FillRect(VTDC,&R,Background);
        break;
      case 2:
        BuffScroll(NumOfLines);
        WinOrgY = WinOrgY - NumOfLines;
        NewOrgY = 0;
        UpdateScroll();
        NewLine(PageStart+CursorY);

        UpdateWindow(HVTWin);

        break;
    }
  }

  void CSInsertLine()
  {
    int i, Count, YEnd;
    LONG SrcPtr, DestPtr;
    RECT R;

    if (CursorY < CursorTop) return;
    if (CursorY > CursorBottom) return;
    if (Param[1]<1) Param[1] = 1;
    UpdateStr();
    UpdateScroll();
    Count = Param[1];
    YEnd = CursorBottom;
    if (CursorY > YEnd) YEnd = NumOfLines-1;
    if (Count > YEnd+1 - CursorY) Count = YEnd+1 - CursorY;
    SrcPtr = GetLinePtr(PageStart+YEnd-Count);
    DestPtr = GetLinePtr(PageStart+YEnd);
    for (i= YEnd-Count ; i>=CursorY ; i--)
    {
      memmove(&(CodeBuff[DestPtr]),&(CodeBuff[SrcPtr]),NumOfColumns);
      memmove(&(AttrBuff[DestPtr]),&(AttrBuff[SrcPtr]),NumOfColumns);
      memmove(&(AttrBuff2[DestPtr]),&(AttrBuff2[SrcPtr]),NumOfColumns);
      SrcPtr = PrevLinePtr(SrcPtr);
      DestPtr = PrevLinePtr(DestPtr);
    }
    for (i = 1 ; i <= Count ; i++)
    {
      memset(&(CodeBuff[DestPtr]),0x20,NumOfColumns);
      memset(&(AttrBuff[DestPtr]),AttrDefault,NumOfColumns);
      memset(&(AttrBuff2[DestPtr]),AttrDefault2,NumOfColumns);
      DestPtr = PrevLinePtr(DestPtr);
    }

    if (Active && CompletelyVisible &&
        (CursorY >= WinOrgY))
    {
      R.left = 0;
      R.right = ScreenWidth;
      R.top = (CursorY-WinOrgY)*FontHeight;
      R.bottom = (YEnd+1-WinOrgY)*FontHeight;
      ScrollWindow(HVTWin,0,FontHeight*Count,&R,&R);
    }
    else
      UpdateRect(VTDC,WinOrgX,CursorY,WinOrgX+WinWidth-1,YEnd);
  }

  void CSLineErase()
  {
    RECT R;

    if (Param[1] == -1) Param[1] = 0;
    UpdateStr();
    UpdateScroll();
    R.top = (CursorY-WinOrgY)*FontHeight;
    R.bottom = R.top+FontHeight;
    switch (Param[1]) {
      /* erase char from cursor to end of line */
      case 0:
        memset(&(CodeLine[CursorX]),0x20,NumOfColumns-CursorX);
        memset(&(AttrLine[CursorX]),AttrDefault,NumOfColumns-CursorX);
        memset(&(AttrLine2[CursorX]),AttrDefault2,NumOfColumns-CursorX);

        R.left = (CursorX-WinOrgX)*FontWidth;
        R.right = ScreenWidth;
        FillRect(VTDC,&R,Background);
        break;
      /* erase char from start of line to cursor */
      case 1:
        memset(CodeLine,0x20,CursorX+1);
        memset(AttrLine,AttrDefault,CursorX+1);
        memset(AttrLine2,AttrDefault2,CursorX+1);
        R.left = 0;
        R.right = (CursorX+1-WinOrgX)*FontWidth;
        FillRect(VTDC,&R,Background);
        break;
      /* erase entire line */
      case 2:
        memset(CodeLine,0x20,NumOfColumns);
        memset(AttrBuff,AttrDefault,NumOfColumns);
        memset(AttrBuff2,AttrDefault2,NumOfColumns);
        R.left = 0;
        R.right = ScreenWidth;
        FillRect(VTDC,&R,Background);
        break;
    }
  }

  void CSDeleteNLines()
  {
    int i, YEnd;
    LONG SrcPtr, DestPtr;
    RECT R;

    if (CursorY < CursorTop) return;
    if (CursorY > CursorBottom) return;
    if (Param[1]<1) Param[1] = 1;
    UpdateStr();
    UpdateScroll();

    YEnd = CursorBottom;
    if (CursorY > YEnd) YEnd = NumOfLines-1;

    if (Param[1] > YEnd+1-CursorY) Param[1] = YEnd+1-CursorY;
    SrcPtr = GetLinePtr(PageStart+CursorY+Param[1]);
    DestPtr = GetLinePtr(PageStart+CursorY);
    for (i=CursorY ; i<= YEnd-Param[1] ; i++)
    {
      memmove(&(CodeBuff[DestPtr]),&(CodeBuff[SrcPtr]),NumOfColumns);
      memmove(&(AttrBuff[DestPtr]),&(AttrBuff[SrcPtr]),NumOfColumns);
      memmove(&(AttrBuff2[DestPtr]),&(AttrBuff2[SrcPtr]),NumOfColumns);
      SrcPtr = NextLinePtr(SrcPtr);
      DestPtr = NextLinePtr(DestPtr);
    }
    for (i = YEnd+1-Param[1] ; i<=YEnd ; i++)
    {
      memset(&(CodeBuff[DestPtr]),0x20,NumOfColumns);
      memset(&(AttrBuff[DestPtr]),AttrDefault,NumOfColumns);
      memset(&(AttrBuff2[DestPtr]),AttrDefault2,NumOfColumns);
      DestPtr = NextLinePtr(DestPtr);
    }
   
    if (Active && CompletelyVisible &&
        (YEnd+1-WinOrgY <= WinHeight))
    {
      R.left = 0;
      R.right = ScreenWidth;
      R.top = (CursorY-WinOrgY)*FontHeight;
      R.bottom = (YEnd+1-WinOrgY)*FontHeight;
      ScrollWindow(HVTWin,0,-FontHeight*Param[1],&R,&R);
      UpdateWindow(HVTWin);
    }
    else
      UpdateRect(VTDC,WinOrgX,CursorY,WinOrgX+WinWidth-1,YEnd);

  }

  void CSDeleteCharacter()
  {
    int Count;

    if (Param[1]<1) Param[1] = 1;
    UpdateStr();
    UpdateScroll();
    Count = Param[1];

    if (ts.Language==IdJapanese)
    {
      EraseKanji(0); /* if cursor is on left harf of a kanji, erase the kanji */
      EraseKanji(1); /* if cursor on right half... */
    }

    if (Count > NumOfColumns-CursorX) Count = NumOfColumns-CursorX;
    memmove(&(CodeLine[CursorX]),&(CodeLine[CursorX+Count]),
            NumOfColumns-Count-CursorX);
    memmove(&(AttrLine[CursorX]),&(AttrLine[CursorX+Count]),
            NumOfColumns-Count-CursorX);
    memmove(&(AttrLine2[CursorX]),&(AttrLine2[CursorX+Count]),
            NumOfColumns-Count-CursorX);
    memset(&(CodeLine[NumOfColumns-Count]),0x20,Count);
    memset(&(AttrLine[NumOfColumns-Count]),AttrDefault,Count);
    memset(&(AttrLine2[NumOfColumns-Count]),AttrDefault2,Count);

    UpdateRect(VTDC,CursorX,CursorY,WinOrgX+WinWidth-1,CursorY);
  }

  void CSEraseCharacter()
  {
    int Count;
    RECT R;

    if (Param[1]<1) Param[1] = 1;
    UpdateStr();
    UpdateScroll();
    Count = Param[1];

    if (ts.Language==IdJapanese)
    {
      EraseKanji(0); /* if cursor is on left harf of a kanji, erase the kanji */
      EraseKanji(1); /* if cursor on right half... */
    }

    if (Count > NumOfColumns-CursorX) Count = NumOfColumns-CursorX;
    memset(&(CodeLine[CursorX]),0x20,Count);
    memset(&(AttrLine[CursorX]),AttrDefault,Count);
    memset(&(AttrLine2[CursorX]),AttrDefault2,Count);

    R.left = (CursorX-WinOrgX)*FontWidth;
    R.right = R.left + Count*FontWidth;
    R.top = (CursorY-WinOrgY)*FontHeight;
    R.bottom = R.top + FontHeight;
    FillRect(VTDC,&R,Background);
  }

  void CSMoveToColumnN()
  {
    if (Param[1]<1) Param[1] = 1;
    Param[1]--;
    if (Param[1] < 0) Param[1] = 0;
    if (Param[1] > NumOfColumns-1) Param[1] = NumOfColumns-1;
    MoveCursor(Param[1],CursorY);
  }

  void CSCursorRight()
  {
    if (Param[1]<1) Param[1] = 1;
    if (CursorX + Param[1] > NumOfColumns-1)
      MoveCursor(NumOfColumns-1,CursorY);
    else
      MoveCursor(CursorX+Param[1],CursorY);
  }

  void CSCursorLeft()
  {
    if (Param[1]<1) Param[1] = 1;
    if (CursorX-Param[1] < 0)
      MoveCursor(0,CursorY);
    else
      MoveCursor(CursorX-Param[1],CursorY);
  }

  void CSMoveToLineN()
  {
    if (Param[1]<1) Param[1] = 1;
    if (RelativeOrgMode)
    {
      if (CursorTop+Param[1]-1 > CursorBottom)
        MoveCursor(CursorX,CursorBottom);
      else
        MoveCursor(CursorX,CursorTop+Param[1]-1);
    }
    else {
      if (Param[1] > NumOfLines)
        MoveCursor(CursorX,NumOfLines-1);
      else
        MoveCursor(CursorX,Param[1]-1);
    }
  }

  void CSMoveToXY()
  {
    int NewX, NewY;

    if (Param[1]<1) Param[1] = 1;
    if ((NParam < 2) || (Param[2]<1)) Param[2] = 1;
    NewX = Param[2] - 1;
    if (NewX > NumOfColumns-1) NewX = NumOfColumns-1;
    if (RelativeOrgMode)
    {
      NewY = CursorTop + Param[1] - 1;
      if (NewY > CursorBottom) NewY = CursorBottom;
    }
    else {
      NewY = Param[1] - 1;
      if (NewY > NumOfLines-1) NewY = NumOfLines-1;
    }
    MoveCursor(NewX,NewY);
  }

  void CSDeleteTabStop()
  {
    int i,j;

    if (Param[1]==-1) Param[1] = 0;
    if (NTabStops>0)
      switch (Param[1]) {
        case 0:
          i = 0;
          while ((TabStops[i]!=CursorX) && (i<NTabStops-1))
            i++;
          if (TabStops[i] == CursorX)
          {
            NTabStops--;
            for (j=i ; j<=NTabStops ; j++)
              TabStops[j] = TabStops[j+1];
          }
          break;
        case 3: NTabStops = 0; break;
      }
  }

  void CS_h_Mode()
  {
    switch (Param[1]) {
      case 4: InsertMode = TRUE; break;
      case 20:
        LFMode = TRUE;
        ts.CRSend = IdCRLF;
        cv.CRSend = IdCRLF;
        break;
      case 38:
        if (ts.AutoWinSwitch>0)
          ChangeEmu = IdTEK; /* Enter TEK Mode */
        break;
      case 59:
	if (ts.Language==IdJapanese)
        { /* kanji terminal */
          Gn[0] = IdASCII;
          Gn[1] = IdKatakana;
          Gn[2] = IdKatakana;
          Gn[3] = IdKanji;
          Glr[0] = 0;
          Glr[1] = 3;
        }
	break;
    }
  }

  void CS_i_Mode()
  {
    if (Param[1]==-1) Param[1] = 0;
    switch (Param[1]) {
      /* print screen */
      case 0: VTPrintInit(TRUE); break;
      /* printer controller mode off */
      case 4: break; /* See 'teraprn.cpp' */
      /* printer controller mode on */
      case 5:
        if (! AutoPrintMode)
          OpenPrnFile();
        PrinterMode = TRUE;
        PrnParseMode = IdPrnNormal;
        break;
    }
  }

  void CS_l_Mode()
  {
    switch (Param[1]) {
      case 4: InsertMode = FALSE; break;
      case 20:
        LFMode = FALSE;
        ts.CRSend = IdCR;
        cv.CRSend = IdCR;
        break;
      case 59:
	if (ts.Language==IdJapanese)
        { /* katakana terminal */
          Gn[0] = IdASCII;
          Gn[1] = IdKatakana;
          Gn[2] = IdKatakana;
          Gn[3] = IdKanji;
          Glr[0] = 0;
          Glr[1] = 3;
        }
	break;
    }
  }

  void CS_n_Mode()
  {
    char Report[16];

    switch (Param[1]) {
      case 5:
	CommBinaryOut(&cv,"\033[0n",4); /* Device Status Report -> Ready */
	break;
      case 6:
        /* Cursor Position Report */
	sprintf(Report,"\033[%u;%uR",CursorY+1,CursorX+1);
        CommBinaryOut(&cv,Report,strlen(Report));
        break;
    }
  }

  void CSSetAttr()
  {
    int i, P;

    UpdateStr();
    for (i=1 ; i<=NParam ; i++)
    {
      P = Param[i];
      if (P<0) P = 0;
      switch (P) {
        /* Clear */
        case 0:
          CharAttr = AttrDefault;
          CharAttr2 = AttrDefault2;
          break;
        /* Bold */
        case 1:
	  CharAttr = CharAttr | AttrBold;
	  break;
        /* Under line */
        case 4:
	  CharAttr = CharAttr | AttrUnder;
	  break;
        /* Blink */
        case 5:
	  CharAttr = CharAttr | AttrBlink;
	  break;
        /* Reverse */
        case 7:
	  CharAttr = CharAttr | AttrReverse;
	  break;
        /* Bold off */
        case 22:
	  CharAttr = CharAttr & ~ AttrBold;
	  break;
        /* Under line off */
        case 24:
	  CharAttr = CharAttr & ~ AttrUnder;
	  break;
        /* Blink off */
        case 25:
	  CharAttr = CharAttr & ~ AttrBlink;
	  break;
        /* Reverse off */
        case 27:
	  CharAttr = CharAttr & ~ AttrReverse;
	  break;
	default:
          /* Text color */
          if ((P>=30) && (P<=37))
	    CharAttr2 = CharAttr2 & (Attr2Back | Attr2BackMask)
                        | (P-30) | Attr2Fore;
	  else if ((P>=40) && (P<=47)) /* Back color */
            CharAttr2 = CharAttr2 & (Attr2Fore | Attr2ForeMask)
                        | ((P-40) << SftAttrBack) | Attr2Back;
	  else if (P==100) /* Reset color attributes */
            CharAttr2 = AttrDefault2;
      }
    }
    SetupDC();
  }

  void CSSetScrollRegion()
  {
    if (Param[1]<1) Param[1] =1;    
    if ((NParam < 2) | (Param[2]<1))
      Param[2] = NumOfLines;
    Param[1]--;
    Param[2]--;
    if (Param[1] > NumOfLines-1)
      Param[1] = NumOfLines-1;
    if (Param[2] > NumOfLines-1)
      Param[2] = NumOfLines-1;
    if (Param[1] >= Param[2]) return;
    CursorTop = Param[1];
    CursorBottom = Param[2];
    if (RelativeOrgMode) MoveCursor(0,CursorTop);
                    else MoveCursor(0,0);
  }

  void CSSunSequence() /* Sun terminal private sequences */
  {
    char Report[16];

    switch (Param[1]) {
      case 8: /* set terminal size */
        if ((Param[2]<=1) || (NParam<2)) Param[2] = 24;
        if ((Param[3]<=1) || (NParam<3)) Param[3] = 80;
        if ((Param[2]==NumOfLines) && (Param[3]==NumOfColumns))
          return;
        UpdateStr();
        UpdateScroll();
        ChangeTerminalSize(Param[3],Param[2]);
        NewLine(PageStart);
        break;
      case 14: /* get window size??? */
        /* Of cource, this is not actual window size */
        CommBinaryOut(&cv,"\033[4;640;480t",12);
        break;
      case 18: /* get terminal size */
        /* Cursor Position Report */
	sprintf(Report,"\033[8;%u;%u;t",NumOfLines,NumOfColumns);
        CommBinaryOut(&cv,Report,strlen(Report));
        break;
    }
  }

  void CSGT(BYTE b)
  {
    switch (b) {
      case 'c': /* second terminal report */
	CommBinaryOut(&cv,"\033[>32;10;2c",11); /* VT382 */
	break;
    }
  }

    void CSQExchangeColor()
    {
      COLORREF ColorRef;

      UpdateStr();
      UpdateScroll();

      ColorRef = ts.VTColor[0];
      ts.VTColor[0] = ts.VTColor[1];
      ts.VTColor[1] = ColorRef;
      ReleaseDC(HVTWin,VTDC);
      DeleteObject(Background);
      Background = CreateSolidBrush(ts.VTColor[1]);
      VTDC = GetDC(HVTWin);
      SetupDC();
      UpdateRect(VTDC,WinOrgX,WinOrgY,WinOrgX+WinWidth-1,WinOrgY+WinHeight-1);
    }

    void CSQ_h_Mode()
    {
      int i;

      for (i = 1 ; i<=NParam ; i++)
        switch (Param[i]) {
          case 1: AppliCursorMode = TRUE; break;
          case 3:
	    if (NumOfColumns < 132)
            {
              UpdateStr();
	      UpdateScroll();
              ChangeTerminalSize(132,NumOfLines);
              NewLine(PageStart);
            }
	    break;
          case 5:
            if (ReverseColor) return;
              ReverseColor = TRUE;
              /* Exchange text/back color */
            CSQExchangeColor();
	    break;
          case 6:
            RelativeOrgMode = TRUE;
            MoveCursor(0,CursorTop);
            break;
          case 7: AutoWrapMode = TRUE; break;
          case 8: AutoRepeatMode = TRUE; break;
          case 19: PrintEX = TRUE; break;
        }
    }

    void CSQ_i_Mode()
    {
      if (Param[1]==-1) Param[1] = 0;
      switch (Param[1]) {
        case 1:
          if (HPrnFile<=0)
            OpenPrnFile();
          PrintCurrentLine(LF);
          if (! AutoPrintMode)
            ClosePrnFile();
          break;
        /* auto print mode off */
        case 4:
	  if (AutoPrintMode)
          {
            ClosePrnFile();
            AutoPrintMode = FALSE;
          }
	  break;
        /* auto print mode on */
        case 5:
	  if (! AutoPrintMode)
          {
            OpenPrnFile();
            AutoPrintMode = TRUE;
          }
	  break;
      }
    }

    void CSQ_l_Mode()
    {
      int i;

      for (i = 1 ; i <= NParam ; i++)
        switch (Param[i]) {
          case 1: AppliCursorMode = FALSE; break;
          case 3:
            UpdateStr();
	    UpdateScroll();
            ChangeTerminalSize(80,NumOfLines);
            NewLine(PageStart);
            break;
          case 5:
            if (! ReverseColor) return;
            ReverseColor = FALSE;
            /* Exchange text/back color */
            CSQExchangeColor();
            break;
          case 6:
            RelativeOrgMode = FALSE;
            MoveCursor(0,0);
            break;
          case 7: AutoWrapMode = FALSE; break;
          case 8: AutoRepeatMode = FALSE; break;
          case 19: PrintEX = FALSE; break;
        }
    }

    void CSQ_n_Mode()
    {
    }

  void CSQuest(BYTE b)
  {
    switch (b) {
      case 'h': CSQ_h_Mode(); break;
      case 'i': CSQ_i_Mode(); break;
      case 'l': CSQ_l_Mode(); break;
      case 'n': CSQ_n_Mode(); break;
    }
  }

  void CSExc(BYTE b)
  {
    switch (b) {
      case 'p':
        /* Software reset */
        UpdateStr();
	UpdateScroll();
        ResetTerminal();
        ScrollHomePos();
        NewLine(PageStart);
        break;
    }
  }

void AnswerTerminalType()
{
  char Tmp[31];

  if (ts.TerminalID<IdVT320)
    strcpy(Tmp,"\033[?");
  else
    strcpy(Tmp,"\233?");

  switch (ts.TerminalID) {
    case IdVT100:
      strcat(Tmp,"1;2");
      break;
    case IdVT100J:
      strcat(Tmp,"5;2");
      break;
    case IdVT101:
      strcat(Tmp,"1;0");
      break;
    case IdVT102:
      strcat(Tmp,"6");
      break;
    case IdVT102J:
      strcat(Tmp,"15");
      break;
    case IdVT220J:
      strcat(Tmp,"62;1;2;5;6;7;8");
      break;
    case IdVT282:
      strcat(Tmp,"62;1;2;4;5;6;7;8;10;11");
      break;
    case IdVT320:
      strcat(Tmp,"63;1;2;6;7;8");
      break;
    case IdVT382:
      strcat(Tmp,"63;1;2;4;5;6;7;8;10;15");
      break;
  }
  strcat(Tmp,"c");

  CommBinaryOut(&cv,Tmp,strlen(Tmp)); /* Report terminal ID */
}

void SaveCursor()
{
  SavedCursor.x = CursorX;
  SavedCursor.y = CursorY;
  SavedAttr = CharAttr;
  SavedAttr2 = CharAttr2;
}

void  RestoreCursor()
{
  UpdateStr();
  if (SavedCursor.x > NumOfColumns-1)
    SavedCursor.x = NumOfColumns-1;
  if (SavedCursor.y > NumOfLines-1)
    SavedCursor.y = NumOfLines-1;
  MoveCursor(SavedCursor.x,SavedCursor.y);
  CharAttr = SavedAttr;
  CharAttr2 = SavedAttr2;
  SetupDC();
}

void ParseCS(BYTE b) /* b is the final char */
{
  switch (ICount) {
    /* no intermediate char */
    case 0:
      switch (Prv) {
        /* no private parameters */
        case 0:
	  switch (b) {
            case '@': CSInsertCharacter(); break;
            case 'A': CSCursorUp(); break;
            case 'B': CSCursorDown(); break;
            case 'C': CSCursorRight(); break;
            case 'D': CSCursorLeft(); break;
            case 'E': CSCursorDown1(); break;
            case 'F': CSCursorUp1(); break;
            case 'G': CSMoveToColumnN(); break;
            case 'H': CSMoveToXY(); break;
            case 'J': CSScreenErase(); break;
            case 'K': CSLineErase(); break;
            case 'L': CSInsertLine(); break;
            case 'M': CSDeleteNLines(); break;
            case 'P': CSDeleteCharacter(); break;
            case 'X': CSEraseCharacter(); break;
            case '`': CSMoveToColumnN(); break;
            case 'a': CSCursorRight(); break;
            case 'c': AnswerTerminalType(); break;
            case 'd': CSMoveToLineN(); break;
            case 'e': CSCursorUp(); break;
            case 'f': CSMoveToXY(); break;
            case 'g': CSDeleteTabStop(); break;
            case 'h': CS_h_Mode(); break;
            case 'i': CS_i_Mode(); break;
            case 'l': CS_l_Mode(); break;
            case 'm': CSSetAttr(); break;
            case 'n': CS_n_Mode(); break;
            case 'r': CSSetScrollRegion(); break;
            case 's': SaveCursor(); break;
            case 't': CSSunSequence(); break;
            case 'u': RestoreCursor(); break;
          } /* of case Prv=0 */
	  break;
        /* private parameter = '>' */
        case '>': CSGT(b); break;
        /* private parameter = '?' */
        case '?': CSQuest(b); break;
      }
      break;
    /* one intermediate char */
    case 1:
      switch (IntChar[1]) {
        /* intermediate char = '!' */
        case '!': CSExc(b); break;
      }
      break;
  } /* of case Icount */

  ParseMode = ModeFirst;
}

void ControlSequence(BYTE b)
{
  if (b<=US)
    ParseControl(b);
  else if ((b>=0x20) && (b<=0x2F))
  {
    if (ICount<IntCharMax) ICount++;
    IntChar[ICount] = b;
  }
  else if ((b>=0x30) && (b<=0x39))
  {
    if (Param[NParam] < 0)
      Param[NParam] = 0; 
    if (Param[NParam]<1000)
      Param[NParam] = Param[NParam]*10 + b - 0x30;
  }
  else if (b==0x3B)
  {
    if (NParam < NParamMax)
    {
      NParam++;
      Param[NParam] = -1;
    }
  }
  else if ((b>=0x3C) && (b<=0x3F))
  {
    if (FirstPrm) Prv = b;
  }
  else if ((b>=0x40) && (b<=0x7E))
    ParseCS(b);
  else if ((b>=0x80) && (b<=0x9F))
    ParseControl(b);

  FirstPrm = FALSE;
}

void DeviceControl(BYTE b)
{
  if (ESCFlag && (b=='\\') || (b==ST))
  {
    ESCFlag = FALSE;
    ParseMode = SavedMode;
    return;
  }

  if (b==ESC)
  {
    ESCFlag = TRUE;
    return;
  }
  else ESCFlag = FALSE;

  if (b<US)
    ParseControl(b);
  else if ((b>=0x30) && (b<=0x39))
  {
    if (Param[NParam] < 0) Param[NParam] = 0; 
    if (Param[NParam]<1000)
      Param[NParam] = Param[NParam]*10 + b - 0x30;
  }
  else if (b==0x3B)
  {
    if (NParam < NParamMax)
    {
      NParam++;
      Param[NParam] = -1;
    }
  }
  else if ((b>=0x40) && (b<=0x7E))
  {
    if (b=='|')
    {
      ParseMode = ModeDCUserKey;
      if (Param[1] < 1) ClearUserKey();
      WaitKeyId = TRUE;
      NewKeyId = 0;
    }
    else ParseMode = ModeSOS;
  }
}

void DCUserKey(BYTE b)
{
  if (ESCFlag && (b=='\\') || (b==ST))
  {
    if (! WaitKeyId) DefineUserKey();
    ESCFlag = FALSE;
    ParseMode = SavedMode;
    return;
  }

  if (b==ESC)
  {
    ESCFlag = TRUE;
    return;
  }
  else ESCFlag = FALSE;

  if (WaitKeyId)
  {
    if ((b>=0x30) && (b<=0x39))
    {
      if (NewKeyId<1000)
        NewKeyId = NewKeyId*10 + b - 0x30;
    }
    else if (b==0x2F)
    {
      WaitKeyId = FALSE;
      WaitHi = TRUE;
      NewKeyLen = 0;
    }
  }
  else {
    if (b==0x3B)
    {
      DefineUserKey();
      WaitKeyId = TRUE;
      NewKeyId = 0;
    }
    else {
      if (NewKeyLen < FuncKeyStrMax)
      {
        if (WaitHi)
        {
          NewKeyStr[NewKeyLen] = ConvHexChar(b) << 4;
          WaitHi = FALSE;
        }
        else {
          NewKeyStr[NewKeyLen] = NewKeyStr[NewKeyLen] +
                                 ConvHexChar(b);
          WaitHi = TRUE;
          NewKeyLen++;
        }
      }
    }
  }
}

void IgnoreString(BYTE b)
{
  if (ESCFlag && (b=='\\') || (b==ST))
    ParseMode = SavedMode;

  if (b==ESC) ESCFlag = TRUE;
         else ESCFlag = FALSE;
}

void ESCSharp(BYTE b)
{
  LONG TmpPtr;
  int i;

  switch (b) {
    case '8':  /* Fill screen with "E" */
      UpdateStr();
      UpdateScroll();
      TmpPtr = GetLinePtr(PageStart);
      for (i = 0 ; i <= NumOfLines-1 ; i++)
      {
        memset(&(CodeBuff[TmpPtr]),'E',NumOfColumns);
        memset(&(AttrBuff[TmpPtr]),AttrDefault,NumOfColumns);
        memset(&(AttrBuff2[TmpPtr]),AttrDefault2,NumOfColumns);
        TmpPtr = NextLinePtr(TmpPtr);
      }
      StrChangeCount = 0;
      UpdateRect(VTDC,WinOrgX,WinOrgY,WinOrgX+WinWidth-1,WinOrgY+WinHeight-1);
      MoveCursor(0,0);
      ParseMode = ModeFirst;
      break;
  }
}

/* select double byte code set */
void ESCDBCSSelect(BYTE b)
{
  int Dist;

  if (ts.Language!=IdJapanese) return;

  switch (ICount) {
    case 1:
      if ((b=='@') || (b=='B'))
      {
        Gn[0] = IdKanji; /* Kanji -> G0 */
        Glr[0] = 0; /* G0->GL */
      }
      break;
    case 2:
      /* Second intermediate char must be
         '(' or ')' or '*' or '+'. */
      Dist = (IntChar[2]-'(') & 3; /* G0 - G3 */
      if ((b=='1') || (b=='3') ||
          (b=='@') || (b=='B'))
      {
        Gn[Dist] = IdKanji; /* Kanji -> G0-3 */
        if (Dist==0) Glr[0] = 0; /* G0->GL */
      }
      break;
  }
}  

void ESCPercent(BYTE b)
{
  switch (b) {
    case '0':
      if (SelectParserFlag && (ts.AutoWinSwitch>0))
        ChangeEmu = IdTEK; /* enter TEK mode */
      break;
  }
}

  /* select single byte code set */
void ESCSBCSSelect(BYTE b)
{
  int Dist;

  /* Intermediate char must be
     '(' or ')' or '*' or '+'.  */
  Dist = (IntChar[1]-'(') & 3; /* G0 - G3 */

  switch (b) {
    case '0': Gn[Dist] = IdSpecial; break;
    case '<': Gn[Dist] = IdASCII; break;
    case '>': Gn[Dist] = IdASCII; break;
    case 'B': Gn[Dist] = IdASCII; break;
    case 'H': Gn[Dist] = IdASCII; break;
    case 'I':
      if (ts.Language==IdJapanese)
        Gn[Dist] = IdKatakana;
      break;
    case 'J': Gn[Dist] = IdASCII; break;
  }

  if (Dist==0) Glr[0] = 0;  /* G0->GL */
}

void ParseEscape(BYTE b) /* b is the final char */
{
  switch (ICount) {
    /* no intermediate char */
    case 0:
      switch (b) {
        case '7': SaveCursor(); break;
        case '8': RestoreCursor(); break;
        case '=': AppliKeyMode = TRUE; break;
        case '>': AppliKeyMode = FALSE; break;
        case 'D': /* IND */
          LineFeed(0);
          break;
        case 'E': /* NEL */
          MoveCursor(0,CursorY);
          LineFeed(0);
          break;
        case 'H': /* HTS */
          SetTabStop();
	  break;
        case 'M': /* RI */
	  CursorUpWithScroll();
	  break;
        case 'N': /* SS2 */
          GLtmp = 2;
          SSflag = TRUE;
          break;
        case 'O': /* SS3 */
          GLtmp = 3;
          SSflag = TRUE;
          break;
        case 'P': /* DCS */
          SavedMode = ParseMode;
          ESCFlag = FALSE;
          NParam = 1;
          Param[1] = -1;
          ParseMode = ModeDCS;
          return;
        case 'X': /* SOS */
          SavedMode = ParseMode;
          ESCFlag = FALSE;
          ParseMode = ModeSOS;
          return;
        case 'Z': AnswerTerminalType(); break;
        case '[': /* CSI */
          ICount = 0;
          FirstPrm = TRUE;
          NParam = 1;
          Param[1] = -1;
          Prv = 0;
          ParseMode = ModeCSI;
          return;
        case '\\': break; /* ST */
        case ']': /* XTERM sequence (OSC) */
          NParam = 1;
          Param[1] = 0;
          ParseMode = ModeXS;
          return;  
        case '^':
        case '_': /* PM, APC */
          SavedMode = ParseMode;
          ESCFlag = FALSE;
          ParseMode = ModeSOS;
          return;
        case 'c': /* Hardware reset */
          ResetTerminal();
          ClearUserKey();
          ClearBuffer();
          if (ts.PortType==IdSerial)
          {
            /* reset serial port */
	    CommResetSerial(&ts,&cv);
          }
          NewLine(PageStart);
          break;
        case 'n': Glr[0] = 2; break; /* LS2 */
        case 'o': Glr[0] = 3; break; /* LS3 */
        case '|': Glr[1] = 3; break; /* LS3R */
        case '}': Glr[1] = 2; break; /* LS2R */
        case '~': Glr[1] = 1; break; /* LS1R */
      } /* end of case Icount=0 */
      break;
    /* one intermediate char */
    case 1:
      switch (IntChar[1]) {
        case '#': ESCSharp(b); break;
        case '$': ESCDBCSSelect(b); break;
        case '%': ESCPercent(b); break;
        case '(':
        case ')':
        case '*':
        case '+':
          ESCSBCSSelect(b);
          break;
      }
      break;
    /* two intermediate char */
    case 2:
      if ((IntChar[1]=='$') &&
          ('('<=IntChar[2]) &&
          (IntChar[2]<='+'))
        ESCDBCSSelect(b);
      break;
  }
  ParseMode = ModeFirst;
}

void EscapeSequence(BYTE b)
{
  if (b<=US)
    ParseControl(b);
  else if ((b>=0x20) && (b<=0x2F))
  {
    if (ICount<IntCharMax) ICount++;
    IntChar[ICount] = b;
  }
  else if ((b>=0x30) && (b<=0x7E))
    ParseEscape(b);
  else if ((b>=0x80) && (b<=0x9F))
    ParseControl(b);

  JustAfterESC = FALSE;
}

void ParseControl(BYTE b)
{
  if (b>=0x80) /* C1 char */
  {
    /* English mode */
    if (ts.Language==IdEnglish)
    {
      if (ts.TerminalID<IdVT220J)
      {
        DispChar(b); /* Disp C1 char in VT100 mode */
        return;
      }
    }
    else { /* Japanese mode */
      if (! ts.Accept8bitCtrl)
        return; /* ignore C1 char */
      /* C1 chars are interpreted as C0 chars in VT100 mode */
      if (ts.TerminalID<IdVT220J)
        b = b & 0x7F;
    }
  }
  switch (b) {
    /* C0 group */
    case ENQ:
      CommBinaryOut(&cv,&ts.Answerback,ts.AnswerbackLen);
      break;
    case BEL: MessageBeep(0); break;
    case BS: BackSpace(); break;
    case HT: Tab(); break;
    case LF:
    case VT: LineFeed(b); break;
    case FF:
      if ((ts.AutoWinSwitch>0) && JustAfterESC)
      {
        CommInsert1Byte(&cv,b);
        CommInsert1Byte(&cv,ESC);
        ChangeEmu = IdTEK;  /* Enter TEK Mode */
      }
      else
        LineFeed(b);
      break;
    case CR:
      CarriageReturn();
      if (ts.CRReceive==IdCRLF)
        CommInsert1Byte(&cv,LF);
      break;
    case SO:
      if ((ts.Language==IdJapanese) &&
          (ts.KanjiCode==IdJIS) &&
          (ts.JIS7Katakana==1))
      Gn[1] = IdKatakana;
      Glr[0] = 1; /* LS1 */
      break;
    case SI: Glr[0] = 0; break; /* LS0 */
    case DLE:
      if (ts.BPAuto)
	ParseMode = ModeDLE; /* Auto B-Plus activation */
      break;
    case CAN:
      if (ts.ZmodemAuto)
        ParseMode = ModeCAN; /* Auto ZMODEM activation */
      else
        ParseMode = ModeFirst;
      break;
    case SUB: ParseMode = ModeFirst; break;
    case ESC:
      ICount = 0;
      JustAfterESC = TRUE;
      ParseMode = ModeESC;
      break;
    case FS:
    case GS:
    case RS:
    case US:
      if (ts.AutoWinSwitch>0)
      {
        CommInsert1Byte(&cv,b);
        ChangeEmu = IdTEK;  /* Enter TEK Mode */
      }
      break;

    /* C1 char */
    case IND: LineFeed(0); break;
    case NEL:
      LineFeed(0);
      CarriageReturn();
      break;
    case HTS: SetTabStop(); break;
    case RI: CursorUpWithScroll(); break;
    case SS2:
      GLtmp = 2;
      SSflag = TRUE;
      break;
    case SS3:
      GLtmp = 3;
      SSflag = TRUE;
      break;
    case DCS:
      SavedMode = ParseMode;
      ESCFlag = FALSE;
      NParam = 1;
      Param[1] = -1;
      ParseMode = ModeDCS;
      break;
    case SOS:
      SavedMode = ParseMode;
      ESCFlag = FALSE;
      ParseMode = ModeSOS;
      break;
    case CSI:
      ICount = 0;
      FirstPrm = TRUE;
      NParam = 1;
      Param[1] = -1;
      Prv = 0;
      ParseMode = ModeCSI;
      break;
    case OSC:
    case PM:
    case APC:
      SavedMode = ParseMode;
      ESCFlag = FALSE;
      ParseMode = ModeSOS;
      break;
  }
}

BOOL CheckKanji(BYTE b)
{
  BOOL Check;

  if (ts.Language!=IdJapanese) return FALSE;
  Check = FALSE;
  if ((b>=0x21) && (b<=0x7e))
    Check = (Gn[Glr[0]]==IdKanji);
  else if ((b>=0xA1) && (b<=0xFE))
    Check = ((ts.KanjiCode == IdEUC) ||
             (ts.KanjiCode == IdJIS) &&
             (ts.JIS7Katakana == 1) &&
             (Gn[Glr[1]]==IdKanji)); 

  if (Check) ConvJIS = TRUE;

  if ((ts.KanjiCode == IdSJIS) &&
      ((0x80<b) && (b<0xa0) || (0xdf<b) && (b<0xfd)))
  {
    Check = TRUE;
    ConvJIS = FALSE;
  }

  return Check;
}

void ParseFirst(BYTE b)
{
  if ((ts.Language==IdJapanese) && KanjiIn)
  {
    if (((! ConvJIS) && (0x3F<b) && (b<0xFD) ||
         ConvJIS && ( (0x20<b) && (b<0x7f) ||
		      (0xa0<b) && (b<0xff) )))
    {
      DispKanji(b);
      KanjiIn = FALSE;
      return;
    }
    else
      if (! ts.CtrlInKanji)
        KanjiIn = FALSE;
  }
        
  if (SSflag)
  {
    if (ts.Language==IdJapanese)
    {
      if (Gn[GLtmp] == IdKanji)
      {
        Kanji = b << 8;
        KanjiIn = TRUE;
        SSflag = FALSE;
        return;
      }
      else if (Gn[GLtmp] == IdKatakana) b = b | 0x80;
    }
    DispChar(b);
    SSflag = FALSE;
    return;
  }

  if ((! EUCsupIn) && (! EUCkanaIn) &&
      (! KanjiIn) && CheckKanji(b))
  {
    Kanji = b << 8;
    KanjiIn = TRUE;
    return;
  }

  if (b<=US)
    ParseControl(b);
  else if (b==0x20)
    DispChar(b);
  else if ((b>=0x21) && (b<=0x7E))
  {
    if (ts.Language==IdJapanese)
    {
      if (EUCsupIn)
      {
        EUCcount--;
        EUCsupIn = (EUCcount==0);
        return;
      }

      if ((Gn[Glr[0]] == IdKatakana) || EUCkanaIn)
      {
        b = b | 0x80;
        EUCkanaIn = FALSE;
      }
    }
    DispChar(b);
  }
  else if ((b>=0x80) && (b<=0x8D))
    ParseControl(b);
  else if (b==0x8E)
  {
    if ((ts.Language==IdJapanese) &&
        (ts.KanjiCode==IdEUC))
      EUCkanaIn = TRUE;
    else
      ParseControl(b);
  }
  else if (b==0x8F)
  {
    if ((ts.Language==IdJapanese) &&
        (ts.KanjiCode==IdEUC))
    {
      EUCcount = 2;
      EUCsupIn = TRUE;
    } else
      ParseControl(b);
  }
  else if ((b>=0x90) && (b<=0x9F))
    ParseControl(b);
  else if (b==0xA0)
    DispChar(0x20);
  else if ((b>=0xA1) && (b<=0xFE))
  {
    if (ts.Language==IdJapanese)
    {
      if (EUCsupIn)
      {
        EUCcount--;
        EUCsupIn = (EUCcount==0);
        return;
      }
      EUCkanaIn = FALSE;

      if (ts.KanjiCode==IdJIS)
      {
        if ((ts.JIS7Katakana==1) &&
            (Gn[Glr[1]] == IdASCII))
          b = b & 0x7f;
      }
    }
    DispChar(b);
  }
}

int VTParse()
{
  BYTE b;
  int c;

  c = CommRead1Byte(&cv,&b);

  if ((c==0) || LButton || MButton || RButton)
    return 0;

  CaretOff();

  ChangeEmu = 0;

  /* Get Device Context */
  VTDC = GetDC(HVTWin);
  while (VTDC==0)
    VTDC = GetDC(HVTWin);

  SetupDC();

  LockBuffer();
  NewLine(PageStart+CursorY);

  while ((c>0) && (ChangeEmu==0))
  {
    if (DebugFlag) DispAllChar(b);
    else if (PrinterMode)
      PrnParse(b);
    else {
      if (! BuffLock) /* for safety */
      {
        LockBuffer();
        NewLine(PageStart+CursorY);
      }
      switch (ParseMode) {
        case ModeFirst: ParseFirst(b); break;
        case ModeESC: EscapeSequence(b); break;
        case ModeDCS: DeviceControl(b); break;
        case ModeDCUserKey: DCUserKey(b); break;
        case ModeSOS: IgnoreString(b); break;
        case ModeCSI: ControlSequence(b); break;
        case ModeXS:  XSequence(b); break;
        case ModeDLE: DLESeen(b); break;
        case ModeCAN: CANSeen(b); break;
        default:
          ParseMode = ModeFirst;
          ParseFirst(b);
      }
    }

    if (ChangeEmu==0)
      c = CommRead1Byte(&cv,&b);
  }

  UpdateStr();
  UpdateScroll();

  UnlockBuffer();

  ReleaseDC(HVTWin,VTDC);
  VTDC = 0;

  CaretOn();

  if (ChangeEmu > 0) ParseMode = ModeFirst;
  return ChangeEmu;
}

